// 2010GeneticLinearProgramming application
// Copyright Mathew Carr (c) 2010
// licensed under gnu affero gpl license 3.0 or any later version

/*

===============

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, VERSION 3 or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    (It's in agpl3.txt)

===============

*/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cstdarg>

#include <cmath>

#include <string>
#include <list>
#include <vector>
#include <map>

#include <iostream>
#include <sstream>
#include <typeinfo>
#include <stdexcept>

//#undef assert
//
//static inline void assert(bool c)
//{
//   if (c) return;
//
//   printf("assert fail");
//   exit(0);
//}

#undef assert
#define ASSERTIONS
#undef NDEBUG
#include <cassert>

#ifdef use_gnu_gmp

#include <GMPxx.h>

typedef mpz_class BigInteger;

std::string bigIntegerToString(const BigInteger &b)
{
   return b.get_str();
}

int bigIntegerToInt(const BigInteger &b)
{
   return b.get_si();
}

unsigned int bigIntegerToUInt(const BigInteger &b)
{
   return b.get_ui();
}
#else

typedef long long BigInteger;

std::string bigIntegerToString(const BigInteger &b)
{
   static char t[64];
   sprintf(t, "%lld", b);
   return std::string(t);
}

int bigIntegerToInt(const BigInteger &b)
{
   return b;
}

unsigned int bigIntegerToUInt(const BigInteger &b)
{
   return b;
}
#endif

#define wait_for_return() system("pause");

// Return a random float in the range 0.0f <= x < 1.0f;
static inline float random_float()
{
   return (((float)rand()) / (((float)RAND_MAX) + 1.0f));
}

// Return a random float in the range lower <= x < upper
static inline float random_float(float lower, float upper)
{
   return lower + ((upper - lower) * random_float());
}

// Return a random int in the range 0 <= x <= max
static inline unsigned int random_int_0_to_(unsigned int max)
{
   if (max == 0) return 0;

   return (unsigned int)(random_float() * (max + 1));
}

// Return a random int in the range 1 <= x <= max
static inline unsigned int random_int_1_to_(unsigned int max)
{
   assert(!(max == 0));

   if (max == 1) return 1;

   return random_int_0_to_(max-1)+1;
}

// Return a random int in the range lower <= x <= upper
static inline int random_int(int lower, int upper)
{
   return lower + random_int_0_to_(upper - lower);
}

#define zz printf("%d, ", __LINE__)

// Return the greater of a and b.
template <typename T>
static inline const T &max(const T &a, const T &b)
{
   return (a > b) ? a : b;
}

// Return the lesser of a and b.
template <typename T>
static inline const T &min(const T &a, const T &b)
{
   return (a < b) ? a : b;
}

// A structure template representing a pair (lower, upper) of bounds
// of type T.
template <typename T>
struct BoundsPair
{
   T lower, upper;

   BoundsPair(T _lower = T(), T _upper = T())
      : lower(_lower), upper(_upper)
   {
   }

   // Is value within the bounds represented by this instance?
   bool Within(const T &value)
   {
      return ((value >= lower) && (value <= upper));
   }
};

// Return t in the form of a std::string.
template <typename T>
static inline std::string String_ToString(const T& t)
{
   std::ostringstream o;
   o << t;
   return o.str();
}

// Returns true if the character c is present in the
// null terminated string s.
static inline bool char_in_str(const char *s, char c)
{
   while (*s)
   {
      if (*s++ == c) return true;
   }

   return false;
}

std::string String_LoseTrailing(const std::string &string, const char *characters)
{
   std::string out = string;

   while (char_in_str(characters, out[out.length() - 1]))
   {
      out = out.substr(0, out.length() - 1);

      if (out == "") break;
   }

   return out;
}

static inline std::string String_LoseTrailingNewlines(const std::string &string)
{
   return String_LoseTrailing(string, "\r\n");
}

static inline std::string String_LoseTrailingSpaces(const std::string &string)
{
   return String_LoseTrailing(string, " ");
}

// Return as a std::string the nth word in a std::string
// containing a series of words seperated by characters
// contained in the null terminated string delimiters.
std::string String_ReturnWord(const std::string &s, unsigned int word, const char *delimiters = " ")
{
   unsigned int scan_position = 0;
   unsigned int string_start  = 0;

   bool searching_for_start_of_string = true;

   //printf("scanning %s\r\n", s.c_str());

   for (scan_position = 0; scan_position < s.length(); scan_position++)
   {
      if (searching_for_start_of_string)
      {
         if (!char_in_str(delimiters, s[scan_position]))
         {
            string_start = scan_position;
            searching_for_start_of_string = false;
         }
      }
      else
      {
         if (char_in_str(delimiters, s[scan_position]))
         {
            unsigned int string_end = scan_position;

            if (word-- == 0)
            {
               //printf("returning sub string: %s (%s) (%d, %d)", s.substr(string_start, string_end - string_start).c_str(), s.c_str(), string_start, string_end - string_start);

               return String_LoseTrailingSpaces(s.substr(string_start, string_end - string_start));
            }

            searching_for_start_of_string = true;
         }
      }
   }

   // If you're looking for additional words beyond the one
   // that's current in the buffer, return empty!
   if (word >= 1) searching_for_start_of_string = true;

   if (searching_for_start_of_string)
   {
      //printf("returning empty string due to character exhaustion\r\n");

      return std::string();
   }
   else
   {
      //printf("returning half string due to character exhaustion\r\n");

      // Return what we have.
      return String_LoseTrailingSpaces(s.substr(string_start));
   }
}

std::string String_ReturnWordRemainder(const std::string &s, unsigned int word, const char *delimiters = " ")
{
   unsigned int scan_position = 0;
   unsigned int string_start  = 0;

   bool searching_for_start_of_string = true;

   for (scan_position = 0; scan_position < s.length(); scan_position++)
   {
      if (searching_for_start_of_string)
      {
         if (!char_in_str(delimiters, s[scan_position]))
         {
            string_start = scan_position;
            searching_for_start_of_string = false;
         }
      }
      else
      {
         if (char_in_str(delimiters, s[scan_position]))
         {
            unsigned int string_end = scan_position;

            if (word-- == 0)
            {
               std::string remainder = s.substr(string_end + 1);

               while ((remainder.length() > 0)
                   && (char_in_str(delimiters, remainder[0])))
               {
                  remainder = remainder.substr(1);
               }

               return remainder;
            }

            searching_for_start_of_string = true;
         }
      }
   }

   return std::string();
}

unsigned int String_CountWordsInString(const std::string &input)
{
   unsigned int c = 0;

   while (String_ReturnWord(input, c) != "")
   {
      c++;
   }

   return c;
}

///////////////////////

// The symbol table is very importnat

// A SymbolicVariableData records contains information on
// a symbol within the symbol table, such as whether or not
// it can be written to, read from, and whether its value
// is important with respect to establishing the degree
// of semantic correlation.
struct SymbolicVariableData
{
   // std::string allows the symbol table to take ownership of the string data.
   std::string    variable_symbolic_name;

   bool write_access; // Can this variable be written to by STORS
   bool read_access;  // Can this variable be  read   to by LOADS

   // Should the value of this variable match the new state after the parse tree
   // high level evaluation?
   bool expected_match_new_state;
   // Should the value of this variable remain unchanged after the parse tree
   // high level evaluation?
   bool expected_match_old_state;

   // Will the contents of this variable be determinate at
   // the beginning of execution?
   bool contents_determinate;

   // Is this symbol actually a symbolic integer constant (like 5)
   bool symbolic_constant;
   BigInteger value; // If so, what is its value?

   SymbolicVariableData()
   {
      variable_symbolic_name = "";

      write_access = false;
      read_access = false;
      expected_match_new_state = false;
      expected_match_old_state = false;
      contents_determinate = false;
      symbolic_constant = false;

      value = 0;
   }
};

// The archetypes allow for the simple expression of the
// nature of a symbol within the symbol table.
enum SYMBOL_TABLE_VARIABLE_ARCHETYPE
{
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_INVALID,

   // Incoming value as function parameter.      Writable, Readable, Value irrelevant, Determinate.
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_PARAMETER,

   // Incidental variable in calculation         Writable, Readable, Value irrelevant, Indeterminate.
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_VARIABLE,

   // Outgoing value from function parameter.    Writable, Readable, Value should match end state. Indeterminate.
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_RESULT,

   // Outgoing value from function parameter.    Writable, Readable, Value should match end state. Determinate.
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_PARAMETER_RESULT,

   // Constant.                                  Not writable, readable, value should match start state. Determinate.
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_CONSTANT,

   // Symbolic constant.                         Not writable, readable, value should match start state. Determinate.
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_SYMBOLIC_CONSTANT,

   // Variable that is not used in calculation   Not writable, not readable, value should match start state. Indeterminate.
   SYMBOL_TABLE_VARIABLE_ARCHETYPE_BYSTANDER,
};

// Apply a SYMBOL_TABLE_VARIABLE_ARCHETYPE to a SymbolicVariableData
// entry to establish its properties.
void SymbolicVariableData_ApplyArchetype(SymbolicVariableData *symbolic_variable_data, SYMBOL_TABLE_VARIABLE_ARCHETYPE archetype)
{
   switch (archetype)
   {
      case (SYMBOL_TABLE_VARIABLE_ARCHETYPE_PARAMETER) :
      {
         symbolic_variable_data->write_access             =  true;
         symbolic_variable_data->read_access              =  true;
         symbolic_variable_data->expected_match_new_state = false;
         symbolic_variable_data->expected_match_old_state = false;
         symbolic_variable_data->contents_determinate     =  true;
      }
      break;
      case (SYMBOL_TABLE_VARIABLE_ARCHETYPE_VARIABLE) :
      {
         symbolic_variable_data->write_access             =  true;
         symbolic_variable_data->read_access              =  true;
         symbolic_variable_data->expected_match_new_state = false;
         symbolic_variable_data->expected_match_old_state = false;
         symbolic_variable_data->contents_determinate     = false;
      }
      break;
      case (SYMBOL_TABLE_VARIABLE_ARCHETYPE_RESULT) :
      {
         symbolic_variable_data->write_access             =  true;
         symbolic_variable_data->read_access              =  true;
         symbolic_variable_data->expected_match_new_state =  true;
         symbolic_variable_data->expected_match_old_state = false;
         symbolic_variable_data->contents_determinate     = false;
      }
      break;
      case (SYMBOL_TABLE_VARIABLE_ARCHETYPE_PARAMETER_RESULT) :
      {
         symbolic_variable_data->write_access             =  true;
         symbolic_variable_data->read_access              =  true;
         symbolic_variable_data->expected_match_new_state =  true;
         symbolic_variable_data->expected_match_old_state = false;
         symbolic_variable_data->contents_determinate     =  true;
      }
      break;
      case (SYMBOL_TABLE_VARIABLE_ARCHETYPE_CONSTANT) :
      {
         symbolic_variable_data->write_access             = false;
         symbolic_variable_data->read_access              =  true;
         symbolic_variable_data->expected_match_new_state = false;
         symbolic_variable_data->expected_match_old_state =  true;
         symbolic_variable_data->contents_determinate     =  true;
      }
      break;
      case (SYMBOL_TABLE_VARIABLE_ARCHETYPE_SYMBOLIC_CONSTANT) :
      {
         symbolic_variable_data->write_access             = false;
         symbolic_variable_data->read_access              =  true;
         symbolic_variable_data->expected_match_new_state = false;
         symbolic_variable_data->expected_match_old_state =  true;
         symbolic_variable_data->contents_determinate     =  true;
         symbolic_variable_data->symbolic_constant        =  true;
      }
      break;
      case (SYMBOL_TABLE_VARIABLE_ARCHETYPE_BYSTANDER) :
      {
         symbolic_variable_data->write_access             = false;
         symbolic_variable_data->read_access              = false;
         symbolic_variable_data->expected_match_new_state = false;
         symbolic_variable_data->expected_match_old_state =  true;
         symbolic_variable_data->contents_determinate     = false;
      }
      break;
      default :
      {
         assert(false);
      }
   }
}

// Entries within the symbol table are accessed using
// a unique SYMBOL_TABLE_KEY.
typedef unsigned int SYMBOL_TABLE_KEY;

// SymbolTableData maps SYMBOL_TABLE_KEYs to SymbolicVariableData records.
typedef std::map<SYMBOL_TABLE_KEY, SymbolicVariableData> SymbolTableData;

// The (default value of unsiged int) is reserved for a 'bad' SYMBOL_TABLE_KEY
// Any access attempted using this key will result in abort.
const SYMBOL_TABLE_KEY SYMBOL_TABLE_KEY_BAD = 0;

// The symbol table contains information on all possible symbols
// available to a program (it forms a pair with the parse tree)
struct SymbolTable
{
   SYMBOL_TABLE_KEY next_symbol_table_key; // Should be unique for every entry into the symbol table.

   SymbolTableData symbol_table_data;

   bool contains_any_writable_variables;
   bool contains_any_readable_variables;

   SymbolTable()
      : next_symbol_table_key(1),
        contains_any_writable_variables(false),
        contains_any_readable_variables(false)
   {
   }
};

SYMBOL_TABLE_KEY SymbolTable_ReverseLookupKeyFromVarname(SymbolTable *_symbol_table,
                                                          const char *_varname,
                                                                bool  _allow_return_bad = false);

// Create a new symbol in the symbol table with the symbolic name
// variable_symbolic_name and the archetype archetype.
SYMBOL_TABLE_KEY SymbolTable_InsertSymbolTableEntry(
                         SymbolTable *symbol_table,
                   const std::string &variable_symbolic_name,
     SYMBOL_TABLE_VARIABLE_ARCHETYPE  archetype,
                          BigInteger  value = 0)
{
   assert(symbol_table != NULL);
   assert(archetype != SYMBOL_TABLE_VARIABLE_ARCHETYPE_INVALID);

   SYMBOL_TABLE_KEY attempt_to_retrieve_existing = SymbolTable_ReverseLookupKeyFromVarname(symbol_table, variable_symbolic_name.c_str(), true);

   if (attempt_to_retrieve_existing != SYMBOL_TABLE_KEY_BAD)
   {
      return attempt_to_retrieve_existing;
   }

   SYMBOL_TABLE_KEY new_item_key = symbol_table->next_symbol_table_key++;
   assert(new_item_key != SYMBOL_TABLE_KEY_BAD);

try{
   SymbolicVariableData data;

   data.variable_symbolic_name = variable_symbolic_name;
   data.value = value;

   SymbolicVariableData_ApplyArchetype(&data, archetype);

   std::pair<SYMBOL_TABLE_KEY, SymbolicVariableData> a(new_item_key, data);

   symbol_table->symbol_table_data.insert(a);
}catch (...)
{
   printf("Map fail.\r\n");
}
   return new_item_key;
}

// Return a pointer to the SymbolicVariableData containing information about
// the symbol given by the SYMBOL_TABLE_KEY _key
SymbolicVariableData *SymbolTable_QuerySymbolTableKey(SymbolTable *_symbol_table,
                                                            const SYMBOL_TABLE_KEY _key)
{
   assert(_symbol_table != NULL);
   assert(_key != SYMBOL_TABLE_KEY_BAD);

   if (_symbol_table->symbol_table_data.find(_key) != _symbol_table->symbol_table_data.end())
   {
      return (&(_symbol_table->symbol_table_data.find(_key)->second));
   }
   else
   {
      // A query to a non existant variable happened.
      // This shouldn't happen because everything that accesses the symbol
      // table should do so by retrieving the keys from the first column
      // of the symbol_table_data_map.

      printf("Key %d is bad.\r\n", _key);

      assert(false);
      return NULL;
   }
}

// Return the SYMBOL_TABLE_KEY that can be used to access the SymbolicVariableData
// relating to the symbol with symbolic name _varname.
SYMBOL_TABLE_KEY SymbolTable_ReverseLookupKeyFromVarname(SymbolTable *_symbol_table,
                                                          const char *_varname,
                                                                bool  _allow_return_bad)
{
   assert(_symbol_table != NULL);
   assert(_varname != NULL);

   for (SymbolTableData::iterator i = _symbol_table->symbol_table_data.begin();
        i != _symbol_table->symbol_table_data.end();
        i++)
   {
      if (strcmp(i->second.variable_symbolic_name.c_str(), _varname) == 0)
      {
         return i->first;
      }

      //printf("%s: it isn't %s.\r\n", _varname, i->second.variable_symbolic_name.c_str());
   }

   if (_allow_return_bad)
   {
      return SYMBOL_TABLE_KEY_BAD;
   }

   printf(
   "Unrecognised symbolic name %s.\r\n"
   "You may have forgotten to add the variable\r\n"
   "to the symbol table section of the program.\r\n"
   "Make sure to declare all numeric constants as\r\n"
   "SYMBOLIC_CONSTANT!\r\n"
   , _varname);

   assert(false);

   return SYMBOL_TABLE_KEY_BAD;
}

// Return a pointer to the SymbolicVariableData containing information about
// the symbol with symbolic name _varname.
SymbolicVariableData *SymbolTable_QuerySymbolTableVarname(SymbolTable *_symbol_table,
                                                          const char *_varname)
{
   assert(_symbol_table != NULL);
   assert(_varname != NULL);

   return SymbolTable_QuerySymbolTableKey(_symbol_table, SymbolTable_ReverseLookupKeyFromVarname(_symbol_table, _varname));
}

// Updates the contains_any_*_variables members of the symbol table
// to allow for quick inspection.
void SymbolTable_ReaffirmContainingTypes(SymbolTable *symbol_table)
{
   symbol_table->contains_any_readable_variables = false;
   symbol_table->contains_any_writable_variables = false;

   for (SymbolTableData::iterator i  = symbol_table->symbol_table_data.begin();
                                  i != symbol_table->symbol_table_data.end();
                                  i++)
   {
      if (i->second.read_access)  symbol_table->contains_any_readable_variables = true;
      if (i->second.write_access) symbol_table->contains_any_writable_variables = true;
   }
}

// enumeration allowing for the retrieval of a random SYMBOL_TABLE_KEY
// relating to a variable that is readable, writable or both.
enum RANDOM_SYMBOL_TABLE_KEY_NATURE
{
   RANDOM_SYMBOL_TABLE_KEY_ANY,
   RANDOM_SYMBOL_TABLE_KEY_WRITABLE,
   RANDOM_SYMBOL_TABLE_KEY_READABLE,
};

// Return a random SYMBOL_TABLE_KEY from symbol_table of nature nature.
SYMBOL_TABLE_KEY SymbolTable_RandomSymbolTableKey(SymbolTable *symbol_table, RANDOM_SYMBOL_TABLE_KEY_NATURE nature)
{
   assert(symbol_table != NULL);

   if (nature == RANDOM_SYMBOL_TABLE_KEY_READABLE)
   {
      assert(symbol_table->contains_any_readable_variables);
   }
   if (nature == RANDOM_SYMBOL_TABLE_KEY_WRITABLE)
   {
      assert(symbol_table->contains_any_writable_variables);
   }

   SymbolTableData::iterator i;

   do
   {
      unsigned int index = random_int_0_to_(symbol_table->symbol_table_data.size() - 1);

      i = symbol_table->symbol_table_data.begin();

      while (index--) i++;
   }
   while (!((nature == RANDOM_SYMBOL_TABLE_KEY_ANY)
         || ((nature == RANDOM_SYMBOL_TABLE_KEY_READABLE) && (i->second.read_access))
         || ((nature == RANDOM_SYMBOL_TABLE_KEY_WRITABLE) && (i->second.write_access))));

   return i->first;
}

// Print a human readable version of the symbol table to
// standard output.
void SymbolTable_Print(SymbolTable *symbol_table)
{
   for (SymbolTableData::iterator i  = symbol_table->symbol_table_data.begin();
                                  i != symbol_table->symbol_table_data.end();
                                  i++)
   {
      printf("Symbol table key '%2d': ", i->first);                      /* SYMBOL_TABLE_KEY number */
      printf("%-8s ", i->second.variable_symbolic_name.c_str());         /* symbolic variable name */
      printf("Flags: ");                                                 /*  */
      printf("W%c ", i->second.write_access              ? 'Y' : 'N');   /* writable */
      printf("R%c ", i->second.read_access               ? 'Y' : 'N');   /* readable */
      printf("N%c ", i->second.expected_match_new_state  ? 'Y' : 'N');   /* must match new state */
      printf("O%c ", i->second.expected_match_old_state  ? 'Y' : 'N');   /* must match old state */
      printf("D%c ", i->second.contents_determinate      ? 'Y' : 'N');   /* contents are determinate at program start */
      printf("S%c ", i->second.symbolic_constant         ? 'Y' : 'N');   /* is a symbolic constant */
      printf("\r\n");
   }
}

///////////////////////

// Parse tree information definitions:

// I define the parse tree operator like symbols here.
// A PTOLS is a symbol in the parse tree that is not a leaf:
// this includes the semicolon, ternary conditionals, loop structures,
// mathematical operations, and indirection tools.

// I define this metadata to allow me to define actions common
// to all instances of a single given type of PTOLS in the parse tree.
// For example, all addition symbols will act to produce an addition effect.


// Parse tree cell. (the homogeneous return type for parse tree HLVM evaluations)
typedef BigInteger PT_VALUE_TYPE;
typedef SYMBOL_TABLE_KEY PT_REFERENCE_TYPE;

// The type that will be used to manage the complexity heuristic.
typedef unsigned int PT_COMPLEXITY_TYPE;

// These are the complexity contributions of the possible actions.
const PT_COMPLEXITY_TYPE PARSE_TREE_COMPLEXITY_EVALUATING_CONSTANT   = 1; // Loading integer constant from node.
const PT_COMPLEXITY_TYPE PARSE_TREE_COMPLEXITY_INITIALISING_CONSTANT = 1; // Placing constant into memory during fitness case creation.

// A value evaluation of a ParseTreeNode returns a ParseTreeEvaluationResult
// that holds the numeric value of the evaluation and the complexity heuristic
// value of the calculation.
struct ParseTreeEvaluationResult
{
   PT_VALUE_TYPE value;

   PT_COMPLEXITY_TYPE complexity_heuristic;

   ParseTreeEvaluationResult(PT_VALUE_TYPE      _value = 0,
                             PT_COMPLEXITY_TYPE _complexity_heuristic = 0)
      : value(_value),
        complexity_heuristic(_complexity_heuristic)
   {
   }
};

// A reference evaluation ('ereferation') of a ParseTreeNode returns a ParseTreeEreferationResult
// that holds a reference type reflecting the variable that it can be used to access,
// together with a complexity heuristic.
struct ParseTreeEreferationResult
{
   PT_REFERENCE_TYPE reference;

   PT_COMPLEXITY_TYPE complexity_heuristic;

   ParseTreeEreferationResult(PT_REFERENCE_TYPE  _reference = 0,
                              PT_COMPLEXITY_TYPE _complexity_heuristic = 0)
      : reference(_reference),
        complexity_heuristic(_complexity_heuristic)
   {
   }
};

struct ParseTreeInteriorNodeSetMember;
struct ParseTreeLeafNodeSetMember;

// ParseTreeEvaluationEnvironment encapsulates an environment
// in which a program given by a ParseTreeNode may be invoked.
struct ParseTreeEvaluationEnvironment
{
   // An environment is linked to a symbol table.
   // The symbol table should be owned by the calling function
   // not this environment structure.
   SymbolTable *symbol_table;

   // The memory of the environment is implemented with std::map
   std::map<SYMBOL_TABLE_KEY, ParseTreeEvaluationResult> variable_memory;

   // Where the 'return' directive is used (for function-like programs)
   // the returned value and complexity heuristic is placed here.
   ParseTreeEvaluationResult returned_value;

   // Was any value returned at all?
   bool return_occured;

   // Total number of ParseTreeEvaluation/EreferationCalls
   unsigned int evaluation_calls;

   // The user may specify a hard limit to the number of calls
   // to prevent infinite loops where such structures are present.
   static const unsigned int EVALUATIONS_LIMIT;
};

const unsigned int ParseTreeEvaluationEnvironment::EVALUATIONS_LIMIT = 500;

// The maximium number of child nodes from a parse tree
// node. Specifying this in advance establishes an upper limit
// on the size complexity of the parse tree, but also has the
// practical effect of allowing for C array syntax to access
// children (without any dynamic memory stuff)
const unsigned int MAXIMUM_PARSE_TREE_OPERATOR_LIKE_SYMBOL_ARITY = 4;

// This enumeration lists the possible actions that may be
// recorded during high level execution of a parse tree program.
enum LOGGED_OPERATION_PERFORMED
{
   LOGGED_OPERATION_INITIALISATION,

   LOGGED_OPERATION_ADDITION,
   LOGGED_OPERATION_SUBTRACTION,
   LOGGED_OPERATION_MULTIPLICATION,
   LOGGED_OPERATION_DIVISION,

   LOGGED_OPERATION_CONSTANT_LOAD,

   LOGGED_OPERATION_RETURN,

   LOGGED_OPERATION__END_OF_LIST,
};

const char *LOGGED_OPERATION_PERFORMED_STRINGS[LOGGED_OPERATION__END_OF_LIST] =
{
   "LOGGED_OPERATION_INITIALISATION",

   "LOGGED_OPERATION_ADDITION",
   "LOGGED_OPERATION_SUBTRACTION",
   "LOGGED_OPERATION_MULTIPLICATION",
   "LOGGED_OPERATION_DIVISION",

   "LOGGED_OPERATION_CONSTANT_LOAD",

   "LOGGED_OPERATION_RETURN",
};

// A record of a single operation of high level evaluation of a
// program in parse tree form.
struct ParseTreeEvaluationLogEntry
{
   LOGGED_OPERATION_PERFORMED operation;

   // Value and complexity heuristic of the values that were 'consumed'
   // in the calculation
   ParseTreeEvaluationResult operand[2];

   // Value and complexity heuristic of the value produced
   // in the calculation
   ParseTreeEvaluationResult result;

   ParseTreeEvaluationLogEntry(LOGGED_OPERATION_PERFORMED _operation,
                               ParseTreeEvaluationResult _operand_0,
                               ParseTreeEvaluationResult _operand_1,
                               ParseTreeEvaluationResult _result)
      : operation(_operation),
        result(_result)
   {
      operand[0] = _operand_0;
      operand[1] = _operand_1;
   }
};

// Print a ParseTreeEvaluationLogEntry in human readable form
void ParseTreeEvaluationLogEntry_Print(const ParseTreeEvaluationLogEntry *entry)
{
   printf("%-35s: %6s <%1d> to %6s <%1d> giving %6s <%1d>.\r\n",
          LOGGED_OPERATION_PERFORMED_STRINGS[entry->operation],
          bigIntegerToString(entry->operand[0].value).c_str(),
          entry->operand[0].complexity_heuristic,
          bigIntegerToString(entry->operand[1].value).c_str(),
          entry->operand[1].complexity_heuristic,
          bigIntegerToString(entry->result.value).c_str(),
          entry->result.complexity_heuristic);
}

// A log holds a vector of ParseTreeEvaluationLogEntry.
typedef std::vector<ParseTreeEvaluationLogEntry> ParseTreeEvaluationLog;

// A ParseTreeNode structure holds a node of a program in parse tree form.
struct ParseTreeNode
{
   // A ParseTreeNode is either an interior or exterior node.
   enum Nature
   {
      NODE_NATURE_INVALID,
      NODE_NATURE_LEAF_NODE,
      NODE_NATURE_INTERIOR_NODE,
   };

   Nature nature;

   // Based on its nature, the ParseTreeNode will have a different payload.

      // An interior ParseTreeNode will contain a
      // ParseTreeInteriorNodeSetMember indicating the
      // language feature it represents, and pointers to
      // child nodes.
      struct
      {
         const ParseTreeInteriorNodeSetMember *interior_node_master;
         ParseTreeNode *child_node[MAXIMUM_PARSE_TREE_OPERATOR_LIKE_SYMBOL_ARITY];
      } as_interior_node;

      // A leaf ParseTreeNode will contain a
      // ParseTreeLeafNodeSetMember indicating the
      // base symbol or constant it represents.
      struct
      {
         // A pointer to the terminal that this node is one of!
         const ParseTreeLeafNodeSetMember *leaf_node_master;
      } as_leaf_node;
};

#define EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE   \
                                ParseTreeNode *node,              \
               ParseTreeEvaluationEnvironment *environment,       \
                       ParseTreeEvaluationLog *log
#define EREFERATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE  \
                                ParseTreeNode *node,              \
               ParseTreeEvaluationEnvironment *environment,       \
                       ParseTreeEvaluationLog *log

// This enumeration allows for some specification of the human readable form
// of PTOLS language features.
enum PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM
{
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INVALID           = 0xFFFF,

   // Should be followed by a newline.
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NEWLINE      = 0x0100,
   // Should not be surrounded by brackets.
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_BRACKETS  = 0x0200,
   // Should not be surrounded by spaces.
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_SPACES    = 0x0400,

   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FIX_MASK          = 0x000F,
   // Should precede the values upon which it acts
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_PREFIX            = 0x0001,
   // Should be interposed between the values upon which it acts.
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX             = 0x0002,
};

// These are the language constructs (operator-like symbol)
enum ParseTreeOperatorLikeSymbolHandle
{
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_PLUS,
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_MINUS,
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_MULTIPLY,
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_DIVIDE,
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_SEMICOLON,
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_ASSIGNMENT,
   PARSE_TREE_OPERATOR_LIKE_SYMBOL_RETURN,

   PARSE_TREE_OPERATOR_LIKE_SYMBOL__END_OF_LIST,
};

// Forward declaration of the functions callable during high level evaluation.
ParseTreeEvaluationResult  parse_tree_node_evaluate__plus      (EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEvaluationResult  parse_tree_node_evaluate__minus     (EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEvaluationResult  parse_tree_node_evaluate__multiply  (EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEvaluationResult  parse_tree_node_evaluate__divide    (EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEvaluationResult  parse_tree_node_evaluate__semicolon (EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEvaluationResult  parse_tree_node_evaluate__assignment(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEvaluationResult  parse_tree_node_evaluate__return    (EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEreferationResult parse_tree_node_ereferate__invalid  (EREFERATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);

// ParseTreeOperatorLikeSymbolMetadata defines the properties of a language feature
// including arity (two for arithmetic calculations, one for unary minus, etc)
struct ParseTreeOperatorLikeSymbolMetadata
{
   // Arity of operation
   unsigned int arity;

   // Pointers to functions performing the operation.
   ParseTreeEvaluationResult  (*evaluation)(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
   ParseTreeEreferationResult (*ereferation)(EREFERATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);

   // How much complexity does evaluating this operation add?
   PT_COMPLEXITY_TYPE complexity_heuristic;

   /* PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM */
   unsigned int source_syntax_form;

   // How does the parse tree operator like symbol look
   // in the various printed representations?
   const char *detailed_list_description;
   const char *source_description;
};

const unsigned int PARSE_TREE_OPERATOR_LIKE_SYMBOL_COUNT = PARSE_TREE_OPERATOR_LIKE_SYMBOL__END_OF_LIST;

// This constant array acts as a table defining the metadata of each operator
// available as an interior node in the parse tree language.
const ParseTreeOperatorLikeSymbolMetadata parse_tree_operator_like_symbol_metadata[PARSE_TREE_OPERATOR_LIKE_SYMBOL_COUNT] =
{
   { 2, parse_tree_node_evaluate__plus,           parse_tree_node_ereferate__invalid,   1,     PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX
                                                                                             | 0,                                                                        "Addition:",         "+"           },

   { 2, parse_tree_node_evaluate__minus,          parse_tree_node_ereferate__invalid,   1,     PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX
                                                                                             | 0,                                                                        "Subtraction:",      "-"           },

   { 2, parse_tree_node_evaluate__multiply,       parse_tree_node_ereferate__invalid,   1,     PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX
                                                                                             | 0,                                                                        "Multiplication:",   "*"           },

   { 2, parse_tree_node_evaluate__divide,         parse_tree_node_ereferate__invalid,   1,     PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX
                                                                                             | 0,                                                                        "Division:",         "/"           },

   { 2, parse_tree_node_evaluate__semicolon,      parse_tree_node_ereferate__invalid,   0,     PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX
                                                                                             | PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NEWLINE
                                                                                             | PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_SPACES
                                                                                             | PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_BRACKETS,      "- semicolon - ",    ";"           },

   { 2, parse_tree_node_evaluate__assignment,     parse_tree_node_ereferate__invalid,   1,     PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX
                                                                                             /*| PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_BRACKETS*/,  "Assignment:",       "="           },

   { 1, parse_tree_node_evaluate__return,         parse_tree_node_ereferate__invalid,   0,     PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_PREFIX
                                                                                             | 0,                                                                        "Return:",           "return "     },
};

// Forward declarations of the functions that act upon 'a node' to get the value.
// The guts of these functions determines the correct function to perform
// depending on the nature and PTOLS of the node and recursively calculates
// the value and complexity heuristic.
ParseTreeEvaluationResult  ParseTreeNode_Evaluate(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);
ParseTreeEreferationResult ParseTreeNode_Ereferate(EREFERATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE);

// A leaf node in the parse tree can only be a variable name currently.
// Symbolic constants are treated as variables with the same name as their value.
enum ParseTreeLeafNodeNature
{
   PARSE_TREE_LEAF_NODE_NATURE_INVALID,
   PARSE_TREE_LEAF_NODE_NATURE_VARNAME,
};

// An available leaf node 'master node'
// (A global archetype to which a leaf node refers to get its properties)
struct ParseTreeLeafNodeSetMember
{
   ParseTreeLeafNodeNature nature;

      struct
      {
         SYMBOL_TABLE_KEY varname_symbol_table_key;
      } as_varname;

};

// An interior node in the parse tree can only be an operator like symbol.
enum ParseTreeInteriorNodeNature
{
   PARSE_TREE_INTERIOR_NODE_NATURE_INVALID,
   PARSE_TREE_INTERIOR_NODE_NATURE_OPERATOR_LIKE_SYMBOL,
};

// An available interior node 'master node'
// (A global archetype to which an interior node refers to get its properties)
struct ParseTreeInteriorNodeSetMember
{
   ParseTreeInteriorNodeNature nature;

      struct
      {
         ParseTreeOperatorLikeSymbolHandle operator_like_symbol;
      } as_operator_like_symbol;

      // Interior nodes can only be language constructs currently.
};

// These functions actually perform the operation defined by the
// operator like symbol.
ParseTreeEvaluationResult parse_tree_node_evaluate__plus(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   const ParseTreeOperatorLikeSymbolHandle &symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

   ParseTreeEvaluationResult a = ParseTreeNode_Evaluate(node->as_interior_node.child_node[0], environment, log);
   ParseTreeEvaluationResult b = ParseTreeNode_Evaluate(node->as_interior_node.child_node[1], environment, log);

   ParseTreeEvaluationResult result;

   result.value                = a.value + b.value;
   result.complexity_heuristic = max(a.complexity_heuristic, b.complexity_heuristic)
                                  + parse_tree_operator_like_symbol_metadata[symbol].complexity_heuristic;

   log->push_back(ParseTreeEvaluationLogEntry(LOGGED_OPERATION_ADDITION, a, b, result));

   return result;
}

ParseTreeEvaluationResult parse_tree_node_evaluate__minus(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   const ParseTreeOperatorLikeSymbolHandle &symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

   ParseTreeEvaluationResult a = ParseTreeNode_Evaluate(node->as_interior_node.child_node[0], environment, log);
   ParseTreeEvaluationResult b = ParseTreeNode_Evaluate(node->as_interior_node.child_node[1], environment, log);

   ParseTreeEvaluationResult result;

   result.value                = a.value - b.value;
   result.complexity_heuristic = max(a.complexity_heuristic, b.complexity_heuristic)
                                  + parse_tree_operator_like_symbol_metadata[symbol].complexity_heuristic;

   log->push_back(ParseTreeEvaluationLogEntry(LOGGED_OPERATION_SUBTRACTION, a, b, result));

   return result;
}

ParseTreeEvaluationResult parse_tree_node_evaluate__multiply(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   const ParseTreeOperatorLikeSymbolHandle &symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

   ParseTreeEvaluationResult a = ParseTreeNode_Evaluate(node->as_interior_node.child_node[0], environment, log);
   ParseTreeEvaluationResult b = ParseTreeNode_Evaluate(node->as_interior_node.child_node[1], environment, log);

   ParseTreeEvaluationResult result;

   result.value                = a.value * b.value;
   result.complexity_heuristic = max(a.complexity_heuristic, b.complexity_heuristic)
                                  + parse_tree_operator_like_symbol_metadata[symbol].complexity_heuristic;

   log->push_back(ParseTreeEvaluationLogEntry(LOGGED_OPERATION_MULTIPLICATION, a, b, result));

   return result;
}

ParseTreeEvaluationResult parse_tree_node_evaluate__divide(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   const ParseTreeOperatorLikeSymbolHandle &symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

   ParseTreeEvaluationResult a = ParseTreeNode_Evaluate(node->as_interior_node.child_node[0], environment, log);
   ParseTreeEvaluationResult b = ParseTreeNode_Evaluate(node->as_interior_node.child_node[1], environment, log);

   ParseTreeEvaluationResult result;

   result.value                = a.value / (b.value != 0 ? b.value : 1);

   result.complexity_heuristic = max(a.complexity_heuristic, b.complexity_heuristic)
                                  + parse_tree_operator_like_symbol_metadata[symbol].complexity_heuristic;

   log->push_back(ParseTreeEvaluationLogEntry(LOGGED_OPERATION_DIVISION, a, b, result));

   return result;
}

// Execute arg1 then arg2
ParseTreeEvaluationResult parse_tree_node_evaluate__semicolon(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   const ParseTreeOperatorLikeSymbolHandle &symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

   ParseTreeEvaluationResult a = ParseTreeNode_Evaluate(node->as_interior_node.child_node[0], environment, log);
   ParseTreeEvaluationResult b = ParseTreeNode_Evaluate(node->as_interior_node.child_node[1], environment, log);

   ParseTreeEvaluationResult result;

   result.value                = b.value;
   result.complexity_heuristic = max(a.complexity_heuristic, b.complexity_heuristic)
                                  + parse_tree_operator_like_symbol_metadata[symbol].complexity_heuristic;

   return result;
}

ParseTreeEvaluationResult parse_tree_node_evaluate__assignment(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   const ParseTreeOperatorLikeSymbolHandle &symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

   ParseTreeEvaluationResult  rvalue = ParseTreeNode_Evaluate(node->as_interior_node.child_node[1], environment, log);

   ParseTreeEreferationResult lvalue = ParseTreeNode_Ereferate(node->as_interior_node.child_node[0], environment, log);

   ParseTreeEvaluationResult result;

   result.value                = rvalue.value;
   result.complexity_heuristic = max(lvalue.complexity_heuristic, rvalue.complexity_heuristic)
                                  + parse_tree_operator_like_symbol_metadata[symbol].complexity_heuristic;

   environment->variable_memory[lvalue.reference] = result;

   return result;
}

ParseTreeEvaluationResult parse_tree_node_evaluate__return(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   const ParseTreeOperatorLikeSymbolHandle &symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

   ParseTreeEvaluationResult value = ParseTreeNode_Evaluate(node->as_interior_node.child_node[0], environment, log);

   value.complexity_heuristic += parse_tree_operator_like_symbol_metadata[symbol].complexity_heuristic;

   log->push_back(ParseTreeEvaluationLogEntry(LOGGED_OPERATION_RETURN, value, value, value));

   environment->returned_value = value;
   environment->return_occured = true;

   return value;
}

// You can't ereferate many of the operator like symbols in the language:
ParseTreeEreferationResult parse_tree_node_ereferate__invalid(EREFERATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   assert(false);

   return ParseTreeEreferationResult();
}
// Ereferation returns a reference version of the variable,
// so you can ereferate 'a' to get a reference type that
// allows assignment to 'a', or you can evaluate 'a' to get
// the value of 'a' at the present time.

// You can't ereferate 'a + 2', but you can evaluate it.
// If you were able to ereferate 'a + 2', you'd be able to write
// 'a + 2 := 5;' in the language!



// A ParseTreeLeafNodeSet is a set of ParseTreeLeafNodeSetMembers.
// These are the primitives that one can build a parse tree out of.
// You have to define all these PTLNSMs first, and then use a
// sort of 'instantiation' function upon them to get the
// actual nodes for inclusion in the parse tree.
ParseTreeLeafNodeSetMember *ParseTreeLeafNodeSetMember_NewParseTreeLeafNodeSetMember_Internal_NewParseTreeLeafNodeSetMember()
{
   ParseTreeLeafNodeSetMember *parse_tree_leaf_node_set_member = new ParseTreeLeafNodeSetMember;

   assert(parse_tree_leaf_node_set_member != NULL);

   return parse_tree_leaf_node_set_member;
}

// Make a new parse tree leaf node set member that, when invoked,
// can create a ParseTreeNode of the varname variety.
ParseTreeLeafNodeSetMember *ParseTreeLeafNodeSetMember_NewParseTreeLeafNodeSetMember_VariableFromSymbolTableKey(SYMBOL_TABLE_KEY varkey)
{
   assert(varkey != SYMBOL_TABLE_KEY_BAD);

   ParseTreeLeafNodeSetMember *parse_tree_leaf_node_set_member = ParseTreeLeafNodeSetMember_NewParseTreeLeafNodeSetMember_Internal_NewParseTreeLeafNodeSetMember();

   parse_tree_leaf_node_set_member->nature = PARSE_TREE_LEAF_NODE_NATURE_VARNAME;

   parse_tree_leaf_node_set_member->as_varname.varname_symbol_table_key = varkey;

   return parse_tree_leaf_node_set_member;
}

// Allocate memory for a ParseTreeInteriorNodeSetMember
ParseTreeInteriorNodeSetMember *ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_Internal_NewParseTreeInteriorNodeSetMember()
{
   ParseTreeInteriorNodeSetMember *parse_tree_interior_node_set_member = new ParseTreeInteriorNodeSetMember;

   assert(parse_tree_interior_node_set_member != NULL);

   return parse_tree_interior_node_set_member;
}

// Make a new parse tree interior node set member that, when invoked,
// can create a ParseTreeNode of the operator like symbol variety.
ParseTreeInteriorNodeSetMember *ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(ParseTreeOperatorLikeSymbolHandle operator_like_symbol)
{
   assert((unsigned int)operator_like_symbol < PARSE_TREE_OPERATOR_LIKE_SYMBOL_COUNT);

   ParseTreeInteriorNodeSetMember *parse_tree_interior_node_set_member = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_Internal_NewParseTreeInteriorNodeSetMember();

   parse_tree_interior_node_set_member->nature = PARSE_TREE_INTERIOR_NODE_NATURE_OPERATOR_LIKE_SYMBOL;

   parse_tree_interior_node_set_member->as_operator_like_symbol.operator_like_symbol = operator_like_symbol;

   return parse_tree_interior_node_set_member;
}

////////////

// Return the arity of the language feature reflected by the ParseTreeNode node.
unsigned int ParseTreeNode_OperatorLikeSymbol_ReturnArity(const ParseTreeNode *node)
{
   assert(node != NULL);
   assert(node->nature == ParseTreeNode::NODE_NATURE_INTERIOR_NODE);

   switch (node->as_interior_node.interior_node_master->nature)
   {
      case (PARSE_TREE_INTERIOR_NODE_NATURE_OPERATOR_LIKE_SYMBOL) :
      {
         // Interrogate the 'master node' referred to by this node.
         return parse_tree_operator_like_symbol_metadata[node->as_interior_node.interior_node_master->nature].arity;
      }
      break;
      default :
      {
         assert(false);
         return 0;
      }
   }

   return 0;
}

// Set a ParseTreeEvaluationEnvironment to default values before execution.
// Copy the values of the symbolic constants from the symbol table to their
// respective places in the map.
void ParseTreeEvaluationEnvironment_Initialise(ParseTreeEvaluationEnvironment *environment, SymbolTable *symbol_table)
{
   assert(environment != NULL);
   assert(symbol_table != NULL);

   environment->symbol_table = symbol_table;

   environment->evaluation_calls = 0;

   environment->return_occured = false;
   environment->returned_value = ParseTreeEvaluationResult();

   for (SymbolTableData::iterator i  = environment->symbol_table->symbol_table_data.begin();
                                  i != environment->symbol_table->symbol_table_data.end();
                                  i++)
   {
      if (i->second.symbolic_constant)
      {
         environment->variable_memory[i->first].value = i->second.value;
      }
      else
      {
         environment->variable_memory[i->first].value = random_int(-100000, 100000);
      }
   }
}

// Evaluates the program given by a given ParseTreeNode in the specified
// environment. Returns a ParseTreeEvaluationResult.
ParseTreeEvaluationResult ParseTreeNode_Evaluate(EVALUATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   assert(node != NULL);
   assert(environment != NULL);

   if (environment->evaluation_calls == ParseTreeEvaluationEnvironment::EVALUATIONS_LIMIT) return ParseTreeEvaluationResult();
   environment->evaluation_calls++;

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         // The item number of this function.
         // Get the metadata associated with this function.
         // Get the evaluation function of this metadata.
         // Invoke this function on the environment and arguments.
         switch (node->as_interior_node.interior_node_master->nature)
         {
            case (PARSE_TREE_INTERIOR_NODE_NATURE_OPERATOR_LIKE_SYMBOL) :
            {
               const ParseTreeInteriorNodeSetMember *interior_node_master = node->as_interior_node.interior_node_master;

               ParseTreeOperatorLikeSymbolHandle operator_like_symbol = interior_node_master->as_operator_like_symbol.operator_like_symbol;

               return (*(parse_tree_operator_like_symbol_metadata[operator_like_symbol].evaluation))(node, environment, log);
            }
            break;
            default :
            {
               assert(false);
            }
         }
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         switch (node->as_leaf_node.leaf_node_master->nature)
         {
            case (PARSE_TREE_LEAF_NODE_NATURE_VARNAME) :
            {
               return environment->variable_memory[node->as_leaf_node.leaf_node_master->as_varname.varname_symbol_table_key];
            }
            break;
            default :
            {
               assert(false);
            }
            break;
         }
      }
      break;
      default :
      {
         assert(false);
      }
      break;
   }

   return ParseTreeEvaluationResult();
}


// Evaluates the program given by a given ParseTreeNode in the specified
// environment. Returns a ParseTreeEvaluationResult.
ParseTreeEreferationResult ParseTreeNode_Ereferate(EREFERATION_OF_PARSE_TREE_OPERATOR_LIKE_SYMBOL_SIGNATURE)
{
   assert(node != NULL);
   assert(environment != NULL);

   if (environment->evaluation_calls == ParseTreeEvaluationEnvironment::EVALUATIONS_LIMIT) return ParseTreeEreferationResult();
   environment->evaluation_calls++;

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         switch (node->as_interior_node.interior_node_master->nature)
         {
            case (PARSE_TREE_INTERIOR_NODE_NATURE_OPERATOR_LIKE_SYMBOL) :
            {
               const ParseTreeInteriorNodeSetMember *interior_node_master = node->as_interior_node.interior_node_master;

               ParseTreeOperatorLikeSymbolHandle operator_like_symbol = interior_node_master->as_operator_like_symbol.operator_like_symbol;

               return (*(parse_tree_operator_like_symbol_metadata[operator_like_symbol].ereferation))(node, environment, log);
            }
            break;
            default :
            {
               assert(false);
            }
         }
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         switch (node->as_leaf_node.leaf_node_master->nature)
         {
            case (PARSE_TREE_LEAF_NODE_NATURE_VARNAME) :
            {
               return ParseTreeEreferationResult(node->as_leaf_node.leaf_node_master->as_varname.varname_symbol_table_key,
                                                 1);
            }
            break;
            default :
            {
               assert(false);
            }
            break;
         }
      }
      break;
      default :
      {
         assert(false);
      }
      break;
   }

   return ParseTreeEreferationResult();
}

// Prints out high level source-like version of the input ParseTreeNode tree
// You need to perform a newline manually after this.
void ParseTreeNode_Print_AsReadableExpression(const ParseTreeNode *node, SymbolTable *symbol_table)
{
   assert(node != NULL);
   assert(symbol_table != NULL);

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(node);

         switch (node->as_interior_node.interior_node_master->nature)
         {
            case (PARSE_TREE_INTERIOR_NODE_NATURE_OPERATOR_LIKE_SYMBOL) :
            {
               unsigned int source_syntax_form = parse_tree_operator_like_symbol_metadata[node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol].source_syntax_form;

               switch (source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FIX_MASK)
               {
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_PREFIX) :
                  {
                     printf("%s", parse_tree_operator_like_symbol_metadata[node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol].source_description);

                     if (!(source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_BRACKETS))
                     {
                        printf("(");
                     }

                     bool first = true;

                     for (unsigned int child_node = 0; child_node < arity; child_node++)
                     {
                        assert(node ->as_interior_node.child_node[child_node] != NULL);

                        if (first == false)
                        {
                           printf(", ");
                        }

                        first = false;

                        ParseTreeNode_Print_AsReadableExpression(node->as_interior_node.child_node[child_node], symbol_table);
                     }

                     if (!(source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_BRACKETS))
                     {
                        printf(")");
                     }
                  }
                  break;
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX) :
                  {
                     if (!(source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_BRACKETS))
                     {
                        printf("(");
                     }

                     bool first = true;

                     for (unsigned int child_node = 0; child_node < arity; child_node++)
                     {
                        if (first == false)
                        {
                           if (!(source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_SPACES))
                           {
                              printf(" ");
                           }

                           printf("%s", parse_tree_operator_like_symbol_metadata[node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol].source_description);

                           if (!(source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_SPACES))
                           {
                              printf(" ");
                           }

                           if (source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NEWLINE)
                           {
                              printf("\r\n");
                           }
                        }

                        first = false;

                        ParseTreeNode_Print_AsReadableExpression(node->as_interior_node.child_node[child_node], symbol_table);
                     }

                     if (!(source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FLAG_NO_BRACKETS))
                     {
                        printf(")");
                     }
                  }
                  break;
                  default:
                  {
                     assert(false);
                  }
               }
            }
            break;
            default :
            {
               assert(false);
            }
         }
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         printf("%s", SymbolTable_QuerySymbolTableKey(symbol_table, node->as_leaf_node.leaf_node_master->as_varname.varname_symbol_table_key)->variable_symbolic_name.c_str());
      }
      break;
      default :
      {
         assert(false);
      }
      break;
   }
}

// Prints out high level source-like version of the input ParseTreeNode tree
// You need to perform a newline manually after this.
std::string ParseTreeNode_Print_ToString(const ParseTreeNode *node, SymbolTable *symbol_table)
{
   assert(node != NULL);
   assert(symbol_table != NULL);

   std::string out;

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(node);

         switch (node->as_interior_node.interior_node_master->nature)
         {
            case (PARSE_TREE_INTERIOR_NODE_NATURE_OPERATOR_LIKE_SYMBOL) :
            {
               unsigned int source_syntax_form = parse_tree_operator_like_symbol_metadata[node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol].source_syntax_form;

               switch (source_syntax_form & PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_FIX_MASK)
               {
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_PREFIX) :
                  {
                     out += std::string(parse_tree_operator_like_symbol_metadata[node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol].source_description);

                        out += std::string("(");

                     bool first = true;

                     for (unsigned int child_node = 0; child_node < arity; child_node++)
                     {
                        assert(node ->as_interior_node.child_node[child_node] != NULL);

                        if (first == false)
                        {
                           out += std::string(",");
                        }

                        first = false;

                        out += ParseTreeNode_Print_ToString(node->as_interior_node.child_node[child_node], symbol_table);
                     }

                        out += std::string(")");
                  }
                  break;
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_SOURCE_SYNTAX_FORM_INFIX) :
                  {
                        out += std::string("(");

                     bool first = true;

                     for (unsigned int child_node = 0; child_node < arity; child_node++)
                     {
                        if (first == false)
                        {
                           out += std::string(parse_tree_operator_like_symbol_metadata[node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol].source_description);
                        }

                        first = false;

                        out += ParseTreeNode_Print_ToString(node->as_interior_node.child_node[child_node], symbol_table);
                     }

                        out += std::string(")");
                  }
                  break;
                  default:
                  {
                     assert(false);
                  }
               }
            }
            break;
            default :
            {
               assert(false);
            }
         }
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         out += std::string(SymbolTable_QuerySymbolTableKey(symbol_table, node->as_leaf_node.leaf_node_master->as_varname.varname_symbol_table_key)->variable_symbolic_name.c_str());
      }
      break;
      default :
      {
         assert(false);
      }
      break;
   }

   return out;
}

// Switches the nodes pointed to by the pointers pointed to by a and b.
void SwitchNodes(ParseTreeNode **a, ParseTreeNode **b)
{
   assert(a != NULL);
   assert(b != NULL);

   ParseTreeNode *t = *a;
   *a = *b;
   *b = t;
}

// Creates a new ParseTreeNode
// Don't call this directly unless you're manually constructing
// a node tree.
ParseTreeNode *ParseTreeNode_Internal_NewNode()
{
   return new ParseTreeNode;
}

// Free the memory used to store a ParseTreeNode.
// Don't call this directly.
void ParseTreeNode_Internal_FreeParseTreeNode(ParseTreeNode *node)
{
   assert(node != NULL);

   delete node;
}

// Free all the memory allocated to the full ParseTreeNode tree
// contained within node n.
// This only works if the tree doesn't contain any undefined children.
// If the tree contains pointers that point to bad nodes
// your program will crash.
void ParseTreeNode_DestroyTree(ParseTreeNode *n)
{
   assert(n != NULL);

   switch (n->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(n);

         for (unsigned int child_node = 0; child_node < arity; child_node++)
         {
            assert(n->as_interior_node.child_node[child_node] != NULL);
            ParseTreeNode_DestroyTree(n->as_interior_node.child_node[child_node]);
         }
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
      }
      break;
      default :
      {
         assert(false);
      }
      break;
   }

   ParseTreeNode_Internal_FreeParseTreeNode(n);
}

// Returns the depth of the tree. A leaf node counts as 1.
unsigned int ParseTreeNode_GetHighestDepth(const ParseTreeNode *node)
{
   assert(node != NULL);

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         unsigned int d = 1;

         unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(node);

         for (unsigned int child_node = 0; child_node < arity; child_node++)
         {
            assert(node->as_interior_node.child_node[child_node] != NULL);
            d = max(d, 1+ParseTreeNode_GetHighestDepth(node->as_interior_node.child_node[child_node]));
         }

         return d;
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         return 1;
      }
      break;
      default :
      {
         assert(false);
      }
      break;
   }

   return 0;
}

// Return the total number of nodes in the given tree, including the
// root.
unsigned int ParseTreeNode_GetTotalNumberOfNodesInTree(const ParseTreeNode *node)
{
   assert(node != NULL);

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         unsigned int d = 1;

         unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(node);

         for (unsigned int child_node = 0; child_node < arity; child_node++)
         {
            assert(node->as_interior_node.child_node[child_node] != NULL);
            d += ParseTreeNode_GetTotalNumberOfNodesInTree(node->as_interior_node.child_node[child_node]);
         }

         return d;
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         return 1;
      }
      break;
      default :
      {
         assert(false);
      }
      break;
   }

   return 0;
}

// Returns a shallow of the tree pointed to by node.
// You need to manually fix the children!!
ParseTreeNode *ParseTreeNode_DuplicateNode(const ParseTreeNode *node)
{
   assert(node != NULL);

   ParseTreeNode *new_tree = ParseTreeNode_Internal_NewNode();

   new_tree->nature = node->nature;

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         new_tree->as_interior_node.interior_node_master = node->as_interior_node.interior_node_master;
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         new_tree->as_leaf_node = node->as_leaf_node;
      }
      break;
      default :
      {
         assert(false);
      }
      break;

   }

   return new_tree;
}

// Returns a DEEP COPY of the tree pointed to by node.
ParseTreeNode *ParseTreeNode_DuplicateTree(const ParseTreeNode *node)
{
   assert(node != NULL);

   ParseTreeNode *new_tree = ParseTreeNode_Internal_NewNode();

   new_tree->nature = node->nature;

   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         new_tree->as_interior_node.interior_node_master = node->as_interior_node.interior_node_master;

         unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(node);

         for (unsigned int child_node = 0; child_node < arity; child_node++)
         {
            new_tree->as_interior_node.child_node[child_node] = ParseTreeNode_DuplicateTree(node->as_interior_node.child_node[child_node]);
         }
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         new_tree->as_leaf_node = node->as_leaf_node;
      }
      break;
      default :
      {
         assert(false);
      }
      break;

   }

   return new_tree;
}

// Instantiate a new ParseTreeNode given the 'master node' ParseTreeLeafNodeSetMember leaf_node_master.
ParseTreeNode *ParseTreeNode_NewLeafNodeFromParseTreeLeafNodeSetMember(const ParseTreeLeafNodeSetMember *leaf_node_master)
{
   assert(leaf_node_master != NULL);

   ParseTreeNode *n = ParseTreeNode_Internal_NewNode();
   assert(n != NULL);

   n->nature = ParseTreeNode::NODE_NATURE_LEAF_NODE;

   n->as_leaf_node.leaf_node_master = leaf_node_master;

   return n;
}

// Instantiate a new ParseTreeNode given the 'master node' ParseTreeInteriorNodeSetMember interior_node_master.
// Don't forget to instantiate some children.
ParseTreeNode *ParseTreeNode_NewIncompleteInteriorNodeFromParseTreeInteriorNodeSetMember(const ParseTreeInteriorNodeSetMember *interior_node_master)
{
   assert(interior_node_master != NULL);

   ParseTreeNode *n = ParseTreeNode_Internal_NewNode();
   assert(n != NULL);

   n->nature = ParseTreeNode::NODE_NATURE_INTERIOR_NODE;

   n->as_interior_node.interior_node_master = interior_node_master;

   return n;
}

////////////////////

// A ParseTreeSourceProgram consists of a program and symbol table pair.
struct ParseTreeSourceProgram
{
   // Pointers to the master nodes referred to by the parse tree program
   // are stored in these maps!
   // You can get the master node of an interior node or leaf node
   // by interrogating these maps.
   std::map<std::string,      ParseTreeInteriorNodeSetMember *> interior_node_set;
   std::map<SYMBOL_TABLE_KEY, ParseTreeLeafNodeSetMember *>     leaf_node_set;

   ParseTreeNode *target_program;

   SymbolTable *symbol_table;
};

// Destroys the entire symbol table and parse tree program pair
// of a program, together with all of the master nodes!
void ParseTreeSourceProgram_DestroySourceProgram(ParseTreeSourceProgram *parse_tree_source_program)
{
   assert(parse_tree_source_program != NULL);

   // Destroy the source program.
   ParseTreeNode_DestroyTree(parse_tree_source_program->target_program);

   // Destroy the master things.
   // We grouped them together into an interior and leaf node sets so
   // that we can destroy all of the members in one swoop, followed
   // by their containing sets, like this!
   for (std::map<std::string, ParseTreeInteriorNodeSetMember *>::iterator i  = parse_tree_source_program->interior_node_set.begin();
                                                                          i != parse_tree_source_program->interior_node_set.end();
                                                                          i++)
   {
      delete i->second;
   }
   for (std::map<SYMBOL_TABLE_KEY, ParseTreeLeafNodeSetMember *>::iterator i  = parse_tree_source_program->leaf_node_set.begin();
                                                                           i != parse_tree_source_program->leaf_node_set.end();
                                                                           i++)
   {
      delete i->second;
   }

   delete parse_tree_source_program->symbol_table;

   delete parse_tree_source_program;
}

////////////////////
////////////////////

std::string String_RemoveOuterBracketsFromString(const std::string &input)
{
   if ((input.at(                 0) == '(')
    && (input.at(input.length() - 1) == ')'))
   {
      unsigned int bracketing_level = 1;

      for (unsigned int i = 1; i < (input.length() - 1); i++)
      {
         if (input.at(i) == '(') bracketing_level++;
         if (input.at(i) == ')') bracketing_level--;

         if (bracketing_level == 0) return input; // This isn't a solid string with brackets either side.
      }

      return String_RemoveOuterBracketsFromString(input.substr(1, input.length() - 2));
   }
   else
   {
      return input;
   }
}

// Transforms a high level program string in infix form
// into a string in prefix form using the priorities map
// to define precedence.
std::string String_PrefixifySourceCodeString(const std::string &input)
{
   static std::map<char, int> priorities;

   static bool initialised = false;

   if (initialised == false)
   {
      priorities[';'] = 10;
      priorities['='] =  9;
      priorities['+'] =  8;
      priorities['-'] =  7;
      priorities['*'] =  6;
      priorities['/'] =  5;
      priorities['@'] = -1;

      initialised = true;
   }

   std::string input_nb = String_RemoveOuterBracketsFromString(input);

   char lowest_char_type = '@';

   unsigned int lowest_char_index = 0;

   unsigned int bracketing_level = 0;

   for (unsigned int i = 0; i < input_nb.length(); i++)
   {
      if (input_nb.at(i) == '(') bracketing_level++;
      if (input_nb.at(i) == ')') bracketing_level--;

      if (bracketing_level == 0)
      {
         if (priorities[input_nb.at(i)] > priorities[lowest_char_type])
         {
            lowest_char_type  = input_nb.at(i);
            lowest_char_index = i;
         }
      }
   }

   // Is this a string of letters, or numbers?
   if (priorities[lowest_char_type] <= 0)
   {
      return input_nb;
   }

   // It's two sections separated by a symbol.
   return   String_ToString(lowest_char_type) + String_ToString(' ')
          + String_PrefixifySourceCodeString(input_nb.substr(0, lowest_char_index)) + String_ToString(' ')
          + String_PrefixifySourceCodeString(input_nb.substr(lowest_char_index + 1));
}

std::string String_RemoveWhitespacesFromString(const std::string &input)
{
   std::string output;

   static std::map<char, bool> discard_characters;

   static bool initialised = false;

   if (initialised == false)
   {
      discard_characters[' ' ] = true;
      discard_characters['\r'] = true;
      discard_characters['\n'] = true;
      discard_characters['\t'] = true;

      initialised = true;
   }

   for (unsigned int i = 0; i < input.length(); i++)
   {
      if (discard_characters[input.at(i)])
      {
         continue;
      }

      output += input.at(i);
   }

   return output;
}

// Each variable in the input high level source program
// has a nature defined by the user.
enum INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE
{
   // Invalid.
   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_NO_NATURE,
   // This variable is only used as input to the program.
   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT,
   // This variable is only used as output.
   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_OUTPUT,
   // This variable is both input and output.
   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT_OUTPUT,
   // The value of this variable is constant during execution
   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_CONSTANT,
   // This variable is a symbolic constants with the same value as its symbolic name.
   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_SYMBOLIC_CONSTANT,
   // This variable is used to hold intermediate values.
   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INTERMEDIATE,
};

// Rudimentary symbol table record for the variables
// in the high level source program.
struct InputSourceProgramSymbolicVariableData
{
   std::string variable_name;

   INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE nature;
};

// Constructing a parse tree from a string of symbols
// requires a persistent structure to keep track of the
// the construction (current symbol, and string)
struct ParseTreeCreationStatus
{
   std::string input_source_code_prefix_form;

   unsigned int current_symbol_considered;

   // The ParseTreeSourceProgram contains the symbol table
   // and the nodes we use to define ParseTreeNodes
   // (the master stuff)
   ParseTreeSourceProgram *program;
};

ParseTreeNode *ParseTreeCreation_CreateParseTreeNode(ParseTreeCreationStatus *status);

// Returns NULL on error.
// Creates a ParseTreeSourceProgram containing both symbol table and parse tree
// from a plaintext high level source file containing the program.
ParseTreeSourceProgram *ParseTreeSourceProgram_ConstructSourceProgramFromFile(const std::string &input_filename)
{
   FILE *i = fopen(input_filename.c_str(), "rb");

   if (i == NULL)
   {
      printf("Input program file does not exist.\r\n");
      return NULL;
   }

   bool reading_program = false;

   std::string input_source_code_read;

   // Read ther lines from the file.
   // If we reach a PROGRAM directive, switch to program reading.

   char read_line[500];
   memset(read_line, 0, 500);

   std::vector<InputSourceProgramSymbolicVariableData> variables;

   do
   {
      fgets(read_line, 500, i);

      if (feof(i)) break;

      std::string read_line_string = String_LoseTrailingNewlines(String_ToString(read_line));
      std::string input_variable_list;

      INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE variable_list_nature = INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_NO_NATURE;

      if (reading_program)
      {
         if (read_line_string.substr(0, strlen("ENDPROGRAM")) == "ENDPROGRAM")
         {
            reading_program = false;
         }
         else
         {
            input_source_code_read += read_line_string;
         }
      }
      else
      if (read_line_string.substr(0, strlen("PROGRAM")) == "PROGRAM")
      {
         reading_program = true;
      }
      else
      if (read_line_string.substr(0, strlen("INPUT ")) == "INPUT ")
      {
         input_variable_list = read_line_string.substr(strlen("INPUT "));

         variable_list_nature = INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT;
      }
      else
      if (read_line_string.substr(0, strlen("OUTPUT ")) == "OUTPUT ")
      {
         input_variable_list = read_line_string.substr(strlen("OUTPUT "));

         variable_list_nature = INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_OUTPUT;
      }
      else
      if (read_line_string.substr(0, strlen("INPUT_OUTPUT ")) == "INPUT_OUTPUT ")
      {
         input_variable_list = read_line_string.substr(strlen("INPUT_OUTPUT "));

         variable_list_nature = INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT_OUTPUT;
      }
      else
      if (read_line_string.substr(0, strlen("CONSTANT ")) == "CONSTANT ")
      {
         input_variable_list = read_line_string.substr(strlen("CONSTANT "));

         variable_list_nature = INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_CONSTANT;
      }
      else
      if (read_line_string.substr(0, strlen("SYMBOLIC_CONSTANT ")) == "SYMBOLIC_CONSTANT ")
      {
         input_variable_list = read_line_string.substr(strlen("SYMBOLIC_CONSTANT "));

         variable_list_nature = INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_SYMBOLIC_CONSTANT;
      }
      else
      if (read_line_string.substr(0, strlen("INTERMEDIATE ")) == "INTERMEDIATE ")
      {
         input_variable_list = read_line_string.substr(strlen("INTERMEDIATE "));

         variable_list_nature = INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INTERMEDIATE;
      }

      if (variable_list_nature != INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_NO_NATURE)
      {
         unsigned int words = String_CountWordsInString(input_variable_list);

         for (unsigned int w = 0; w < words; w++)
         {
            std::string variable_name = String_ReturnWord(input_variable_list, w);

            variable_name = String_RemoveWhitespacesFromString(variable_name);

            InputSourceProgramSymbolicVariableData v;

            v.variable_name = variable_name;
            v.nature = variable_list_nature;

            variables.push_back(v);
         }
      }
   } while (true);

   fclose(i);

   i = NULL;

   std::string input_source_code_nospaces = String_RemoveWhitespacesFromString(input_source_code_read);
   std::string input_source_code_prefix   = String_PrefixifySourceCodeString(input_source_code_nospaces);

   printf("Read:\r\n%s\r\n",     input_source_code_read.c_str());
   printf("Nospaced:\r\n%s\r\n", input_source_code_nospaces.c_str());
   printf("Prefixed:\r\n%s\r\n", input_source_code_prefix.c_str());

   printf("--\r\n");

   for (std::vector<InputSourceProgramSymbolicVariableData>::iterator i = variables.begin(); i != variables.end(); i++)
   {
      printf("Variable %10s is an %s.\r\n", i->variable_name.c_str(),
                                  i->nature == INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT             ? "INPUT"
                                : i->nature == INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_OUTPUT            ? "OUTPUT"
                                : i->nature == INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT_OUTPUT      ? "INPUT_OUTPUT"
                                : i->nature == INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_CONSTANT          ? "CONSTANT"
                                : i->nature == INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_SYMBOLIC_CONSTANT ? "SYMBOLIC_CONSTANT"
                                : i->nature == INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INTERMEDIATE      ? "INTERMEDIATE"
                                : "MYSTERY");
   }

   ParseTreeSourceProgram *program = new ParseTreeSourceProgram;

   ParseTreeInteriorNodeSetMember *operator_like_symbol_plus       = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(PARSE_TREE_OPERATOR_LIKE_SYMBOL_PLUS);
   ParseTreeInteriorNodeSetMember *operator_like_symbol_minus      = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(PARSE_TREE_OPERATOR_LIKE_SYMBOL_MINUS);
   ParseTreeInteriorNodeSetMember *operator_like_symbol_multiply   = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(PARSE_TREE_OPERATOR_LIKE_SYMBOL_MULTIPLY);
   ParseTreeInteriorNodeSetMember *operator_like_symbol_divide     = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(PARSE_TREE_OPERATOR_LIKE_SYMBOL_DIVIDE);
   ParseTreeInteriorNodeSetMember *operator_like_symbol_semicolon  = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(PARSE_TREE_OPERATOR_LIKE_SYMBOL_SEMICOLON);
   ParseTreeInteriorNodeSetMember *operator_like_symbol_assignment = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(PARSE_TREE_OPERATOR_LIKE_SYMBOL_ASSIGNMENT);

   // Put the master things into the sets.
   program->interior_node_set["+"] = operator_like_symbol_plus;
   program->interior_node_set["-"] = operator_like_symbol_minus;
   program->interior_node_set["*"] = operator_like_symbol_multiply;
   program->interior_node_set["/"] = operator_like_symbol_divide;
   program->interior_node_set[";"] = operator_like_symbol_semicolon;
   program->interior_node_set["="] = operator_like_symbol_assignment;

   program->symbol_table = new SymbolTable;

   for (std::vector<InputSourceProgramSymbolicVariableData>::iterator i = variables.begin(); i != variables.end(); i++)
   {
      bool constant = false;

      BigInteger value = 0;

      SYMBOL_TABLE_VARIABLE_ARCHETYPE archetype = SYMBOL_TABLE_VARIABLE_ARCHETYPE_BYSTANDER;

      switch (i->nature)
      {
         case (INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT) :
         {
            archetype = SYMBOL_TABLE_VARIABLE_ARCHETYPE_PARAMETER;
         }
         break;
         case (INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_OUTPUT) :
         {
            archetype = SYMBOL_TABLE_VARIABLE_ARCHETYPE_RESULT;
         }
         break;
         case (INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INPUT_OUTPUT) :
         {
            archetype = SYMBOL_TABLE_VARIABLE_ARCHETYPE_PARAMETER_RESULT;
         }
         break;
         case (INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_CONSTANT) :
         {
            archetype = SYMBOL_TABLE_VARIABLE_ARCHETYPE_CONSTANT;
         }
         break;
         case (INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_SYMBOLIC_CONSTANT) :
         {
            archetype = SYMBOL_TABLE_VARIABLE_ARCHETYPE_SYMBOLIC_CONSTANT;

            value = atoi(i->variable_name.c_str());
            constant = true;
         }
         break;
         case (INPUT_SOURCE_PROGRAM_SYMBOLIC_VARIABLE_NATURE_INTERMEDIATE) :
         {
            archetype = SYMBOL_TABLE_VARIABLE_ARCHETYPE_VARIABLE;
         }
         break;
         default :
         {
            assert(false);
         }
      }

      SymbolTable_InsertSymbolTableEntry(program->symbol_table,
                                         /* constant ? (String_ToString("_c") + i->variable_name)
                                                  : (i->variable_name), */
                                         (i->variable_name),
                                         archetype,
                                         value);

      SYMBOL_TABLE_KEY key = SymbolTable_ReverseLookupKeyFromVarname(program->symbol_table, i->variable_name.c_str());

      program->leaf_node_set[key] = ParseTreeLeafNodeSetMember_NewParseTreeLeafNodeSetMember_VariableFromSymbolTableKey(key);
   }

   SymbolTable_ReaffirmContainingTypes(program->symbol_table);

   SymbolTable_Print(program->symbol_table);

   ParseTreeCreationStatus status;

   status.current_symbol_considered = 0;
   status.input_source_code_prefix_form = input_source_code_prefix;
   status.program = program;

   program->target_program = ParseTreeCreation_CreateParseTreeNode(&status);

   return program;
}

// Create a ParseTreeNode from the next node along in the string
// held within the ParseTreeCreationStatus structure.
ParseTreeNode *ParseTreeCreation_CreateParseTreeNode(ParseTreeCreationStatus *status)
{
   std::string current_symbol = String_ReturnWord(status->input_source_code_prefix_form,
                                                  status->current_symbol_considered);

   printf("Considering node %d which is %s.\r\n", status->current_symbol_considered, current_symbol.c_str());

   ParseTreeNode *node;

   if ((current_symbol == ";") /* Language features */
    || (current_symbol == "=")
    || (current_symbol == "+")
    || (current_symbol == "-")
    || (current_symbol == "*")
    || (current_symbol == "/"))
   {
      node = ParseTreeNode_NewIncompleteInteriorNodeFromParseTreeInteriorNodeSetMember(status->program->interior_node_set[current_symbol]);

      status->current_symbol_considered++;

      node->as_interior_node.child_node[0] = ParseTreeCreation_CreateParseTreeNode(status);

      status->current_symbol_considered++;

      node->as_interior_node.child_node[1] = ParseTreeCreation_CreateParseTreeNode(status);
   }
   else
   {
      SYMBOL_TABLE_KEY current_symbol_table_key = SymbolTable_ReverseLookupKeyFromVarname(status->program->symbol_table, current_symbol.c_str());

      node = ParseTreeNode_NewLeafNodeFromParseTreeLeafNodeSetMember(status->program->leaf_node_set[current_symbol_table_key]);
   }

   return node;
}

void ParseTreeSourceProgram_CalculateOperatorLikeSymbolPopulations(
                                                     ParseTreeNode *program,
    std::map<const ParseTreeInteriorNodeSetMember *, unsigned int> &population)
{
   // Affect yourself.
   population[program->as_interior_node.interior_node_master]++;

   // Consider children.
   unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(program);

   for (unsigned int child = 0; child < arity; child++)
   {
      if (program->as_interior_node.child_node[child]->nature == ParseTreeNode::NODE_NATURE_INTERIOR_NODE)
      {
         ParseTreeSourceProgram_CalculateOperatorLikeSymbolPopulations(program->as_interior_node.child_node[child], population);
      }
   }
}

////////////////////
////////////////////

unsigned int INSTRUCTION_EXECUTIONS_MAX_COUNT;

// Defines a sort of array wrapper that's supposed to hold a series of
// unsigned integers that are all unique.
struct IndexSeries
{
   unsigned int length;

   unsigned int *value;
};

// Return a new index series of length _length.
IndexSeries *IndexSeries_NewIndexSeries(unsigned int _length)
{
   IndexSeries *index_series = new IndexSeries;

   index_series->length = _length;

   index_series->value = new unsigned int[_length];

   assert(index_series != NULL);
   assert(index_series->value != NULL);

   return index_series;
}

// Free an index series.
void IndexSeries_DestroyIndexSeries(IndexSeries *index_series)
{
   assert(index_series != NULL);
   assert(index_series->value != NULL);

   delete[] index_series->value;
   delete index_series;
}

// Given a population size and a tournament_size,
// return tournament_size unique indices in the
// range 0 <= x < population size
IndexSeries *IndexSeries_ReturnTournament(unsigned int population_size,
                                          unsigned int tournament_size)
{
   assert(tournament_size <= population_size);
   assert(tournament_size > 0);

   IndexSeries *new_tournament = IndexSeries_NewIndexSeries(tournament_size);

   // A tournament of everybody is everybody!
   if (population_size == tournament_size)
   {
      for (unsigned int index = 0; index < tournament_size; index++)
      {
         new_tournament->value[index] = index;
      }
   }
   else
   {
      new_tournament->value[0] = random_int_0_to_(population_size - 1);

      for (unsigned int index = 1; index < tournament_size; index++)
      {
         bool good;
         unsigned int new_index;

         do
         {
            good = true;

            new_index = random_int_0_to_(population_size - 1);

            for (unsigned int existing_index = 0; existing_index < index; existing_index++)
            {
               if (new_tournament->value[existing_index] == new_index)
               {
                  good = false;
               }
            }
         } while (!good);

         new_tournament->value[index] = new_index;
      }
   }

   for (unsigned int tournament_index = 0; tournament_index < tournament_size; tournament_index++)
   {
      assert(new_tournament->value[tournament_index] < population_size);
   }

   return new_tournament;
}

/////////////////////////////////////////

struct VirtualMachineInstructionExecutionRecord;

typedef std::vector<VirtualMachineInstructionExecutionRecord> VirtualMachineExecutionLog;

// This execution signature is shared between all low level implementation functions
#define VM_INSTRUCTION_EXECUTION_SIGNATURE \
   VirtualMachine *vm,              \
   Instruction    *i,               \
   VirtualMachineExecutionLog *log

// Virtual machine definition.

// Total number of registers in the virtual machine including PC and SP
const unsigned int VM_REGISTER_COUNT = 10;

// Program counter is the highest register.
const unsigned int VM_REGISTER_PROGRAM_COUNTER       = VM_REGISTER_COUNT - 1;
// Stack pointer counter is the second highest register.
const unsigned int VM_REGISTER_STACK_POINTER         = VM_REGISTER_PROGRAM_COUNTER - 1;

unsigned int vm_register_general_purpose_count;
unsigned int vm_register_final_general_purpose;

// Indexed memory (if used)
const unsigned int VM_INDEXED_MEMORY_TOTAL_SIZE         = 32;
const unsigned int VM_INDEXED_MEMORY_STACK_SECTION_SIZE =  8;
const int VM_INDEXED_MEMORY_STACK_SECTION_INDEX_HIGHEST = VM_INDEXED_MEMORY_TOTAL_SIZE - 1;
const int VM_INDEXED_MEMORY_STACK_SECTION_INDEX_LOWEST  = VM_INDEXED_MEMORY_STACK_SECTION_INDEX_HIGHEST
                                                                    - VM_INDEXED_MEMORY_STACK_SECTION_SIZE + 1;
const unsigned int VM_INDEXED_MEMORY_FREE_SECTION_SIZE  = VM_INDEXED_MEMORY_TOTAL_SIZE
                                                           - VM_INDEXED_MEMORY_STACK_SECTION_SIZE;

// Return a random register from all available registers.
static inline unsigned int VirtualMachine_RandomRegister()
{
   return random_int_0_to_(VM_REGISTER_COUNT - 1);
}

// Return a random general purpose register from all available registers.
static inline unsigned int VirtualMachine_RandomGeneralPurposeRegister()
{
   return random_int_0_to_(vm_register_final_general_purpose);
}

// An 'array' of strings containing the human readable mnemonic for each register.
std::vector<std::string> VM_REGISTER_MNEMONIC;

// The type of all values considered by the virtual machine.
typedef BigInteger VM_TYPE;

// A virtual machine (an execution instance) has registers, memory and state flags:
struct VirtualMachine
{
   VM_TYPE registers[VM_REGISTER_COUNT];

   // Associative memory is used.
   std::map<SYMBOL_TABLE_KEY, VM_TYPE> data;

   unsigned int state_running; // Set this to zero is the virtual machine should stop.
   unsigned int state_bad;     // Set this to 1 if there is any bad.

   unsigned int total_instructions;

   // You need a symbol table to run the virtual machine
   // in order to decode the symbols.
   SymbolTable *symbol_table;
};

// Populate the human readable mnemonics of the registers.
void VirtualMachine_InitialiseStaticMnemonics()
{
   for (unsigned int r = 0; r < VM_REGISTER_COUNT; r++)
   {
      if (r == VM_REGISTER_PROGRAM_COUNTER)
      {
         VM_REGISTER_MNEMONIC.push_back(std::string("PC"));
      }
      else if (r == VM_REGISTER_STACK_POINTER)
      {
         VM_REGISTER_MNEMONIC.push_back(std::string("SP"));
      }
      else
      {
         char s[10] = "";

         sprintf(s, "%d", r);

         VM_REGISTER_MNEMONIC.push_back(std::string(s));
      }
   }
}

// Fill the VirtualMachine with default values.
// Copy the starting values from the symbol table into the virtual machine.
void VirtualMachine_Initialise(VirtualMachine *vm, SymbolTable *symbol_table)
{
   assert(vm != NULL);
   assert(symbol_table != NULL);

   for (unsigned int r = 0; r <= vm_register_final_general_purpose; r++)
   {
      vm->registers[r] = 0;
   }

   vm->registers[VM_REGISTER_PROGRAM_COUNTER] = 0;
   vm->registers[VM_REGISTER_STACK_POINTER]   = VM_INDEXED_MEMORY_TOTAL_SIZE - 1;

   for (unsigned int c = 0; c < VM_INDEXED_MEMORY_TOTAL_SIZE; c++)
   {
      vm->data[c] = 0;
   }

   for (SymbolTableData::iterator i  = symbol_table->symbol_table_data.begin();
                                  i != symbol_table->symbol_table_data.end();
                                  i++)
   {
      if (i->second.symbolic_constant)
      {
         //printf("%s = %d\r\n", i->second.variable_symbolic_name.c_str(), i->second.value);

         vm->data[i->first] = i->second.value;
      }
   }

   vm->state_bad     = 0;
   vm->state_running = 1;

   vm->total_instructions = 0;

   vm->symbol_table = symbol_table;
}

// Copy the state of one VirtualMachine into another.
void VirtualMachine_InitialiseFromExisting(VirtualMachine *vm, VirtualMachine *vm_existing)
{
   memcpy(vm, vm_existing, sizeof(VirtualMachine));
}

// Return an int indicating if the given index can refer
// to some part of indexed memory.
// It's fine to write directly into stack territory...
int VirtualMachine_IsValidMemoryLocation(int memory_location)
{
   if (memory_location < 0) return 0;
   if (memory_location >= (int)VM_INDEXED_MEMORY_TOTAL_SIZE) return 0;

   return 1;
}

// The VM can halt due to one of the following error states.
enum VM_STATE_BAD_REASON
{
   VM_STATE_BAD_REASON_ZERO,
   VM_STATE_BAD_REASON_HALT_INSTRUCTION_ENCOUNTERED,
   VM_STATE_BAD_REASON_INSTRUCTION_OUT_OF_RANGE,
   VM_STATE_BAD_REASON_INSTRUCTION_UNUSUAL_INPUT,
   VM_STATE_BAD_REASON_STACK_OVERFLOW,
   VM_STATE_BAD_REASON_MEMORY_ACCESS_OUT_OF_BOUNDS,
   VM_STATE_BAD_REASON_PROGRAM_COUNTER_OUT_OF_RANGE,
   VM_STATE_BAD_REASON_DIVISION_BY_ZERO,
   VM_STATE_BAD_REASON_INSTRUCTION_LIMIT_REACHED,
   VM_STATE_BAD_REASON_UNIDENTIFIED_ERROR_STATE,
   VM_STATE_BAD_REASON__END_OF_LIST,
};

const unsigned int VM_STATE_BAD_REASON_COUNT = VM_STATE_BAD_REASON__END_OF_LIST;

const char *VM_STATE_BAD_REASONS_STRINGS[VM_STATE_BAD_REASON_COUNT] =
{
   "VM_STATE_BAD_REASON_ZERO",
   "VM_STATE_BAD_REASON_HALT_INSTRUCTION_ENCOUNTERED", // This isn't actually that bad.
   "VM_STATE_BAD_REASON_INSTRUCTION_OUT_OF_RANGE",
   "VM_STATE_BAD_REASON_INSTRUCTION_UNUSUAL_INPUT",
   "VM_STATE_BAD_REASON_STACK_OVERFLOW",
   "VM_STATE_BAD_REASON_MEMORY_ACCESS_OUT_OF_BOUNDS",
   "VM_STATE_BAD_REASON_PROGRAM_COUNTER_OUT_OF_RANGE",
   "VM_STATE_BAD_REASON_DIVISION_BY_ZERO",
   "VM_STATE_BAD_REASON_INSTRUCTION_LIMIT_REACHED",
   "VM_STATE_BAD_REASON_BAD_REASON_OUT_OF_RANGE_OH_DEAR",
};

// Abort execution with the following bad state.
void VirtualMachine_Abort(VirtualMachine *vm, VM_STATE_BAD_REASON reason)
{
   assert(vm != NULL);

   vm->state_bad = reason;
   vm->state_running = 0;
}

// Return the human readable bad state of the current virtual machine.
const char *VirtualMachine_ExplainCurrentBadState(VirtualMachine *vm)
{
   assert(vm != NULL);

   if (vm->state_bad < VM_STATE_BAD_REASON_COUNT)
   {
      return VM_STATE_BAD_REASONS_STRINGS[vm->state_bad];
   }

   return VM_STATE_BAD_REASONS_STRINGS[VM_STATE_BAD_REASON__END_OF_LIST];
}

// These are all of the possible operations available to the virtual machine
// InstructionString language.
enum VM_INSTRUCTION_OPERATION
{
   VMI_C_HALT,

   VMI_A_ADD,
   VMI_A_SUB,
   VMI_A_MUL,
   VMI_A_DIVP,

   VMI_S_PUSH,
   VMI_S_POP,
   VMI_S_TOP,

   VMI_M_LOADS,
   VMI_M_STORS,

   VMI_V_LOADV,

   VMI_J_JRELT,
   VMI_J_JRELE,
   VMI_J_JREGE,
   VMI_J_JREGT,
   VMI_J_JREEQ,
   VMI_J_JRENE,

   VMI__END_OF_LIST,
};

const unsigned int VM_INSTRUCTION_OPERATION_COUNT = VMI__END_OF_LIST;

// Define this to void dynamic memory allocation in handling InstructionStrings.
const unsigned int INSTRUCTION_OPERANDS_MAX_COUNT = 3;

// Polymorphic operand for instructions in the low level language.
struct InstructionOperand
{
          unsigned int as_register;
      SYMBOL_TABLE_KEY as_symbol_table_key;
               VM_TYPE as_direct_value;
};

// Instructions consist of an operation and some number of operands
// of the type defined above.
struct Instruction
{
   VM_INSTRUCTION_OPERATION operation;

   InstructionOperand operand[INSTRUCTION_OPERANDS_MAX_COUNT];
};

// An InstructionString holds a complete assembly program.
typedef std::list<Instruction> InstructionString;

// Return a pointer to the nth instruction in the InstructionString is.
Instruction *InstructionString_RetrieveInstruction(InstructionString *is, unsigned int instruction_cell_index)
{
   assert(is != NULL);
   assert(instruction_cell_index < is->size());

   InstructionString::iterator i = is->begin();

   advance(i, instruction_cell_index);

   return (&(*(i)));
}

// Return an int indicating if it is safe to access the nth instruction in
// InstructionString is.
int InstructionString_IsInstructionIndexWithinRange(InstructionString *is, int program_counter)
{
   assert(is != NULL);

   if (program_counter < 0) return 0;
   if (program_counter >= (int)is->size()) return 0;

   return 1;
}

unsigned int InstructionString_ReturnHighestRegisterUsed(InstructionString *is)
{
   assert(is != NULL);

   unsigned int highest_register = 0;

   for (unsigned int i = 0; i < is->size(); i++)
   {
      Instruction *ins = InstructionString_RetrieveInstruction(is, i);

      if (ins->operation != VMI_C_HALT)
      {
         highest_register = max(ins->operand[0].as_register, highest_register);
      }
   }

   return highest_register;
}

// Construct a new InstructionString by taking the first start1
// instructions of str1, then count2 instructions from str2 starting
// at start2, then the remaining instructions from str1 that appear
// count1 instructions after start1.
InstructionString *InstructionString_ConstructNewCompositeString(
   InstructionString *str1,
   InstructionString *str2,
   unsigned int start1, /* These are all in instructions. */
   unsigned int count1,
   unsigned int start2,
   unsigned int count2)
{
   assert(str1 != NULL);
   assert(str2 != NULL);

   assert(start1 <= str1->size());
   assert(start2 <= str2->size());

   assert(count1 <= str1->size());
   assert(count2 <= str2->size());

   InstructionString *composite = new InstructionString;

   InstructionString::iterator end_of_str1_copy = str1->begin();
   advance(end_of_str1_copy, start1);

   InstructionString::iterator start_of_str1_copy2 = str1->begin();
   advance(start_of_str1_copy2, start1+count1);

   InstructionString::iterator start_of_str2_copy = str2->begin();
   advance(start_of_str2_copy, start2);

   InstructionString::iterator end_of_str2_copy = str2->begin();
   advance(end_of_str2_copy, start2+count2);

   composite->insert(composite->begin(), str1->begin(), end_of_str1_copy);
   composite->insert(composite->begin(), start_of_str2_copy, end_of_str2_copy);
   composite->insert(composite->begin(), start_of_str1_copy2, str1->end());

   return composite;
}

// Duplicate the InstructionString is.
InstructionString *InstructionString_DuplicateInstructionString(InstructionString *is)
{
   assert(is != NULL);

   assert(is->size() != 0);

   InstructionString *new_is = new InstructionString;
   *new_is = *is;

   return new_is;
}

// Append the instructions within is2 to the end of is1.
void InstructionString_ConcatenateInstructionStringsInPlace(InstructionString *is1, InstructionString *is2)
{
   assert(is1 != NULL);
   assert(is2 != NULL);

   is1->insert(is1->end(), is2->begin(), is2->end());
}

// Return a new InstructionString containing the instructions from is1
// followed by the instructions from is2.
InstructionString *InstructionString_ConcatenateInstructionStrings(InstructionString *is1, InstructionString *is2)
{
   assert(is1 != NULL);
   assert(is2 != NULL);

   InstructionString *new_is = InstructionString_DuplicateInstructionString(is1);
   InstructionString_ConcatenateInstructionStringsInPlace(new_is, is2);

   return new_is;
}

void InstructionString_AppendInstruction(InstructionString *str, Instruction *i)
{
   assert(str != NULL);
   assert(i != NULL);

   str->push_back(*i);
}

void InstructionString_InsertInstruction(InstructionString *str, unsigned int insertion_position, Instruction *i)
{
   assert(str != NULL);
   assert(i != NULL);

   InstructionString::iterator position = str->begin();
   advance(position, insertion_position);

   str->insert(position, *i);
}


void Instruction_Print(Instruction *i, SymbolTable *symbol_table, bool with_semantics);

//////////////////

// Perform a basic tree walking on the program 'program' and
// produce an appropriate output instruction string -appended-
// to output_string.
void TreeWalkingCompiler_BasicCompileParseTreeSourceProgram(ParseTreeSourceProgram *program,
                                                                     ParseTreeNode *node,
                                                                      unsigned int  register_destination,
                                                                 InstructionString *output_string)
{
   switch (node->nature)
   {
      case (ParseTreeNode::NODE_NATURE_INTERIOR_NODE) :
      {
         const ParseTreeOperatorLikeSymbolHandle &operator_like_symbol = node->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol;

         switch (operator_like_symbol)
         {
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_SEMICOLON) :
            {
               TreeWalkingCompiler_BasicCompileParseTreeSourceProgram(program,
                                                                      node->as_interior_node.child_node[0],
                                                                      register_destination,
                                                                      output_string);

               TreeWalkingCompiler_BasicCompileParseTreeSourceProgram(program,
                                                                      node->as_interior_node.child_node[1],
                                                                      register_destination,
                                                                      output_string);
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_ASSIGNMENT) :
            {
               TreeWalkingCompiler_BasicCompileParseTreeSourceProgram(program,
                                                                      node->as_interior_node.child_node[1],
                                                                      register_destination,
                                                                      output_string);

               Instruction assignment;

               assignment.operation = VMI_M_STORS;

               assignment.operand[0].as_register = register_destination;
               assignment.operand[1].as_symbol_table_key = node->as_interior_node.child_node[0]->as_leaf_node.leaf_node_master->as_varname.varname_symbol_table_key;

               InstructionString_AppendInstruction(output_string, &assignment);
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_PLUS) :
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MINUS) :
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MULTIPLY) :
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_DIVIDE) :
            {
               TreeWalkingCompiler_BasicCompileParseTreeSourceProgram(program,
                                                                      node->as_interior_node.child_node[0],
                                                                      register_destination,
                                                                      output_string);

               TreeWalkingCompiler_BasicCompileParseTreeSourceProgram(program,
                                                                      node->as_interior_node.child_node[1],
                                                                      register_destination + 1,
                                                                      output_string);

               Instruction calculation;

               switch (operator_like_symbol)
               {
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_PLUS)     : calculation.operation = VMI_A_ADD;  break;
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MINUS)    : calculation.operation = VMI_A_SUB;  break;
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MULTIPLY) : calculation.operation = VMI_A_MUL;  break;
                  case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_DIVIDE)   : calculation.operation = VMI_A_DIVP; break;
                  default :
                  {
                     assert(false);
                  }
               }

               calculation.operand[0].as_register = register_destination;
               calculation.operand[1].as_register = register_destination;
               calculation.operand[2].as_register = register_destination + 1;

               InstructionString_AppendInstruction(output_string, &calculation);
            }
            break;
            default :
            {
               assert(false);
            }
         }
      }
      break;
      case (ParseTreeNode::NODE_NATURE_LEAF_NODE) :
      {
         Instruction loading;

         loading.operation = VMI_M_LOADS;

         loading.operand[0].as_register = register_destination;
         loading.operand[1].as_symbol_table_key = node->as_leaf_node.leaf_node_master->as_varname.varname_symbol_table_key;

         InstructionString_AppendInstruction(output_string, &loading);
      }
      break;
      default :
      {
         assert(false);
      }
   }
}

//////////////////

// The properties of the execution of an operation in
// the low level language
struct VMInstructionOperationMetadata
{
   // How is it written.
   const char *instruction_mnemonic;

   // How is it executed.
   void (*implementation)(VM_INSTRUCTION_EXECUTION_SIGNATURE);

   // How many operands does it take.
   unsigned int operands;
};

extern VMInstructionOperationMetadata vm_instruction_operation_metadata[];

// The result of an execution of an operation in the low level language
// can be placed in a register or a symbol table location.
struct VirtualMachineInstructionExecutionRecordLocation
{
          unsigned int as_register;
      SYMBOL_TABLE_KEY as_symbol_table_key;
};

// This structure records the action taken during the execution
// of a single instruction in the low level language.
struct VirtualMachineInstructionExecutionRecord
{
   VM_INSTRUCTION_OPERATION instruction;

                                            VM_TYPE result_value;
   VirtualMachineInstructionExecutionRecordLocation result_location;

                                            VM_TYPE operand_1_value;
   VirtualMachineInstructionExecutionRecordLocation operand_1_location;
                                            VM_TYPE operand_2_value;
   VirtualMachineInstructionExecutionRecordLocation operand_2_location;

   VM_TYPE registers_before[VM_REGISTER_COUNT];
   VM_TYPE registers_after[VM_REGISTER_COUNT];
};

// Print a VirtualMachineInstructionExecutionRecord in human readable form.
void VirtualMachineInstructionExecutionRecord_Print(VirtualMachineInstructionExecutionRecord *r, SymbolTable *symbol_table)
{
//   printf("%-5s [%u] %8d - ([%u] %8d, [%u] %8d)\r\n",
//          vm_instruction_operation_metadata[r->instruction].instruction_mnemonic,
//          r->result_register,    r->result_value,
//          r->operand_1_register, r->operand_1_value,
//          r->operand_2_register, r->operand_2_value);


   switch (r->instruction)
   {
      case (VMI_C_HALT) :
      {
         printf("Halt");
      }
      break;

      case (VMI_A_ADD)  :
      case (VMI_A_SUB)  :
      case (VMI_A_MUL)  :
      case (VMI_A_DIVP) :
      {
         unsigned int &register_a = r->result_location.as_register;
         unsigned int &register_b = r->operand_1_location.as_register;
         unsigned int &register_c = r->operand_2_location.as_register;

         const char *register_result    = VM_REGISTER_MNEMONIC[register_a].c_str();
         const char *register_operand_1 = VM_REGISTER_MNEMONIC[register_b].c_str();
         const char *register_operand_2 = VM_REGISTER_MNEMONIC[register_c].c_str();

         const char *op = "Unknown";

         switch (r->instruction)
         {
            case (VMI_A_ADD)  : op = "Add";      break;
            case (VMI_A_SUB)  : op = "Subtract"; break;
            case (VMI_A_MUL)  : op = "Multiply"; break;
            case (VMI_A_DIVP) : op = "Divide";   break;

            default :
            {
               assert(false); // HOW?!
            }
         }

         printf("%s %s in r%s, %s in r%s to get %s in r%s", op,
                                                            bigIntegerToString(r->operand_1_value).c_str(), register_operand_1,
                                                            bigIntegerToString(r->operand_2_value).c_str(), register_operand_2,
                                                            bigIntegerToString(r->result_value).c_str(),    register_result);
      }
      break;

      case (VMI_S_PUSH) :
      {
         unsigned int &register_a = r->result_location.as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         printf("Push %s in r%s", bigIntegerToString(r->result_value).c_str(), register_result);
      }
      break;

      case (VMI_S_POP) :
      {
         unsigned int &register_a = r->result_location.as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         printf("Pop to get %s in r%s", bigIntegerToString(r->result_value).c_str(), register_result);
      }
      break;

      case (VMI_S_TOP) :
      {
         unsigned int &register_a = r->result_location.as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         printf("Top to get %s in r%s", bigIntegerToString(r->result_value).c_str(), register_result);
      }
      break;

      case (VMI_M_LOADS) :
      {
         unsigned int &register_a = r->result_location.as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         const char *variable_symbolic_name_string = SymbolTable_QuerySymbolTableKey(symbol_table, r->operand_1_location.as_symbol_table_key)->variable_symbolic_name.c_str();

         printf("Loads %s in r%s from %s", bigIntegerToString(r->result_value).c_str(), register_result, variable_symbolic_name_string);
      }
      break;

      case (VMI_M_STORS) :
      {
         unsigned int &register_a = r->result_location.as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         const char *variable_symbolic_name_string = SymbolTable_QuerySymbolTableKey(symbol_table, r->operand_1_location.as_symbol_table_key)->variable_symbolic_name.c_str();

         printf("Stors %s in r%s to %s", bigIntegerToString(r->result_value).c_str(), register_result, variable_symbolic_name_string);
      }
      break;
//
//      case (VMI_J_JRELT) :
//      case (VMI_J_JRELE) :
//      case (VMI_J_JREGE) :
//      case (VMI_J_JREGT) :
//      case (VMI_J_JREEQ) :
//      case (VMI_J_JRENE) :
//      {
//         unsigned int &register_a = i->operand[0].as_register;
//         unsigned int &register_b = i->operand[1].as_register;
//         unsigned int &register_c = i->operand[2].as_register;
//
//         const char *register_result    = VM_REGISTER_MNEMONIC[register_a].c_str();
//         const char *register_operand_1 = VM_REGISTER_MNEMONIC[register_b].c_str();
//         const char *register_operand_2 = VM_REGISTER_MNEMONIC[register_c].c_str();
//
//         const char *register_program_counter = VM_REGISTER_MNEMONIC[VM_REGISTER_PROGRAM_COUNTER].c_str();
//
//         const char *predicate = "?";
//
//         switch (i->operation)
//         {
//            case (VMI_J_JRELT) : predicate = "<" ; break;
//            case (VMI_J_JRELE) : predicate = "<="; break;
//            case (VMI_J_JREGE) : predicate = ">="; break;
//            case (VMI_J_JREGT) : predicate = ">" ; break;
//            case (VMI_J_JREEQ) : predicate = "=="; break;
//            case (VMI_J_JRENE) : predicate = "!="; break;
//
//            default :
//            {
//               assert(false);
//            }
//         }
//
//         printf("if (r[%s] %s r[%s]) r[%s] += r[%s]", register_result, predicate, register_operand_1, register_program_counter, register_operand_2);
//      }
//      break;

      default :
      {
         assert(false);
      }
   }

   printf("\r\n");
}

//unsigned int register_argument_modulo(int i)
//{
//   int a = ((i > 0) ? (i) : (-i));
//
//   return a % VM_REGISTER_COUNT;
//}

// Is a register in range?
#define assert_register_good(_r_) assert(_r_ < VM_REGISTER_COUNT)
// Is a symbol good?
#define assert_symbol_good(_s_)   assert(SymbolTable_QuerySymbolTableKey(vm->symbol_table, _s_))

// The system halts.
void vm_instruction_implementation_vmi_c_halt(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_HALT_INSTRUCTION_ENCOUNTERED);
   return;
}

void vm_instruction_implementation_vmi_a_add(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   unsigned int &register_a = i->operand[0].as_register;
   unsigned int &register_b = i->operand[1].as_register;
   unsigned int &register_c = i->operand[2].as_register;

   assert_register_good(register_a);
   assert_register_good(register_b);
   assert_register_good(register_c);

   VM_TYPE value_l = vm->registers[register_b];
   VM_TYPE value_r = vm->registers[register_c];

   VM_TYPE result  = value_l + value_r;

   vm->registers[register_a] = result;

   VirtualMachineInstructionExecutionRecord record;

   record.instruction = VMI_A_ADD;
   record.result_value = result;
   record.result_location.as_register = register_a;
   record.operand_1_value = value_l;
   record.operand_1_location.as_register = register_b;
   record.operand_2_value = value_r;
   record.operand_2_location.as_register = register_c;

   log->push_back(record);
}

void vm_instruction_implementation_vmi_a_sub(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   unsigned int &register_a = i->operand[0].as_register;
   unsigned int &register_b = i->operand[1].as_register;
   unsigned int &register_c = i->operand[2].as_register;

   assert_register_good(register_a);
   assert_register_good(register_b);
   assert_register_good(register_c);

   VM_TYPE value_l = vm->registers[register_b];
   VM_TYPE value_r = vm->registers[register_c];

   VM_TYPE result  = value_l - value_r;

   vm->registers[register_a] = result;

   VirtualMachineInstructionExecutionRecord record;

   record.instruction = VMI_A_SUB;
   record.result_value = result;
   record.result_location.as_register = register_a;
   record.operand_1_value = value_l;
   record.operand_1_location.as_register = register_b;
   record.operand_2_value = value_r;
   record.operand_2_location.as_register = register_c;

   log->push_back(record);
}

void vm_instruction_implementation_vmi_a_mul(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   unsigned int &register_a = i->operand[0].as_register;
   unsigned int &register_b = i->operand[1].as_register;
   unsigned int &register_c = i->operand[2].as_register;

   assert_register_good(register_a);
   assert_register_good(register_b);
   assert_register_good(register_c);

   VM_TYPE value_l = vm->registers[register_b];
   VM_TYPE value_r = vm->registers[register_c];

   VM_TYPE result  = value_l * value_r;

   vm->registers[register_a] = result;

   VirtualMachineInstructionExecutionRecord record;

   record.instruction = VMI_A_MUL;
   record.result_value = result;
   record.result_location.as_register = register_a;
   record.operand_1_value = value_l;
   record.operand_1_location.as_register = register_b;
   record.operand_2_value = value_r;
   record.operand_2_location.as_register = register_c;

   log->push_back(record);
}

void vm_instruction_implementation_vmi_a_divp(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   unsigned int &register_a = i->operand[0].as_register;
   unsigned int &register_b = i->operand[1].as_register;
   unsigned int &register_c = i->operand[2].as_register;

   assert_register_good(register_a);
   assert_register_good(register_b);
   assert_register_good(register_c);

   VM_TYPE value_l = vm->registers[register_b];
   VM_TYPE value_r = vm->registers[register_c];

//#define DIVISION_BY_ZERO_IS_ERROR

#ifdef DIVISION_BY_ZERO_IS_ERROR
   if (value_r == 0)
   {
      VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_DIVISION_BY_ZERO);
      return;
   }

   VM_TYPE result  = value_l / value_r;
#else
   VM_TYPE result  = value_l / (value_r != 0 ? value_r : 1);
#endif

   vm->registers[register_a] = result;

   VirtualMachineInstructionExecutionRecord record;

   record.instruction = VMI_A_DIVP;
   record.result_value = result;
   record.result_location.as_register = register_a;
   record.operand_1_value = value_l;
   record.operand_1_location.as_register = register_b;
   record.operand_2_value = value_r;
   record.operand_2_location.as_register = register_c;

   log->push_back(record);
}

void vm_instruction_implementation_vmi_s_push(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   assert(false);

//   unsigned int &register_a = i->operand[0].as_register;
//
//   assert_register_good(register_a);
//
//   if (vm->registers[VM_REGISTER_STACK_POINTER] < VM_INDEXED_MEMORY_STACK_SECTION_INDEX_LOWEST)
//   {
//      VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_STACK_OVERFLOW);
//   }
//
//   vm->data[vm->registers[VM_REGISTER_STACK_POINTER]] = vm->registers[register_a];
//
//   vm->registers[VM_REGISTER_STACK_POINTER] -= 1;
}

void vm_instruction_implementation_vmi_s_pop(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   assert(false);

//   unsigned int &register_a = i->operand[0].as_register;
//
//   assert_register_good(register_a);
//
//   if (vm->registers[VM_REGISTER_STACK_POINTER] >= VM_INDEXED_MEMORY_STACK_SECTION_INDEX_HIGHEST)
//   {
//      VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_STACK_OVERFLOW);
//   }
//
//   VM_TYPE result = vm->data[++vm->registers[VM_REGISTER_STACK_POINTER]];
//
//   vm->registers[register_a] = result;
//
//   VirtualMachineInstructionExecutionRecord record;
//
//   record.instruction = VMI_S_POP;
//   record.result_value = result;
//   record.result_location.as_register = register_a;
//   record.operand_1_value = result;
//   record.operand_1_location.as_register = register_a;
//   record.operand_2_value = result;
//   record.operand_2_location.as_register = register_a;
//
//   log->push_back(record);
}

void vm_instruction_implementation_vmi_s_top(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   assert(false);

//   unsigned int &register_a = i->operand[0].as_register;
//
//   assert_register_good(register_a);
//
//   if (vm->registers[VM_REGISTER_STACK_POINTER] >= VM_INDEXED_MEMORY_STACK_SECTION_INDEX_HIGHEST)
//   {
//      VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_STACK_OVERFLOW);
//   }
//
//   VM_TYPE result = vm->data[vm->registers[VM_REGISTER_STACK_POINTER] + 1];
//
//   vm->registers[register_a] = result;
//
//   VirtualMachineInstructionExecutionRecord record;
//
//   record.instruction = VMI_S_TOP;
//   record.result_value = result;
//   record.result_location.as_register = register_a;
//   record.operand_1_value = result;
//   record.operand_1_location.as_register = register_a;
//   record.operand_2_value = result;
//   record.operand_2_location.as_register = register_a;
//
//   log->push_back(record);
}

void vm_instruction_implementation_vmi_v_loadv(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   unsigned int &register_a = i->operand[0].as_register;

   assert_register_good(register_a);

   VM_TYPE result = i->operand[1].as_direct_value;

   vm->registers[register_a] = result;

   VirtualMachineInstructionExecutionRecord record;

   record.instruction = VMI_V_LOADV;
   record.result_value = result;
   record.result_location.as_register = register_a;
   record.operand_1_value = result;
   record.operand_1_location.as_register = register_a;
   record.operand_2_value = result;
   record.operand_2_location.as_register = register_a;

   log->push_back(record);
}

void vm_instruction_implementation_vmi_m_loads(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   unsigned int &register_a = i->operand[0].as_register;

   SYMBOL_TABLE_KEY &symbol = i->operand[1].as_symbol_table_key;

   assert_register_good(register_a);
   assert_symbol_good(symbol);

   VM_TYPE result = vm->data[symbol];

   vm->registers[register_a] = result;

   VirtualMachineInstructionExecutionRecord record;

   record.instruction = VMI_M_LOADS;
   record.result_value = result;
   record.result_location.as_register = register_a;
   record.operand_1_value = result;
   record.operand_1_location.as_symbol_table_key = symbol;
   record.operand_2_value = result;
   record.operand_2_location.as_register = register_a;

   log->push_back(record);
}

void vm_instruction_implementation_vmi_m_stors(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   unsigned int &register_a = i->operand[0].as_register;

   SYMBOL_TABLE_KEY &symbol = i->operand[1].as_symbol_table_key;

   assert_register_good(register_a);
   assert_symbol_good(symbol);

   VM_TYPE result = vm->registers[register_a];

   vm->data[symbol] = result;

   VirtualMachineInstructionExecutionRecord record;

   record.instruction = VMI_M_STORS;
   record.result_value = result;
   record.result_location.as_register = register_a;
   record.operand_1_value = result;
   record.operand_1_location.as_symbol_table_key = symbol;
   record.operand_2_value = result;
   record.operand_2_location.as_register = register_a;

   log->push_back(record);
}

enum JUMP_PREDICATE
{
   JUMP_PREDICATE_INVALID,
   JUMP_PREDICATE_LESS_THAN,
   JUMP_PREDICATE_LESS_THAN_OR_EQUAL,
   JUMP_PREDICATE_GREATER_THAN_OR_EQUAL,
   JUMP_PREDICATE_GREATER_THAN,
   JUMP_PREDICATE_EQUAL,
   JUMP_PREDICATE_NOT_EQUAL,
};

void vm_instruction_implementation_vmi_j_generic_relative_jump(VM_INSTRUCTION_EXECUTION_SIGNATURE, JUMP_PREDICATE predicate)
{
   unsigned int &register_a = i->operand[0].as_register;
   unsigned int &register_b = i->operand[1].as_register;
   unsigned int &register_c = i->operand[2].as_register;

   assert_register_good(register_a);
   assert_register_good(register_b);
   assert_register_good(register_c);

   VM_TYPE &a = vm->registers[register_a];
   VM_TYPE &b = vm->registers[register_b];

   VM_TYPE &distance = vm->registers[register_c];

   bool condition_met = false;

   switch (predicate)
   {
      case (JUMP_PREDICATE_LESS_THAN)             : condition_met = (a <  b); break;
      case (JUMP_PREDICATE_LESS_THAN_OR_EQUAL)    : condition_met = (a <= b); break;
      case (JUMP_PREDICATE_GREATER_THAN_OR_EQUAL) : condition_met = (a >= b); break;
      case (JUMP_PREDICATE_GREATER_THAN)          : condition_met = (a >  b); break;
      case (JUMP_PREDICATE_EQUAL)                 : condition_met = (a == b); break;
      case (JUMP_PREDICATE_NOT_EQUAL)             : condition_met = (a != b); break;
      default : assert(false); break;
   }

   if (condition_met)
   {
      vm->registers[VM_REGISTER_PROGRAM_COUNTER] += distance;
   }

//   log->push_back(VirtualMachineInstructionExecutionRecord((VM_INSTRUCTION_OPERATION)i->operation,
//                                                           result,  register_a,
//                                                           value_l, 0,
//                                                           value_r, register_c));
}

void vm_instruction_implementation_vmi_j_jrelt(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   vm_instruction_implementation_vmi_j_generic_relative_jump(vm, i, log, JUMP_PREDICATE_LESS_THAN);
}

void vm_instruction_implementation_vmi_j_jrele(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   vm_instruction_implementation_vmi_j_generic_relative_jump(vm, i, log, JUMP_PREDICATE_LESS_THAN_OR_EQUAL);
}

void vm_instruction_implementation_vmi_j_jrege(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   vm_instruction_implementation_vmi_j_generic_relative_jump(vm, i, log, JUMP_PREDICATE_GREATER_THAN_OR_EQUAL);
}

void vm_instruction_implementation_vmi_j_jregt(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   vm_instruction_implementation_vmi_j_generic_relative_jump(vm, i, log, JUMP_PREDICATE_GREATER_THAN);
}

void vm_instruction_implementation_vmi_j_jreeq(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   vm_instruction_implementation_vmi_j_generic_relative_jump(vm, i, log, JUMP_PREDICATE_EQUAL);
}

void vm_instruction_implementation_vmi_j_jrene(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   vm_instruction_implementation_vmi_j_generic_relative_jump(vm, i, log, JUMP_PREDICATE_NOT_EQUAL);
}

VMInstructionOperationMetadata vm_instruction_operation_metadata[VM_INSTRUCTION_OPERATION_COUNT] =
{
   {"HALT"  , vm_instruction_implementation_vmi_c_halt,   0, },

   {"ADD"   , vm_instruction_implementation_vmi_a_add,    3, },
   {"SUB"   , vm_instruction_implementation_vmi_a_sub,    3, },
   {"MUL"   , vm_instruction_implementation_vmi_a_mul,    3, },
   {"DIVP"  , vm_instruction_implementation_vmi_a_divp,   3, },

   {"PUSH"  , vm_instruction_implementation_vmi_s_push,   1, },
   {"POP"   , vm_instruction_implementation_vmi_s_pop,    1, },
   {"TOP"   , vm_instruction_implementation_vmi_s_top,    1, },

   {"LOADS" , vm_instruction_implementation_vmi_m_loads,  2, },
   {"STORS" , vm_instruction_implementation_vmi_m_stors,  2, },

   {"LOADV" , vm_instruction_implementation_vmi_v_loadv,  2, },

   {"JRELT" , vm_instruction_implementation_vmi_j_jrelt,  3, },
   {"JRELE" , vm_instruction_implementation_vmi_j_jrele,  3, },
   {"JREGE" , vm_instruction_implementation_vmi_j_jrege,  3, },
   {"JREGT" , vm_instruction_implementation_vmi_j_jregt,  3, },
   {"JREEQ" , vm_instruction_implementation_vmi_j_jreeq,  3, },
   {"JRENE" , vm_instruction_implementation_vmi_j_jrene,  3, },
};

// Return an int indicating if the instruction is executable in
// a virtual machine (is the operation part valid?)
int VirtualMachineInstruction_IsValidInstruction(Instruction *i)
{
   assert(i != NULL);

   if (i->operation >= VMI__END_OF_LIST) return 0;

   return 1;
}

// Execute the given instruction in the given VirtualMachine.
void VirtualMachineInstruction_ExecuteInstruction(VM_INSTRUCTION_EXECUTION_SIGNATURE)
{
   if (VirtualMachineInstruction_IsValidInstruction(i))
   {
      VM_TYPE registers_before[VM_REGISTER_COUNT];

      for (unsigned int r = 0; r < VM_REGISTER_COUNT; r++)
      {
         registers_before[r] = vm->registers[r];
      }

      int log_size_before = log->size();

      (*(vm_instruction_operation_metadata[i->operation].implementation))(vm, i, log);

      int log_size_after  = log->size();

      if (log_size_before != log_size_after)
      {
         // Add the register values to the most recent log entry.
         VirtualMachineInstructionExecutionRecord *record = &(*(log->end() - 1));

         for (unsigned int r = 0; r < VM_REGISTER_COUNT; r++)
         {
            record->registers_before[r] = registers_before[r];
            record->registers_after[r] = vm->registers[r];
         }
      }

      vm->total_instructions++;
   }
   else
   {
      // Illegal instruction
      VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_INSTRUCTION_OUT_OF_RANGE);
      return;
   }
}

// Print the contents of all registers in the given VirtualMachine state.
void VirtualMachine_PrintRegisters(VirtualMachine *vm)
{
   assert(vm != NULL);

   for (unsigned int r = 0; r < VM_REGISTER_COUNT; r++)
   {
      if (r > 0)
      {
         printf(", ");
      }

      printf("R%d", r);

      if (r == VM_REGISTER_PROGRAM_COUNTER)
      {
         printf(" (PC)");
      }

      printf(" = %s", bigIntegerToString(vm->registers[r]).c_str());
   }
   printf("\r\n");
}

// Print the contents of the indexed memory in the given VirtualMachine state.
//void VirtualMachine_PrintMemory(VirtualMachine *vm)
//{
//   assert(vm != NULL);
//
//   for (unsigned int r = 0; r < VM_INDEXED_MEMORY_FREE_SECTION_SIZE; r++)
//   {
//      printf("Memory[%3d] = %d\r\n", r, vm->data[r]);
//   }
//}

// Advance the program counter in the given VirtualMachine and
// execute the next instruction.
void VirtualMachineInstructionString_Advance(            VirtualMachine *vm,
                                                      InstructionString *is,
                                             VirtualMachineExecutionLog *log)
{
   assert(vm != NULL);
   assert(is != NULL);

   unsigned int program_counter = bigIntegerToInt(vm->registers[VM_REGISTER_PROGRAM_COUNTER]);

   vm->registers[VM_REGISTER_PROGRAM_COUNTER] += 1;

   if (!InstructionString_IsInstructionIndexWithinRange(is, program_counter))
   {
      VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_PROGRAM_COUNTER_OUT_OF_RANGE);
      return;
   }

   Instruction *i = InstructionString_RetrieveInstruction(is, program_counter);

   VirtualMachineInstruction_ExecuteInstruction(vm, i, log);
}

void Instruction_Print(Instruction *i, SymbolTable *symbol_table, bool with_semantics);

// The VirtualMachine can be executed in either interactive or non-interactive
// mode. The interactive mode prints out the current state of the VirtualMachine
// at the end of every instruction and waits for user input.
// The non interactive mode does not print and does not wait.
enum VM_EXECUTE_INTERACTIVE_FLAG
{
   VM_EXECUTE_NONINTERACTIVE,
   VM_EXECUTE_INTERACTIVE,
};

// Execute the given InstructionString in the given VirtualMachine
void VirtualMachineInstructionString_ExecuteInstructionString(            VirtualMachine *vm,
                                                                       InstructionString *is,
                                                              VirtualMachineExecutionLog *log,
                                                                            unsigned int  instruction_limit,
                                                             VM_EXECUTE_INTERACTIVE_FLAG  interactive)
{
   assert(vm != NULL);
   assert(is != NULL);
   assert(log != NULL);
   assert(instruction_limit != 0);

   while (vm->state_running)
   {
      if (vm->total_instructions == instruction_limit)
      {
         VirtualMachine_Abort(vm, VM_STATE_BAD_REASON_INSTRUCTION_LIMIT_REACHED);
      }
      else
      {
         if (interactive)
         {
            if (vm->state_bad == 0)
            {
               VirtualMachine_PrintRegisters(vm);
               //VirtualMachine_PrintMemory(vm);

               int program_counter_as_int = bigIntegerToInt(vm->registers[VM_REGISTER_PROGRAM_COUNTER]);

               if (InstructionString_IsInstructionIndexWithinRange(is, program_counter_as_int))
               {
                  Instruction_Print(InstructionString_RetrieveInstruction(is, program_counter_as_int), vm->symbol_table, true);
               }
               else
               {
                  printf("Program counter is out of program range.\r\n");
               }
            }
            else
            {
               printf("Encountered VirtualMachine BAD STATE: '%s'\r\n", VirtualMachine_ExplainCurrentBadState(vm));
            }

            wait_for_return();
         }

         VirtualMachineInstructionString_Advance(vm, is, log);
      }
   }

   // We might use the bad state as an index into an array later,
   // so let's be safe!
   if (vm->state_bad > VM_STATE_BAD_REASON_UNIDENTIFIED_ERROR_STATE)
   {
      vm->state_bad = VM_STATE_BAD_REASON_UNIDENTIFIED_ERROR_STATE;
   }
}

// Print a human readable version of the semantics of an instruction.
void VirtualMachineInstruction_PrintSemantics(Instruction *i, SymbolTable *symbol_table)
{
   assert(i != NULL);
   assert(symbol_table != NULL);

   switch (i->operation)
   {
      case (VMI_C_HALT) :
      {
         printf("Halt");
      }
      break;

      case (VMI_A_ADD)  :
      case (VMI_A_SUB)  :
      case (VMI_A_MUL)  :
      case (VMI_A_DIVP) :
      {
         unsigned int &register_a = i->operand[0].as_register;
         unsigned int &register_b = i->operand[1].as_register;
         unsigned int &register_c = i->operand[2].as_register;

         const char *register_result    = VM_REGISTER_MNEMONIC[register_a].c_str();
         const char *register_operand_1 = VM_REGISTER_MNEMONIC[register_b].c_str();
         const char *register_operand_2 = VM_REGISTER_MNEMONIC[register_c].c_str();

         if (i->operation == VMI_A_DIVP)
         {
            printf("r[%s] = divp(r[%s], r[%s])", register_result, register_operand_1, register_operand_2);
            break;
         }

         char op = '?';

         switch (i->operation)
         {
            case (VMI_A_ADD)  : op = '+'; break;
            case (VMI_A_SUB)  : op = '-'; break;
            case (VMI_A_MUL)  : op = '*'; break;

            default :
            {
               assert(false); // HOW?!
            }
         }

         printf("r[%s] = r[%s] %c r[%s]", register_result, register_operand_1, op, register_operand_2);
      }
      break;

      case (VMI_S_PUSH) :
      {
         unsigned int &register_a = i->operand[0].as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         printf("PUSH(r[%s])", register_result);
      }
      break;

      case (VMI_S_POP) :
      {
         unsigned int &register_a = i->operand[0].as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         printf("r[%s] = POP()", register_result);
      }
      break;

      case (VMI_S_TOP) :
      {
         unsigned int &register_a = i->operand[0].as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         printf("r[%s] = TOP()", register_result);
      }
      break;

      case (VMI_M_LOADS) :
      {
         unsigned int &register_a = i->operand[0].as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         const char *variable_symbolic_name_string = SymbolTable_QuerySymbolTableKey(symbol_table, i->operand[1].as_symbol_table_key)->variable_symbolic_name.c_str();

         printf("r[%s] = %s", register_result, variable_symbolic_name_string);
      }
      break;

      case (VMI_M_STORS) :
      {
         unsigned int &register_a = i->operand[0].as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         const char *variable_symbolic_name_string = SymbolTable_QuerySymbolTableKey(symbol_table, i->operand[1].as_symbol_table_key)->variable_symbolic_name.c_str();

         printf("%s = r[%s]", variable_symbolic_name_string, register_result);
      }
      break;

      case (VMI_V_LOADV) :
      {
         unsigned int &register_a = i->operand[0].as_register;

         const char *register_result = VM_REGISTER_MNEMONIC[register_a].c_str();

         BigInteger value = i->operand[1].as_direct_value;

         printf("r[%s] = %s", register_result, bigIntegerToString(value).c_str());
      }
      break;

      case (VMI_J_JRELT) :
      case (VMI_J_JRELE) :
      case (VMI_J_JREGE) :
      case (VMI_J_JREGT) :
      case (VMI_J_JREEQ) :
      case (VMI_J_JRENE) :
      {
         unsigned int &register_a = i->operand[0].as_register;
         unsigned int &register_b = i->operand[1].as_register;
         unsigned int &register_c = i->operand[2].as_register;

         const char *register_result    = VM_REGISTER_MNEMONIC[register_a].c_str();
         const char *register_operand_1 = VM_REGISTER_MNEMONIC[register_b].c_str();
         const char *register_operand_2 = VM_REGISTER_MNEMONIC[register_c].c_str();

         const char *register_program_counter = VM_REGISTER_MNEMONIC[VM_REGISTER_PROGRAM_COUNTER].c_str();

         const char *predicate = "?";

         switch (i->operation)
         {
            case (VMI_J_JRELT) : predicate = "<" ; break;
            case (VMI_J_JRELE) : predicate = "<="; break;
            case (VMI_J_JREGE) : predicate = ">="; break;
            case (VMI_J_JREGT) : predicate = ">" ; break;
            case (VMI_J_JREEQ) : predicate = "=="; break;
            case (VMI_J_JRENE) : predicate = "!="; break;

            default :
            {
               assert(false);
            }
         }

         printf("if (r[%s] %s r[%s]) r[%s] += r[%s]", register_result, predicate, register_operand_1, register_program_counter, register_operand_2);
      }
      break;

      default :
      {
         assert(false);
      }
   }

   printf(";");
}

// Print an assembly like representation of the given instruction.
void Instruction_Print(Instruction *i, SymbolTable *symbol_table, bool with_semantics)
{
   assert(i != NULL);

   printf("%5s ", vm_instruction_operation_metadata[i->operation].instruction_mnemonic);

   unsigned int &r_a = i->operand[0].as_register;
   unsigned int &r_b = i->operand[1].as_register;
   unsigned int &r_c = i->operand[2].as_register;

   switch(i->operation)
   {
      case (VMI_C_HALT)  : printf("----------");
      break;

      case (VMI_A_ADD)   :
      case (VMI_A_SUB)   :
      case (VMI_A_MUL)   :
      case (VMI_A_DIVP)  :
      case (VMI_J_JRELT) :
      case (VMI_J_JRELE) :
      case (VMI_J_JREGE) :
      case (VMI_J_JREGT) :
      case (VMI_J_JREEQ) :
      case (VMI_J_JRENE) : printf("%2d,  %2d,  %2d  ", r_a, r_b, r_c);
      break;

      case (VMI_S_PUSH)  :
      case (VMI_S_POP)   :
      case (VMI_S_TOP)   : printf("%2d            ", r_a);
      break;

      case (VMI_V_LOADV) : printf("%2d, %5s     ", r_a, bigIntegerToString(i->operand[1].as_direct_value).c_str());
      break;

      case (VMI_M_LOADS) :
      case (VMI_M_STORS) : printf("%2d,  %-8s ", r_a, SymbolTable_QuerySymbolTableKey(symbol_table, i->operand[1].as_symbol_table_key)->variable_symbolic_name.c_str());
      break;

      default : assert(false);
   }

   if (with_semantics)
   {
      printf("    => ");
      VirtualMachineInstruction_PrintSemantics(i, symbol_table);
   }

   printf("\r\n");
}

// Print an assembly like representation of an InstructionString
void InstructionString_Print(InstructionString *is, SymbolTable *symbol_table, bool with_semantics = false)
{
   assert(is != NULL);

   for (unsigned int ins = 0; ins < is->size(); ins++)
   {
      printf("Instruction %4d: ", ins);

      Instruction_Print(InstructionString_RetrieveInstruction(is, ins), symbol_table, with_semantics);
   }
}

// An InstructionProbability links a low level VM_INSTRUCTION_OPERATION
// to a probability.
struct InstructionProbability
{
   VM_INSTRUCTION_OPERATION operation;

                      float probability;

                      float cumulative_probability;

   InstructionProbability(VM_INSTRUCTION_OPERATION _operation,
                          float _probability)
      : operation(_operation), probability(_probability)
   {
   }
};

// The InstructionSetProbabilistic contains a list of VM_INSTRUCTION_OPERATIONs
// and their respective proabilities.
typedef std::vector<InstructionProbability> InstructionSetProbabilistic;

// This function normalises and calculates the cumulative probabilities of all
// the VM_INSTRUCTION_OPERATIONs in the InstructionSetProbabilistic.
// After this, you can use a 0.0f-1.0f random number to select an instruction
// from the set.
void InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(InstructionSetProbabilistic *isp)
{
   assert(isp != NULL);

   float total_probability = 0.0f;

   for (unsigned int ip = 0; ip < isp->size(); ip++)
   {
      total_probability += isp->at(ip).probability;
   }

   float cumulative_probability = 0.0f;

   for (unsigned int ip = 0; ip < isp->size(); ip++)
   {
      isp->at(ip).probability /= total_probability;

      // This holds the upper bound of the instruction in the 0.0f - 1.0f interval!
      isp->at(ip).cumulative_probability = cumulative_probability += isp->at(ip).probability;
   }
}

// Print a human readable version of the contents of this InstructionSetProbabilistic
void InstructionSetProbabilistic_Print(InstructionSetProbabilistic *instruction_set)
{
   for (InstructionSetProbabilistic::iterator i  = instruction_set->begin();
                                              i != instruction_set->end();
                                              i++)
   {
      printf("%-10s, %5.3f (<= %5.3f)\r\n",
                vm_instruction_operation_metadata[i->operation].instruction_mnemonic,
                i->probability,
                i->cumulative_probability);
   }
}

// Fill in the fields of this instruction randomly
// in a sensible way: LOADS can select
// a random symbol table key and a random register
// to load... ADD can select three random registers, etc.

// If the operand number is less than zero, we fill in the instruction completely.
// If the operand number is non zero, we fill in the appropriate argument.
// If the operand number is out of range for the current instruction
// we assertion fail.
void Instruction_FurnishInstruction(Instruction *instruction, SymbolTable *symbol_table, int operand_no = -1)
{
   assert(instruction != NULL);

   assert(operand_no < (int)INSTRUCTION_OPERANDS_MAX_COUNT);

   VM_INSTRUCTION_OPERATION &operation = instruction->operation;

   switch (operation)
   {
      case (VMI_C_HALT)  : return;

      case (VMI_A_ADD)   :
      case (VMI_A_SUB)   :
      case (VMI_A_MUL)   :
      case (VMI_A_DIVP)  :
      case (VMI_J_JRELT) :
      case (VMI_J_JRELE) :
      case (VMI_J_JREGE) :
      case (VMI_J_JREGT) :
      case (VMI_J_JREEQ) :
      case (VMI_J_JRENE) :
      {
         for (unsigned int operand = 0; operand < INSTRUCTION_OPERANDS_MAX_COUNT; operand++)
         {
            if ((operand_no < 0) || (operand_no == (int)operand))
            {
               instruction->operand[operand].as_register = VirtualMachine_RandomGeneralPurposeRegister();
            }
         }
      }
      break;

      case (VMI_S_PUSH)  :
      case (VMI_S_POP)   :
      case (VMI_S_TOP)   :
      {
         instruction->operand[0].as_register = VirtualMachine_RandomGeneralPurposeRegister();
      }
      break;

      case (VMI_V_LOADV) :
      {
         if ((operand_no < 0) || (operand_no == 0))
         {
            instruction->operand[0].as_register = VirtualMachine_RandomGeneralPurposeRegister();
         }

         if ((operand_no < 0) || (operand_no == 1))
         {
            int g = random_int(-10, 10);

            instruction->operand[1].as_direct_value = g;
         }
      }
      break;

      case (VMI_M_LOADS) :
      case (VMI_M_STORS) :
      {
         if ((operand_no < 0) || (operand_no == 0))
         {
            instruction->operand[0].as_register = VirtualMachine_RandomGeneralPurposeRegister();
         }

         if ((operand_no < 0) || (operand_no == 1))
         {
            RANDOM_SYMBOL_TABLE_KEY_NATURE nature = (operation == VMI_M_LOADS) ? RANDOM_SYMBOL_TABLE_KEY_READABLE
                                                                               : RANDOM_SYMBOL_TABLE_KEY_WRITABLE;

            instruction->operand[1].as_symbol_table_key = SymbolTable_RandomSymbolTableKey(symbol_table, nature);
         }
      }
      break;

      default :
      {
         assert(false);
      }
   }
}

// Return a random operation from a normalised InstructionSetProbabilistic.
VM_INSTRUCTION_OPERATION InstructionSetProbabilistic_RandomOperation(InstructionSetProbabilistic *i)
{
   // What instruction goes here?
   unsigned int instruction_in_set;

   // The instruction who's cumulative probability is higher than this float value!
   float random_value = random_float();

   for (instruction_in_set = 0; instruction_in_set < i->size(); instruction_in_set++)
   {
      if (random_value < (i->at(instruction_in_set).cumulative_probability))
      {
         break;
      }
   }

   // Numerical errors (or a non cumulativesed instruction set)
   // might make the above test fail. Let's just jam it at
   // the last instruction in the list.
   if (instruction_in_set == i->size())
   {
      instruction_in_set--;
   }

   assert(instruction_in_set < i->size());

   VM_INSTRUCTION_OPERATION operation = i->at(instruction_in_set).operation;

   return operation;
}

// Return an InstructionString consisting of length_in_instructions
// random instructions using the probabilities stored in the
// InstructionSetProbabilistic and the given SymbolTable.

// The instruction set should be normalise and cumulativesed before inserting
// into this function!
InstructionString *InstructionString_NewRandomInstructionStringFromInstructionSetProbabilistic(
                              InstructionSetProbabilistic *i,
                              SymbolTable *symbol_table,
                              unsigned int length_in_instructions)
{
   assert(i != NULL);
   assert(length_in_instructions != 0);

   InstructionString *s = new InstructionString;

   for (unsigned int instruction_slot = 0; instruction_slot < length_in_instructions; instruction_slot++)
   {
      s->push_back(Instruction());
      Instruction *current_instruction = (&(*(--s->end())));

      // The current instruction's operation is the operation
      // in the instruction set list at the selected instruction_in_set slot!
      current_instruction->operation = InstructionSetProbabilistic_RandomOperation(i);

      // Complete the random intialisation of this instruction given its type.
      Instruction_FurnishInstruction(current_instruction, symbol_table);
   }

   return s;
}

// A GeneticLinearProgram candidate consists of an InstructionString
// and some other metadata indicating its performance.
struct GeneticLinearProgram
{
   InstructionString *main_program;

   float fitness;

   unsigned int hits;
};

// Insert records into the ParseTreeEvaluationLog
// reflecting the automatic initialisation of
// SymbolTable entries with known values.
void ParseTreeEvaluationLog_InsertInitialisationSteps(std::map<SYMBOL_TABLE_KEY, VM_TYPE> *starting_values,
                                                      SymbolTable *symbol_table,
                                                      ParseTreeEvaluationLog *log)
{
   // Insert records as if we loaded the starting values into memory manually.
   for (SymbolTableData::iterator i = symbol_table->symbol_table_data.begin(); i != symbol_table->symbol_table_data.end(); i++)
   {
      const std::pair<SYMBOL_TABLE_KEY const, SymbolicVariableData> *s = &(*i);

      log->push_back(ParseTreeEvaluationLogEntry(LOGGED_OPERATION_INITIALISATION,
                                                 ParseTreeEvaluationResult(),
                                                 ParseTreeEvaluationResult(),
                                                 ParseTreeEvaluationResult((*starting_values)[s->first], PARSE_TREE_COMPLEXITY_INITIALISING_CONSTANT)));
   }
}

// A fitness case consists of:
struct FitnessCase
{
   // Starting variable valuations.
   std::map<SYMBOL_TABLE_KEY, VM_TYPE> starting_values;

   // Ending variable valuations.
   std::map<SYMBOL_TABLE_KEY, VM_TYPE> ending_values;

   // Desired return value.
   VM_TYPE return_value;

   // Model operation checklist.
   ParseTreeEvaluationLog model_operation_log;
};

// Construct fitness_case_count fitness cases reflecting the
// behaviour of the target program.
void FitnessCase_InitialiseCases(
      ParseTreeNode *target_program,
        SymbolTable *symbol_table,
        FitnessCase *fitness_case_array,
       unsigned int  fitness_case_count)
{
   // Bounds for variables!
   std::map<SYMBOL_TABLE_KEY, BoundsPair<int> > starting_variable_bounds;

   for (SymbolTableData::iterator r = symbol_table->symbol_table_data.begin(); r != symbol_table->symbol_table_data.end(); r++)
   {
      SYMBOL_TABLE_KEY const     &varkey = r->first;

      const SymbolicVariableData &data   = r->second; (void)data;

      // All variables have this bounds for now.
      starting_variable_bounds[varkey] = BoundsPair<int>(5, 50000);
   }

   // For each fitness case:
   for (unsigned int ci = 0; ci < fitness_case_count; ci++)
   {
      FitnessCase &c = fitness_case_array[ci];

      ParseTreeEvaluationEnvironment environment;
      ParseTreeEvaluationEnvironment_Initialise(&environment, symbol_table);

      // Produce some random values for the starting
      // state variables.
      for (SymbolTableData::iterator r = symbol_table->symbol_table_data.begin(); r != symbol_table->symbol_table_data.end(); r++)
      {
         SYMBOL_TABLE_KEY const     &varkey = r->first;

         const SymbolicVariableData &data   = r->second; (void)data;

         if (data.symbolic_constant)
         {
            c.starting_values[varkey] = data.value;
         }
         else
         {
            c.starting_values[varkey] = random_int(starting_variable_bounds[varkey].lower,
                                                   starting_variable_bounds[varkey].upper);
         }

         environment.variable_memory[varkey] = ParseTreeEvaluationResult(c.starting_values[varkey], 0);
      }


//      printf("Program:\r\n");
//      ParseTreeNode_Print_AsReadableExpression(target_program);
//      printf("\r\n");
//      printf("\r\n");
//
//      unsigned int sym;
//
//      sym = 0;
//
//      printf("Before:\r\n");
//      for (SymbolTableData::iterator r = symbol_table.symbol_table_data.begin(); r != symbol_table.symbol_table_data.end(); r++, sym++)
//      {
//         SYMBOL_TABLE_KEY const         &varkey = r->first;
//
//         const SymbolicVariableData &data    = r->second;
//
//         printf("Varname %2d: %7s = %5d\r\n", sym, varname, environment.indexed_memory[data.memory_mapped_location]);
//      }
//      printf("\r\n");

      // Let's run the program, and see what happens.
      ParseTreeEvaluationLog_InsertInitialisationSteps(&c.starting_values, symbol_table, &c.model_operation_log);

      ParseTreeNode_Evaluate(target_program, &environment, &c.model_operation_log);

      // Retrieve the ending values for the variables.
      for (SymbolTableData::iterator r = symbol_table->symbol_table_data.begin(); r != symbol_table->symbol_table_data.end(); r++)
      {
         SYMBOL_TABLE_KEY const     &varkey = r->first;

         const SymbolicVariableData &data   = r->second; (void)data;

         c.ending_values[varkey] = environment.variable_memory[varkey].value;
      }

      if (environment.return_occured)
      {
         c.return_value = environment.returned_value.value;
      }
      else
      {
         c.return_value = 0;
      }

//      sym = 0;
//
//      printf("After:\r\n");
//      for (SymbolTableData::iterator r = symbol_table.symbol_table_data.begin(); r != symbol_table.symbol_table_data.end(); r++, sym++)
//      {
//         SYMBOL_TABLE_KEY const         &varkey = r->first;
//
//         const SymbolicVariableData &data    = r->second;
//
//         printf("Varname %2d: %7s = %5d\r\n", sym, varname, environment.indexed_memory[data.memory_mapped_location]);
//      }
//      printf("\r\n");

      for (unsigned int log_entry = 0; log_entry < c.model_operation_log.size(); log_entry++)
      {
         ParseTreeEvaluationLogEntry_Print(&(c.model_operation_log[log_entry]));
      }
      printf("\r\n");

      wait_for_return();
   }
}

// VMERS contains the status of a register at a
// given point in the program.
struct VirtualMachineExecutionRegisterStatus
{
   bool register_contents_determinate_resulting;

   bool register_read_from_this_instruction;
   bool register_written_to_this_instruction;

   bool register_changed_value_due_to_this_instruction;
};

// There is a VMERS for each register in the virtual machine.
struct VirtualMachineExecutionRegisterStatusRecord
{
   VirtualMachineExecutionRegisterStatus registers[VM_REGISTER_COUNT];
};

// VMERST contains the VMERSR before and after each instruction execution.
typedef std::vector<VirtualMachineExecutionRegisterStatusRecord> VirtualMachineExecutionRegisterStatusTimeline;

// Interpret a VirtualMachineExecutionLog and populate a
// VirtualMachineExecutionRegisterStatusLine containing
// the status of each register before and after each instruction.
void VirtualMachineExecutionRegisterStatusTimeline_PopulateFromVirtualMachineExecutionLog(
                           VirtualMachineExecutionRegisterStatusTimeline *timeline,
                                              VirtualMachineExecutionLog *log)
{
   // Assume that a record already exists in the VirtualMachineExecutionLog
   // indicating that the values from the symbol table
   // have been loaded from indexed memory into the main
   // memory (the C function call model) and that
   // the register determinacy has already been determined.
   // (So, record 0 in the log has register determinacies for
   // the state of the system before instruction 0 has been executed.)

   assert(timeline != NULL);
   assert(log != NULL);

   unsigned int r = 0;

   for (VirtualMachineExecutionLog::iterator i = log->begin(); i != log->end(); i++)
   {
      VirtualMachineExecutionRegisterStatusRecord *previous_record = &(timeline->at(r++));

      VirtualMachineInstructionExecutionRecord *current_instruction = &(*i);

      // This is the new timeline step we're populating.
      VirtualMachineExecutionRegisterStatusRecord current_record;

      // Now that we know the register determinacies from the
      // previous time step, we can use the current instruction
      // to work out the states of everything!

      for (unsigned int r = 0; r < VM_REGISTER_COUNT; r++)
      {
         current_record.registers[r].register_contents_determinate_resulting =
            previous_record->registers[r].register_contents_determinate_resulting;
         current_record.registers[r].register_read_from_this_instruction  = false;
         current_record.registers[r].register_written_to_this_instruction = false;

         current_record.registers[r].register_changed_value_due_to_this_instruction =
            (current_instruction->registers_after[r] != current_instruction->registers_before[r]);
      }

      // Depending on the instruction executed, do different analysis.
      switch (current_instruction->instruction)
      {
         case (VMI_A_ADD)  :
         case (VMI_A_SUB)  :
         case (VMI_A_MUL)  :
         case (VMI_A_DIVP) :
         {
            // Arithmetic operations read the operand registers and
            // write to the result register.
            current_record.registers[current_instruction->operand_1_location.as_register].register_read_from_this_instruction = true;
            current_record.registers[current_instruction->operand_2_location.as_register].register_read_from_this_instruction = true;
            bool result_is_determinate =    previous_record->registers[current_instruction->operand_1_location.as_register].register_contents_determinate_resulting
                                         && previous_record->registers[current_instruction->operand_2_location.as_register].register_contents_determinate_resulting;

            current_record.registers[current_instruction->result_location.as_register].register_written_to_this_instruction = true;
            current_record.registers[current_instruction->result_location.as_register].register_contents_determinate_resulting = result_is_determinate;
         }
         break;
         case (VMI_V_LOADV) :
         {
            // Loading a constant directly.
            current_record.registers[current_instruction->result_location.as_register].register_written_to_this_instruction = true;
            current_record.registers[current_instruction->result_location.as_register].register_contents_determinate_resulting = true;
         }
         break;
         case (VMI_M_LOADS) :
         {
            // Loading from memory.
            current_record.registers[current_instruction->result_location.as_register].register_written_to_this_instruction = true;
            current_record.registers[current_instruction->result_location.as_register].register_contents_determinate_resulting = true;
         }
         break;
         case (VMI_M_STORS) :
         {
            // Storing a value to memory
            current_record.registers[current_instruction->result_location.as_register].register_read_from_this_instruction = true;
         }
         break;
//         case (VMI_RJ_JLT) :
//         case (VMI_RJ_JLE) :
//         case (VMI_RJ_JGE) :
//         case (VMI_RJ_JGT) :
//         case (VMI_RJ_JEQ) :
//         case (VMI_RJ_JNE) :
//         {
//            // Jumps read the values of registers a and c
//            current_record.registers[current_instruction->result_register].register_read_from_this_instruction = true;
//            current_record.registers[current_instruction->operand_2_register].register_read_from_this_instruction = true;
//         }
//         break;
         default :
         {
            assert(false);
         }
      }

      timeline->push_back(current_record);
   }
}

// Initialise a VirtualMachine by copying the starting state
// of a FitnessCase
void VirtualMachine_FitnessCase_PopulateStartingState(VirtualMachine *v, SymbolTable *symbol_table, FitnessCase *c)
{
   for (SymbolTableData::iterator r = symbol_table->symbol_table_data.begin(); r != symbol_table->symbol_table_data.end(); r++)
   {
      SYMBOL_TABLE_KEY const     &varkey = r->first;

      const SymbolicVariableData &data   = r->second; (void)data;

      v->data[varkey] = c->starting_values[varkey];
   }
}

// Inspect a ParseTreeEvaluationLogEntry/VirtualMachineInstructionExecutionRecord
// pair to see if they reflect the same (or similar) instructions.
bool ParseTreeEvaluationLogEntry_InstructionCorrelates(ParseTreeEvaluationLogEntry *le, VirtualMachineInstructionExecutionRecord *vi)
{
   switch (le->operation)
   {
      case (LOGGED_OPERATION_ADDITION) :
      {
         switch (vi->instruction)
         {
            case (VMI_A_ADD)  : return true;
            default           : return false;
         }
      }
      break;
      case (LOGGED_OPERATION_SUBTRACTION) :
      {
         switch (vi->instruction)
         {
            case (VMI_A_SUB)  : return true;
            default           : return false;
         }
      }
      break;
      case (LOGGED_OPERATION_MULTIPLICATION) :
      {
         switch (vi->instruction)
         {
            case (VMI_A_MUL)  : return true;
            default           : return false;
         }
      }
      break;
      case (LOGGED_OPERATION_DIVISION) :
      {
         switch (vi->instruction)
         {
            case (VMI_A_DIVP) : return true;
            default           : return false;
         }
      }
      break;
      default :
      {
         return false;
      }
   }
}

// Function for sorting an array of GeneticLinearProgram using qsort.
int qsort_GeneticLinearProgramByFitness(const void *a, const void *b)
{
   assert(a != NULL);
   assert(b != NULL);

   GeneticLinearProgram *program_a = **((GeneticLinearProgram ***)a);
   GeneticLinearProgram *program_b = **((GeneticLinearProgram ***)b);

   assert(program_a != NULL);
   assert(program_b != NULL);

//   if (program_a->fitness >  program_b->fitness) return -1;
//   if (program_a->fitness == program_b->fitness) return  0;
//   if (program_a->fitness <  program_b->fitness) return +1;
   if (program_a->fitness <  program_b->fitness) return +1; else return -1;

   return -1;
}

// Creates a new GeneticLinearProgram
// Don't call this directly unless you're manually constructing
// a program.
GeneticLinearProgram *GeneticLinearProgram_Internal_NewProgram()
{
   GeneticLinearProgram *p = new GeneticLinearProgram;

   return p;
}

// Free the memory used for storing a GeneticLinearProgram
// Don't call this directly.
void GeneticLinearProgram_Internal_FreeProgram(GeneticLinearProgram *node)
{
   assert(node != NULL);

   delete node;
}

// Create a new random GeneticLinearProgram with length_in_instructions
// instructions using the InstructionSetProbabilistic i.
GeneticLinearProgram *GeneticLinearProgram_NewRandomProgram(InstructionSetProbabilistic *i,
                                                                            SymbolTable *symbol_table,
                                                                           unsigned int  length_in_instructions)
{
   GeneticLinearProgram *p = GeneticLinearProgram_Internal_NewProgram();

   assert(p != NULL);

   p->main_program = InstructionString_NewRandomInstructionStringFromInstructionSetProbabilistic(i, symbol_table, length_in_instructions);

   assert(p->main_program);

   return p;
}

// Crossover produces two programs.
struct GeneticLinearProgram_Crossover_Pair
{
   GeneticLinearProgram *program[2];
};

// Duplicates a GeneticLinearProgram to produce a new GeneticLinearProgram.
GeneticLinearProgram *GeneticLinearProgram_DuplicateProgram(GeneticLinearProgram *program)
{
   assert(program != NULL);

   GeneticLinearProgram *new_program = GeneticLinearProgram_Internal_NewProgram();

   assert(new_program != NULL);

   new_program->main_program = InstructionString_DuplicateInstructionString(program->main_program);

   return new_program;
}

// Produce two child programs by executing the crossover operation
// on two parent programs.
GeneticLinearProgram_Crossover_Pair GeneticLinearProgram_Crossover2(GeneticLinearProgram *p0, GeneticLinearProgram *p1, unsigned int maximum_length_in_instructions)
{
   assert(p0 != NULL);
   assert(p1 != NULL);

   // Return a pair of GeneticLinearPrograms produced by crossover of the two programs given in p1 and p2.
   GeneticLinearProgram_Crossover_Pair output;

   output.program[0] = GeneticLinearProgram_Internal_NewProgram();
   output.program[1] = GeneticLinearProgram_Internal_NewProgram();

   // These are the parameters to the string compositing function!
   unsigned int instruction_to_start_switching_for_program[2];
   unsigned int instruction_count_to_switch_for_program[2];

   int invalid_switch = 1;

   while (invalid_switch)
   {
      invalid_switch = 0;

      // Program 0's switch can start anywhere.
      instruction_to_start_switching_for_program[0] = random_int(0, p0->main_program->size());

      // Program 0's switch length is dependent on
      // how many instructions are remaining!
      instruction_count_to_switch_for_program[0]    = random_int(0,  p0->main_program->size()
                                                                     - instruction_to_start_switching_for_program[0]);

      // Program 1's switch can start anywhere.
      instruction_to_start_switching_for_program[1] = random_int(0, p1->main_program->size());

      // Program 0's switch length is dependent on
      // how many instructions are remaining!
      instruction_count_to_switch_for_program[1]    = random_int(0,  p1->main_program->size()
                                                                     - instruction_to_start_switching_for_program[1]);

      // We can't allow zero sized strings to be created! (either way)
      unsigned int composite_length_in_instructions_p010 = p0->main_program->size() - instruction_count_to_switch_for_program[0] + instruction_count_to_switch_for_program[1];
      unsigned int composite_length_in_instructions_p101 = p1->main_program->size() - instruction_count_to_switch_for_program[1] + instruction_count_to_switch_for_program[0];

      if (composite_length_in_instructions_p010 == 0)
      {
         invalid_switch = 1;
         continue;
      }

      if (composite_length_in_instructions_p101 == 0)
      {
         invalid_switch = 1;
         continue;
      }

      if (composite_length_in_instructions_p010 > maximum_length_in_instructions)
      {
         invalid_switch = 1;
         continue;
      }

      if (composite_length_in_instructions_p101 > maximum_length_in_instructions)
      {
         invalid_switch = 1;
         continue;
      }
   }

   // Interleave the strings in the format 111112222221111.
   output.program[0]->main_program = InstructionString_ConstructNewCompositeString(
                                                             p0->main_program,
                                                             p1->main_program,
                                                             instruction_to_start_switching_for_program[0],
                                                             instruction_count_to_switch_for_program[0],
                                                             instruction_to_start_switching_for_program[1],
                                                             instruction_count_to_switch_for_program[1]);

   // Now do inverse!
   output.program[1]->main_program = InstructionString_ConstructNewCompositeString(
                                                             p1->main_program,
                                                             p0->main_program,
                                                             instruction_to_start_switching_for_program[1],
                                                             instruction_count_to_switch_for_program[1],
                                                             instruction_to_start_switching_for_program[0],
                                                             instruction_count_to_switch_for_program[0]);

   // The inverse operation interleaves the strings in the way 22221111111222.

   return output;
}

// Free all the memory allocated to the full GeneticLinearProgram
void GeneticLinearProgram_DestroyProgram(GeneticLinearProgram *p)
{
   assert(p != NULL);

   delete p->main_program;

   GeneticLinearProgram_Internal_FreeProgram(p);
}

// Mutate a GeneticLinearProgram program by
// altering a random instruction.
GeneticLinearProgram *GeneticLinearProgram_MutateProgram_AlterRandomInstruction(GeneticLinearProgram *program, InstructionSetProbabilistic *is, SymbolTable *symbol_table)
{
   assert(program != NULL);

   GeneticLinearProgram *new_program = GeneticLinearProgram_DuplicateProgram(program);

   // Do a mutation!
   // Pick an instruction!
   unsigned int in = random_int(0, new_program->main_program->size() - 1);

   Instruction *i = InstructionString_RetrieveInstruction(new_program->main_program, in);

   const float MUTATION_RATE_MUTATE_OPERATION = 0.5f;
   const float MUTATION_RATE_MUTATE_OPERAND   = 1.0f - MUTATION_RATE_MUTATE_OPERATION;

   const float CUMULATIVE_MUTATION_RATE_MUTATE_OPERATION = MUTATION_RATE_MUTATE_OPERATION;
   const float CUMULATIVE_MUTATION_RATE_MUTATE_OPERAND   = CUMULATIVE_MUTATION_RATE_MUTATE_OPERATION
                                                             + MUTATION_RATE_MUTATE_OPERAND;

   (void)CUMULATIVE_MUTATION_RATE_MUTATE_OPERAND;

   float mutation_target_float = random_float();

   if ((mutation_target_float < CUMULATIVE_MUTATION_RATE_MUTATE_OPERATION)
    || (vm_instruction_operation_metadata[i->operation].operands == 0))
   {
      // Pick a new operation.
      // This involves DISCARDING THE ENTIRE INSTRUCTION!
      i->operation = InstructionSetProbabilistic_RandomOperation(is);

      Instruction_FurnishInstruction(i, symbol_table);
   }
   else
   {
      unsigned int operand_to_mutate = random_int_0_to_(vm_instruction_operation_metadata[i->operation].operands - 1);

      Instruction_FurnishInstruction(i, symbol_table, operand_to_mutate);
   }

   return new_program;
}

// Mutate a GeneticLinearProgram program by
// deleting a random instruction.
GeneticLinearProgram *GeneticLinearProgram_MutateProgram_DeleteRandomInstruction(GeneticLinearProgram *program)
{
   assert(program != NULL);

   // Can't take an instruction from a program of length one.
   if (program->main_program->size() == 1)
   {
      return GeneticLinearProgram_DuplicateProgram(program);
   }

   unsigned int instruction_to_omit = random_int_0_to_(program->main_program->size() - 1);

   GeneticLinearProgram *new_program = GeneticLinearProgram_Internal_NewProgram();

   new_program->main_program = InstructionString_ConstructNewCompositeString(program->main_program,
                                                                             program->main_program,
                                                                             instruction_to_omit,
                                                                             1,
                                                                             0,
                                                                             0);

   return new_program;
}

// Mutate a GeneticLinearProgram program by
// inserting a random instruction.
GeneticLinearProgram *GeneticLinearProgram_MutateProgram_InsertRandomInstruction(GeneticLinearProgram *program, SymbolTable *symbol_table, InstructionSetProbabilistic *instruction_set)
{
   assert(program != NULL);

   unsigned int position_to_insert_at = random_int_0_to_(program->main_program->size());

   GeneticLinearProgram *new_program = GeneticLinearProgram_Internal_NewProgram();

   new_program->main_program = InstructionString_DuplicateInstructionString(program->main_program);

   Instruction instruction;
   instruction.operation = InstructionSetProbabilistic_RandomOperation(instruction_set);

   Instruction_FurnishInstruction(&instruction, symbol_table, -1);

   InstructionString_InsertInstruction(new_program->main_program, position_to_insert_at, &instruction);

   return new_program;
}

// Mutate a GeneticLinearProgram program by
// performing an operation chosed from the above randomly.
GeneticLinearProgram *GeneticLinearProgram_MutateProgram(GeneticLinearProgram *program, InstructionSetProbabilistic *is, SymbolTable *symbol_table)
{
   assert(program != NULL);
   assert(symbol_table != NULL);

   const float MUTATION_RATE_MUTATE_DELETE_RANDOM_INSTRUCTION = 0.3334f;
   const float MUTATION_RATE_MUTATE_ALTER_RANDOM_INSTRUCTION  = 0.3333f;
   const float MUTATION_RATE_MUTATE_INSERT_RANDOM_INSTRUCTION = 0.3333f;

   const float CUMULATIVE_MUTATION_RATE_MUTATE_DELETE_RANDOM_INSTRUCTION = MUTATION_RATE_MUTATE_DELETE_RANDOM_INSTRUCTION;
   const float CUMULATIVE_MUTATION_RATE_MUTATE_ALTER_RANDOM_INSTRUCTION  = CUMULATIVE_MUTATION_RATE_MUTATE_DELETE_RANDOM_INSTRUCTION
                                                                             + MUTATION_RATE_MUTATE_ALTER_RANDOM_INSTRUCTION;
   const float CUMULATIVE_MUTATION_RATE_MUTATE_INSERT_RANDOM_INSTRUCTION = CUMULATIVE_MUTATION_RATE_MUTATE_ALTER_RANDOM_INSTRUCTION
                                                                             + MUTATION_RATE_MUTATE_INSERT_RANDOM_INSTRUCTION;

   (void)CUMULATIVE_MUTATION_RATE_MUTATE_INSERT_RANDOM_INSTRUCTION;

   float mutation_action_float = random_float();

   if (mutation_action_float < CUMULATIVE_MUTATION_RATE_MUTATE_DELETE_RANDOM_INSTRUCTION)
   {
      return GeneticLinearProgram_MutateProgram_DeleteRandomInstruction(program);
   }
   else
   if (mutation_action_float < CUMULATIVE_MUTATION_RATE_MUTATE_ALTER_RANDOM_INSTRUCTION)
   {
      return GeneticLinearProgram_MutateProgram_AlterRandomInstruction(program, is, symbol_table);
   }
   else
   {
      return GeneticLinearProgram_MutateProgram_InsertRandomInstruction(program, symbol_table, is);
   }
}

// A sorted genetic linear program population
// consists of a list of pointers to the slots in the population.
typedef std::list<GeneticLinearProgram **> SortedGeneticLinearProgramPopulation;

// Insert a new program into the SGLPP a, retaining the sorted property.
// This has linear complexity, but assumes that most programs have
// very high fitness, so the actual cost will be minimal.
unsigned int SortedGeneticLinearProgramPopulation_Insert(SortedGeneticLinearProgramPopulation &a, GeneticLinearProgram **p)
{
   SortedGeneticLinearProgramPopulation::iterator i = a.begin();

   unsigned int position = 0;

   for (; i != a.end(); i++)
   {
      float &current_fitness = (***i).fitness;

      if (current_fitness < (**p).fitness)
      {
         break;
      }

      position++;
   }

//   printf("Inserting %f at position %d.\r\n", (*p)->fitness, position);
//   printf("\r\n");
//
//   wait_for_return();

   a.insert(i, p);

   return position;
}

// Return an iterator to the program at position l
// Searches iteratively from the closest edge.
SortedGeneticLinearProgramPopulation::iterator SortedGeneticLinearProgramPopulation_At(SortedGeneticLinearProgramPopulation &a, unsigned int l)
{
   if (l < (a.size()/2))
   {
      SortedGeneticLinearProgramPopulation::iterator i = a.begin();

      while (l--)
      {
         i++;
      }

      return i;
   }
   else
   {
      SortedGeneticLinearProgramPopulation::iterator i = a.end();

      l = a.size() - l;

      while (l--)
      {
         i--;
      }

      return i;
   }
}


// Parameters for the evolution system.
struct EvolutionSystem_Parameters
{
   unsigned int population_size;

   // New random programs will be produced in a ramp between
   // these sizes.
   unsigned int initial_maximum_program_size_in_instructions;
   unsigned int initial_minimum_program_size_in_instructions;

   // Reattempt crossover if programs above this size are produced.
   unsigned int discard_crossover_programs_of_size_above;

   // For this steady state experiment, the variable 'new_creation' refers to
   // the number of new entities created thus far in the steady state system.
   unsigned int max_creations_per_run;

   unsigned int tournament_size_for_crossover;
   unsigned int tournament_size_for_reproduction;
   unsigned int tournament_size_for_mutation;

   unsigned int tournament_size_for_deletion;

   unsigned int max_runs;

   unsigned int repeat_experiment;

          float rate_of_crossover;
          float rate_of_mutation;
          float rate_of_reproduction;

   unsigned int fitness_case_training_set_count;
   unsigned int fitness_case_test_set_count;

          float maximum_acceptable_fitness;

   unsigned int instruction_executions_max_count;

   unsigned int vm_register_general_purpose_count;

           bool adaptive_instruction_set;
          float adaptive_instruction_set_weighting;

           bool attempt_embryonic_refinement;
          float population_embryo_contribution;

   unsigned int reporting_granularity_in_new_creations;
          float reporting_delay_in_seconds;

           bool reporting;

           bool interactive;

    std::string output_file;
};

EvolutionSystem_Parameters EvolutionSystem_Parameters_LoadFromFile(const std::string &file)
{
   EvolutionSystem_Parameters p;

   FILE *f = fopen(file.c_str(), "rb");

   if (f == NULL)
   {
      printf("File %s cannot be opened.\r\n", file.c_str());
      exit(0);
   }

   const unsigned int READ_BUFFER_LENGTH = 5000;

   char read_line[READ_BUFFER_LENGTH];
   memset(read_line, 0, READ_BUFFER_LENGTH);

   do
   {
      fgets(read_line, READ_BUFFER_LENGTH, f);

      if (feof(f)) break;

      char *read_from = read_line;

      while (*read_from == ' ')
      {
         if ((read_from - read_line) >= (int)strlen(read_line)) break;

         read_from++;
      }
      if ((read_from - read_line) >= (int)strlen(read_line)) continue;

      std::string read_string = String_LoseTrailingNewlines(std::string(read_from));

      std::string first  = String_ReturnWord(read_string, 0);
      std::string second = String_ReturnWord(read_string, 1);

      #define load_unsigned_int(ELEMENT) \
      if (first == #ELEMENT) {sscanf(second.c_str(), "%d", &(p.ELEMENT)); printf("Loaded '"#ELEMENT" = %d'\r\n", p.ELEMENT);}
      #define load_float(ELEMENT) \
      if (first == #ELEMENT) {sscanf(second.c_str(), "%f", &(p.ELEMENT)); printf("Loaded '"#ELEMENT" = %f'\r\n", p.ELEMENT);}
      #define load_bool(ELEMENT) \
      if (first == #ELEMENT) {p.ELEMENT = (stricmp(second.c_str(), "true") == 0); printf("Loaded '"#ELEMENT" = %s'\r\n", p.ELEMENT ? "true" : "false");}

   load_unsigned_int(population_size);

   load_unsigned_int(initial_maximum_program_size_in_instructions);
   load_unsigned_int(initial_minimum_program_size_in_instructions);

   load_unsigned_int(discard_crossover_programs_of_size_above);

   load_unsigned_int(max_creations_per_run);

   load_unsigned_int(tournament_size_for_crossover);
   load_unsigned_int(tournament_size_for_reproduction);
   load_unsigned_int(tournament_size_for_mutation);

   load_unsigned_int(tournament_size_for_deletion);

   load_unsigned_int(max_runs);
   load_unsigned_int(repeat_experiment);

          load_float(rate_of_crossover);
          load_float(rate_of_mutation);
          load_float(rate_of_reproduction);

   load_unsigned_int(fitness_case_training_set_count);
   load_unsigned_int(fitness_case_test_set_count);

          load_float(maximum_acceptable_fitness);

   load_unsigned_int(instruction_executions_max_count);

   load_unsigned_int(vm_register_general_purpose_count);

           load_bool(adaptive_instruction_set);
          load_float(adaptive_instruction_set_weighting);

          load_float(population_embryo_contribution);

   load_unsigned_int(reporting_granularity_in_new_creations);
          load_float(reporting_delay_in_seconds);

           load_bool(reporting);

           load_bool(interactive);
   } while (true);

   fclose(f);

   return p;
}

// Report on the status of evolution.
struct EvolutionSystem_Report
{
   bool success_on_training_set;
   bool success_on_test_set;

   unsigned int total_candidates_processed;
   unsigned int runs_attempted;
   unsigned int total_instructions_executed;

   EvolutionSystem_Report()
      : success_on_training_set(false),
        success_on_test_set(false),
        total_candidates_processed(0),
        runs_attempted(0),
        total_instructions_executed(0)
   {}
};



// Populate the fitness field of a GeneticLinearProgram
// by calculating its fitness against the fitness case training set.
void GeneticLinearProgram_FitnessEvaluationOnTrainingSet(
            GeneticLinearProgram *p,
      EvolutionSystem_Parameters *param,
                     FitnessCase *fitness_case_training_set,
                     SymbolTable *symbol_table,
                            bool  alert = false)
{
   assert(p != NULL);
   assert(param != NULL);
   assert(symbol_table != NULL);
   assert(fitness_case_training_set != NULL);

   float fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_COUNT];

   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_ZERO                        ] =  0.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_HALT_INSTRUCTION_ENCOUNTERED] =  0.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_INSTRUCTION_OUT_OF_RANGE    ] = 20.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_INSTRUCTION_UNUSUAL_INPUT   ] = 30.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_MEMORY_ACCESS_OUT_OF_BOUNDS ] = 30.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_PROGRAM_COUNTER_OUT_OF_RANGE] = 10.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_DIVISION_BY_ZERO            ] = 30.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_INSTRUCTION_LIMIT_REACHED   ] = 30.0f;
   fitness_modifiers_for_bad_states[VM_STATE_BAD_REASON_UNIDENTIFIED_ERROR_STATE    ] = 50.0f;

   bool indeterminate_read_fails = true;
   bool indeterminate_write_fails = true;

   float fitness = 10000000.0f;

   p->hits = 0;

   VirtualMachine *v = new VirtualMachine;

   // T is the training set cases.

   //printf("This program:\r\n");
   //InstructionString_Print(p->main_program);

   for (unsigned int t = 0; t < param->fitness_case_training_set_count; t++)
   {
      // Blank virtual machine
      VirtualMachine_Initialise(v, symbol_table);

      // New log.
      VirtualMachineExecutionLog log;

      // This is the case we're inspecting.
      FitnessCase &c = fitness_case_training_set[t];

      // Populate the virtual machine's memory space with
      // the variables values from the case.
      VirtualMachine_FitnessCase_PopulateStartingState(v, symbol_table, &c);

      VirtualMachineInstructionString_ExecuteInstructionString(v,
                                                               p->main_program,
                                                               &log,
                                                               INSTRUCTION_EXECUTIONS_MAX_COUNT,
                                                               VM_EXECUTE_NONINTERACTIVE);

      float fitness_penalties = 0.0f;
      float fitness_bonuses   = 0.0f;
      float fitness_variable_penalty_for_error = 0.0f;

      // Identify hit.
      bool hit = true;

      // Hits have to have all the variables with the correct value.
      for (SymbolTableData::iterator r  = symbol_table->symbol_table_data.begin();
                                     r != symbol_table->symbol_table_data.end();
                                     r++)
      {
         SYMBOL_TABLE_KEY const     &varkey = r->first;

         const SymbolicVariableData &data   = r->second;

         if ((!data.expected_match_new_state)
          && (!data.expected_match_old_state))
         {
            continue;
         }

         VM_TYPE desired = 0, actual = v->data[varkey];

         if (data.expected_match_new_state)
         {
            desired = c.ending_values[varkey];
         }
         else
         if (data.expected_match_old_state)
         {
            desired = c.starting_values[varkey];
         }

         if (actual != desired)
         {
            const float FITNESS_VARIABLE_PENALTY_FOR_ERROR_MAXIMUM = 1000.0f;

            BigInteger error = desired - actual;

            fitness_variable_penalty_for_error = fabs((float)bigIntegerToInt(error));

            fitness_variable_penalty_for_error = min(fitness_variable_penalty_for_error,  FITNESS_VARIABLE_PENALTY_FOR_ERROR_MAXIMUM);
            fitness_variable_penalty_for_error = max(fitness_variable_penalty_for_error, -FITNESS_VARIABLE_PENALTY_FOR_ERROR_MAXIMUM);

            hit = false;

            if (alert)
            {
               //printf("Symbol %s fail. %s instead of %s.\r\n", data.variable_symbolic_name.c_str(), bigIntegerToString(actual).c_str(), bigIntegerToString(desired).c_str());
            }
         }
      }

      VirtualMachineExecutionRegisterStatusTimeline timeline;

      // This record represents the pre-VM actions that may be
      // taken to place the input values into the registers
      // at program start.
      VirtualMachineExecutionRegisterStatusRecord initial_register_state;

      const bool registers_initialised_to_zero = false;

      if (registers_initialised_to_zero)
      {
         for (unsigned int r = 0; r <= vm_register_final_general_purpose; r++)
         {
            initial_register_state.registers[r].register_contents_determinate_resulting = true;
            initial_register_state.registers[r].register_written_to_this_instruction = true;

            v->registers[r] = 0;
         }
      }
      else
      {
         for (unsigned int r = 0; r <= vm_register_final_general_purpose; r++)
         {
            initial_register_state.registers[r].register_contents_determinate_resulting = false;
            initial_register_state.registers[r].register_written_to_this_instruction    = false;

            unsigned int magnitude = random_int(20, 1000);
            bool sign = random_float() < 0.5f;

            v->registers[r] = sign ? magnitude : -magnitude;
         }

         indeterminate_read_fails  = true;
         indeterminate_write_fails = true;
      }

      timeline.push_back(initial_register_state);

      VirtualMachineExecutionRegisterStatusTimeline_PopulateFromVirtualMachineExecutionLog(&timeline, &log);


//      for (unsigned int r = 0; r < log.size(); r++)
//      {
//         VirtualMachineInstructionExecutionRecord_Print(&log[r], symbol_table);
//      }
//      system("pause");
//
//      for (VirtualMachineExecutionRegisterStatusTimeline::iterator i = timeline.begin(); i != timeline.end(); i++)
//      {
//         for (unsigned int r = 0; r < VM_REGISTER_COUNT; r++)
//         {
//            printf("r%d:%c ", r, i->registers[r].register_contents_determinate_resulting ? 'Y' : 'N');
//         }
//
//         printf("\r\n");
//      }



      // Now we have a timeline that indicates when the registers were read from and written to...
      // We now step through the timeline and try to work out when
      // rewrites were attempted on registers, when bad reads
      // were attempted, and when dangling writes were left.

      bool register_dangling_write[vm_register_final_general_purpose];

      for (unsigned int gpr = 0; gpr <= vm_register_final_general_purpose; gpr++)
      {
         // The registers have dangling writes if the input values
         // were placed into them at program start.
         register_dangling_write[gpr] = initial_register_state.registers[gpr].register_written_to_this_instruction;
      }

      const float FITNESS_PENALTY_FOR_REGISTER_DANGLING_WRITE                     =   20.0f;
      const float FITNESS_PENALTY_FOR_REGISTER_REWRITING                          =   50.0f;
      const float FITNESS_PENALTY_FOR_REGISTER_INDETERMINATE_READ                 = 3300.0f;
      const float FITNESS_PENALTY_FOR_NON_HIT                                     =108000.0f;
      const float FITNESS_PENALTY_FOR_STORING_INDETERMINATE_VALUE                 =  500.0f;

      const float FITNESS_PENALTY_FOR_ALTERING_PROGRAM_COUNTER_DIRECTLY           = 5000.0f;

      const float FITNESS_BONUS_FOR_READING_VALUE_FROM_MEMORY                     =    -4.0f;
      const float FITNESS_BONUS_FOR_WRITING_TO_VARYING_VARIABLE                   =    -0.0f;
      const float FITNESS_BONUS_FOR_WRITING_TO_VARYING_VARIABLE_EXPECTED_CHANGE   =    -3.0f;
      const float FITNESS_BONUS_FOR_REGISTER_DETERMINATE_READ                     =    -2.0f;
      const float FITNESS_BONUS_FOR_PRODUCING_USEFUL_VALUE                        =   -71.0f;
      const float FITNESS_BONUS_FOR_PRODUCING_USEFUL_CONSTANT                     = -3871.0f;
      const float FITNESS_BONUS_FOR_PRODUCING_USEFUL_VALUE_THEN_CORRECT_OPERATION =   -73.0f;

      VirtualMachineExecutionRegisterStatusTimeline::iterator before = timeline.begin();
      VirtualMachineExecutionRegisterStatusTimeline::iterator after  = timeline.begin(); after++;

      VirtualMachineExecutionLog::iterator instruction = log.begin();

      // Are the bonuses still available?
      // Multiple claims are probably not a good idea.
      unsigned int useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[c.model_operation_log.size()];

      for (unsigned int r = 0; r < c.model_operation_log.size(); r++)
      {
         const unsigned int USEFUL_VALUE_BONUS_REMAINING_FOR_ITEM_IN_PARSE_TREE_EVALUATION_LOG = 2;

         useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[r] = USEFUL_VALUE_BONUS_REMAINING_FOR_ITEM_IN_PARSE_TREE_EVALUATION_LOG;
      }

      for (; before != timeline.end(); before++, after++, instruction++)
      {
         VirtualMachineExecutionRegisterStatusRecord *b = &(*before);
         VirtualMachineExecutionRegisterStatusRecord *a = &(*after );

         VirtualMachineInstructionExecutionRecord *i = &(*instruction);

         // Detect whether the instruction execution we're
         // considering resulted in a reading of an indeterminate register.

         if (after != timeline.end())
         {
            if (a->registers[VM_REGISTER_PROGRAM_COUNTER].register_written_to_this_instruction == true)
            {
               fitness_penalties += FITNESS_PENALTY_FOR_ALTERING_PROGRAM_COUNTER_DIRECTLY;
            }

            for (unsigned int r = 0; r <= vm_register_final_general_purpose; r++)
            {
               if (b->registers[r].register_contents_determinate_resulting == false)
               {
                  if (a->registers[r].register_read_from_this_instruction == true)
                  {
                     // You read from a register whose value was possibly random!
                     fitness_penalties += FITNESS_PENALTY_FOR_REGISTER_INDETERMINATE_READ;

                     hit = hit && !indeterminate_read_fails;
                  }
               }
               else
               {
                  if (a->registers[r].register_read_from_this_instruction == true)
                  {
                     // Remove the dangling write flag for registers that were
                     // read during the last instruction
                     register_dangling_write[r] = false;

                     // You read from a valid register.
                     fitness_bonuses += FITNESS_BONUS_FOR_REGISTER_DETERMINATE_READ;
                  }

                  if (a->registers[r].register_written_to_this_instruction == true)
                  {
                     if (register_dangling_write[r] == true)
                     {
                        // You wrote to the same register twice
                        // without inspecting its value at least once!
                        fitness_penalties += FITNESS_PENALTY_FOR_REGISTER_REWRITING;
                     }

                     register_dangling_write[r] = true;
                  }
               }
            }

            switch (i->instruction)
            {
               case (VMI_A_ADD)  :
               case (VMI_A_SUB)  :
               case (VMI_A_MUL)  :
               case (VMI_A_DIVP) :
               {
                  // Does this operation change the target register?
                  if (after->registers[i->result_location.as_register].register_changed_value_due_to_this_instruction)
                  {
                     // Does this addition produce a nice value?
                     VM_TYPE &vm_result = i->result_value;

                     // Search the log of the fitness case!
                     unsigned int item = 0;
                     for (ParseTreeEvaluationLog::iterator lei = c.model_operation_log.begin(); lei != c.model_operation_log.end(); lei++, item++)
                     {
                        ParseTreeEvaluationLogEntry *e = &(*lei);

                        if (useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[item] == 0) continue;

                        if (e->result.value == vm_result)
                        {
                           fitness_bonuses += FITNESS_BONUS_FOR_PRODUCING_USEFUL_VALUE * (1+e->result.complexity_heuristic);
                           useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[item]--;

                           // Was the useful operation produced using an operation that resembles
                           // the operation undertaken by the parse tree?
                           if (ParseTreeEvaluationLogEntry_InstructionCorrelates(e, i))
                           {
                              /* TODO: Not used. *///fitness_bonuses += FITNESS_BONUS_FOR_PRODUCING_USEFUL_VALUE_THEN_CORRECT_OPERATION;
                           }

                           break;
                        }
                     }
                  }
               }
               break;
               case (VMI_M_LOADS) :
               {
                  // Does this operation change the target register?
                  if (after->registers[i->result_location.as_register].register_changed_value_due_to_this_instruction)
                  {
                     // Does this memory load retrieve a nice value?
                     VM_TYPE &vm_result = i->result_value;

                     // Search the log of the fitness case!
                     unsigned int item = 0;
                     for (ParseTreeEvaluationLog::iterator lei = c.model_operation_log.begin(); lei != c.model_operation_log.end(); lei++, item++)
                     {
                        ParseTreeEvaluationLogEntry *e = &(*lei);

                        if (useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[item] == 0) continue;

                        if (e->result.value == vm_result)
                        {
                           fitness_bonuses += FITNESS_BONUS_FOR_PRODUCING_USEFUL_VALUE;
                           fitness_bonuses += FITNESS_BONUS_FOR_READING_VALUE_FROM_MEMORY;
                           useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[item]--;

                           break;
                        }
                     }
                  }
               }
               break;
               case (VMI_M_STORS) :
               {
                  // What kind of variable was written to?
                  const SymbolicVariableData *symbolic_variable_data = SymbolTable_QuerySymbolTableKey(symbol_table, i->operand_1_location.as_symbol_table_key);

                  if (b->registers[i->result_location.as_register].register_contents_determinate_resulting == false)
                  {
                     // You calculated garbage and then stored it in memory.
                     fitness_penalties += FITNESS_PENALTY_FOR_STORING_INDETERMINATE_VALUE;

                     hit = hit && !indeterminate_write_fails;
                  }
                  else
                  {
                     if (symbolic_variable_data->expected_match_new_state)
                     {
                        fitness_bonuses += FITNESS_BONUS_FOR_WRITING_TO_VARYING_VARIABLE_EXPECTED_CHANGE;
                     }
                     else
                     {
                        fitness_bonuses += FITNESS_BONUS_FOR_WRITING_TO_VARYING_VARIABLE;
                     }
                  }
               }
               break;
               case (VMI_V_LOADV) :
               {
                  // Does this operation change the target register?
                  if (after->registers[i->result_location.as_register].register_changed_value_due_to_this_instruction)
                  {
                     // Does this constant load retrieve a nice value?
                     VM_TYPE &vm_result = i->result_value;

                     // Search the log of the fitness case!
                     unsigned int item = 0;
                     for (ParseTreeEvaluationLog::iterator lei = c.model_operation_log.begin(); lei != c.model_operation_log.end(); lei++, item++)
                     {
                        ParseTreeEvaluationLogEntry *e = &(*lei);

                        if (useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[item] == 0) continue;

                        if (e->result.value == vm_result)
                        {
                           if (e->operation == LOGGED_OPERATION_CONSTANT_LOAD)
                           {
                              fitness_bonuses += FITNESS_BONUS_FOR_PRODUCING_USEFUL_VALUE;

                              fitness_bonuses += FITNESS_BONUS_FOR_PRODUCING_USEFUL_CONSTANT;

                              useful_value_bonus_remaining_for_item_in_parse_tree_evaluation_log[item]--;
                           }

                           break;
                        }
                     }
                  }
               }
               break;
               case (VMI_C_HALT) :
               {

               }
               break;
               default :
               {
                  printf("Bad instruction: %s.\r\n", vm_instruction_operation_metadata[i->instruction].instruction_mnemonic);
                  assert(false);
               }
            }
         }
         else
         {
            // Check for dangling writes.
            for (unsigned int r = 0; r <= vm_register_final_general_purpose; r++)
            {
               if (register_dangling_write[r] == true)
               {
                  // You wrote to this register, but never
                  // inspected its value since then...
                  // so why did you write to it?
                  fitness_penalties += FITNESS_PENALTY_FOR_REGISTER_DANGLING_WRITE;
               }
            }
         }
      }

      if (hit)
      {
         p->hits++;
      }

      if (!hit)
      {
         fitness += fitness_penalties;
         fitness += fitness_bonuses;

         fitness += FITNESS_PENALTY_FOR_NON_HIT;
         fitness += fitness_variable_penalty_for_error;
      }
      else
      {
         fitness += fitness_modifiers_for_bad_states[v->state_bad];

         const float FITNESS_PENALTY_FOR_HIT_PROGRAMS_BY_INSTRUCTION_LENGTH = 100.0f;

         fitness += FITNESS_PENALTY_FOR_HIT_PROGRAMS_BY_INSTRUCTION_LENGTH
                      * p->main_program->size();
      }

   }

//   printf("This program's fitness = %f.\r\n", fitness);

   p->fitness = fitness;

//   ParseTreeEvaluationLog &log = fitness_case_training_set[0].model_operation_log;
//
//   for (unsigned int log_entry = 0; log_entry < log.size(); log_entry++)
//   {
//      ParseTreeEvaluationLogEntry_Print(&(log[log_entry]));
//   }
//
//   wait_for_return();

   delete v;
}

// Return a bool indicating if the given GeneticLinearProgram
// passes all of the fitness cases in the test set.
bool GeneticLinearProgram_AcceptanceTestOnTestSet(
            GeneticLinearProgram *p,
      EvolutionSystem_Parameters *param,
                     FitnessCase *fitness_case_test_set,
                     SymbolTable *symbol_table)
{
   assert(p != NULL);
   assert(param != NULL);
   assert(symbol_table != NULL);
   assert(fitness_case_test_set != NULL);

   bool hit_allowed_on_bad_state[VM_STATE_BAD_REASON_COUNT];

   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_ZERO                        ] = true;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_HALT_INSTRUCTION_ENCOUNTERED] = true;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_INSTRUCTION_OUT_OF_RANGE    ] = false;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_INSTRUCTION_UNUSUAL_INPUT   ] = false;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_MEMORY_ACCESS_OUT_OF_BOUNDS ] = false;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_PROGRAM_COUNTER_OUT_OF_RANGE] = true;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_DIVISION_BY_ZERO            ] = false;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_INSTRUCTION_LIMIT_REACHED   ] = false;
   hit_allowed_on_bad_state[VM_STATE_BAD_REASON_UNIDENTIFIED_ERROR_STATE    ] = false;

   unsigned int hits = 0;

   VirtualMachine *v = new VirtualMachine;

   for (unsigned int t = 0; t < param->fitness_case_test_set_count; t++)
   {
      // Blank virtual machine
      VirtualMachine_Initialise(v, symbol_table);

      // New log.
      VirtualMachineExecutionLog log;

      // This is the case we're inspecting.
      FitnessCase &c = fitness_case_test_set[t];

      // Populate the virtual machine's memory space with
      // the variables values from the case.
      VirtualMachine_FitnessCase_PopulateStartingState(v, symbol_table, &c);

      VirtualMachineInstructionString_ExecuteInstructionString(v,
                                                               p->main_program,
                                                               &log,
                                                               INSTRUCTION_EXECUTIONS_MAX_COUNT,
                                                               VM_EXECUTE_NONINTERACTIVE);

      // Identify hit.
      bool hit = true;

      // Hits have to have all the variables with the correct value.
      for (SymbolTableData::iterator r  = symbol_table->symbol_table_data.begin();
                                     r != symbol_table->symbol_table_data.end();
                                     r++)
      {
         SYMBOL_TABLE_KEY const     &varkey = r->first;

         const SymbolicVariableData &data   = r->second; (void)data;

         if ((!data.expected_match_new_state)
          && (!data.expected_match_old_state))
         {
            continue;
         }

         VM_TYPE desired = 0, actual = v->data[varkey];

         if (data.expected_match_new_state)
         {
            desired = c.ending_values[varkey];
         }
         else
         if (data.expected_match_old_state)
         {
            desired = c.starting_values[varkey];
         }

         if (actual != desired)
         {
            hit = false;
         }
      }

      hit = hit && hit_allowed_on_bad_state[v->state_bad];

      if (hit) hits++;
   }

   delete v;

   if (hits == param->fitness_case_test_set_count)
   {
      return true;
   }
   else
   {
      return false;
   }
}

// Attempt to evolve an InstructionString with the same
// semantics as the given input parse tree.
void EvolutionSystem_EvolveInstructionStringFromParseTree(
                   ParseTreeNode  *target_program,
          EvolutionSystem_Report  *output_report,
               InstructionString **output_string,
      EvolutionSystem_Parameters  *param,
                            bool   reporting,
     InstructionSetProbabilistic  *instruction_set,
                     SymbolTable  *symbol_table,
               InstructionString  *input_string_embryo)
{
   assert(target_program  != NULL);
   assert(output_report   != NULL);
   assert(output_string   != NULL);
   assert(param           != NULL);
   assert(instruction_set != NULL);
   assert(symbol_table    != NULL);

   SymbolTable_ReaffirmContainingTypes(symbol_table);

   vm_register_general_purpose_count = param->vm_register_general_purpose_count;
   vm_register_final_general_purpose = vm_register_general_purpose_count - 1;

   INSTRUCTION_EXECUTIONS_MAX_COUNT = param->instruction_executions_max_count;

   const unsigned int FITNESS_CASE_TOTAL_COUNT = param->fitness_case_training_set_count
                                                + param->fitness_case_test_set_count;

   FitnessCase fitness_case[FITNESS_CASE_TOTAL_COUNT];

   FitnessCase *fitness_case_training_set = fitness_case;
   FitnessCase *fitness_case_test_set     = fitness_case + param->fitness_case_training_set_count;

   FitnessCase_InitialiseCases(target_program, symbol_table, fitness_case, FITNESS_CASE_TOTAL_COUNT);

   const float CUMULATIVE_RATE_OF_CROSSOVER    = param->rate_of_crossover;
   const float CUMULATIVE_RATE_OF_REPRODUCTION = param->rate_of_reproduction + CUMULATIVE_RATE_OF_CROSSOVER;
   const float CUMULATIVE_RATE_OF_MUTATION     = param->rate_of_mutation     + CUMULATIVE_RATE_OF_REPRODUCTION;

   clock_t report_time = clock();
   clock_t start_time = report_time;

   *output_report = EvolutionSystem_Report();

   bool acceptable_program_found = false;

   InstructionString *acceptable_program = NULL;

   printf("\r\n");
   printf("\r\n");
   printf("BEGIN EVOLUTION:\r\n");

   printf("Trying to evolve program:\r\n");

   printf("\r\n    ");
   ParseTreeNode_Print_AsReadableExpression(target_program, symbol_table);
   printf("\r\n");

   printf("\r\n");

      FILE *f = fopen(param->output_file.c_str(), "ab");
      fprintf(f, "evolve %s\r\n",
                  ParseTreeNode_Print_ToString(target_program, symbol_table).c_str());
      fclose(f);

   SymbolTable_Print(symbol_table);

   InstructionSetProbabilistic instruction_set_adapted;

   if (param->adaptive_instruction_set)
   {
      InstructionSetProbabilistic adaptation_contribution;

      std::map<const ParseTreeInteriorNodeSetMember *, unsigned int> interior_node_population;

      ParseTreeSourceProgram_CalculateOperatorLikeSymbolPopulations(target_program, interior_node_population);

      for (std::map<const ParseTreeInteriorNodeSetMember *, unsigned int>::iterator i  = interior_node_population.begin();
                                                                                    i != interior_node_population.end();
                                                                                    i++)
      {
         switch (i->first->as_operator_like_symbol.operator_like_symbol)
         {
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_PLUS) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_ADD,  (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MINUS) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_SUB,  (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MULTIPLY) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_MUL,  (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_DIVIDE) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_DIVP, (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_ASSIGNMENT) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_M_STORS, (float)i->second));
            }
            break;
            default :
            {
               // Don't care.
            }
         }
      }
      InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(&adaptation_contribution);

      instruction_set_adapted = *instruction_set;
      instruction_set_adapted.insert(instruction_set_adapted.end(), adaptation_contribution.begin(), adaptation_contribution.end());

      InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(&instruction_set_adapted);
   }
   else
   {
      instruction_set_adapted = *instruction_set;

      InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(&instruction_set_adapted);
   }

   printf("\r\n\r\nInstruction Set:\r\n");
   InstructionSetProbabilistic_Print(&instruction_set_adapted);

   for (unsigned int run = 0; run < param->max_runs; run++)
   {
      output_report->runs_attempted++;

      //printf("START OF RUN %d\r\n", run);

      // This is where the programs are actually stored.
      GeneticLinearProgram **population = new GeneticLinearProgram *[param->population_size];

      // Generate programs through 'ramped full'
      const unsigned int LENGTH_RAMP_SIZE = param->population_size / ((param->initial_maximum_program_size_in_instructions
                                                                         - param->initial_minimum_program_size_in_instructions) + 1);

      // Construct a series of programs of ramped length, that begin with programs
      // that are INITIAL_MINIMUM_PROGRAM_SIZE_IN_INSTRUCTIONS
      // and end at INITIAL_MAXIMUM_PROGRAM_SIZE_IN_INSTRUCTIONS

      //printf("New population member:\r\n");
      for (unsigned int n = 0; n < param->population_size; n++)
      {
         //printf(" %d", n);

         unsigned int current_ramp_step = n / LENGTH_RAMP_SIZE;

         unsigned int current_size_in_instructions = param->initial_minimum_program_size_in_instructions
                                                      + current_ramp_step;

         GeneticLinearProgram *new_program;

         bool create_random_program = false;

         create_random_program = (!((input_string_embryo != NULL) && (param->attempt_embryonic_refinement) && (n < (param->population_embryo_contribution * param->population_size))));

         if (create_random_program)
         {
            while (true)
            {
               new_program = GeneticLinearProgram_NewRandomProgram(instruction_set,
                                                                   symbol_table,
                                                                   current_size_in_instructions);

               if (new_program != NULL) break;
            }
         }
         else
         {
            new_program = GeneticLinearProgram_Internal_NewProgram();

            new_program->main_program = InstructionString_DuplicateInstructionString(input_string_embryo);

            new_program->hits = 0;
         }

         output_report->total_candidates_processed++;

         assert(new_program != NULL);
         population[n] = new_program;
      }

      float best_fitness = 99999999999.0f;

      // Make an array of pointers to all our population
      // The sort it by fitness!
      SortedGeneticLinearProgramPopulation sorted_population;

      unsigned int new_creation = 0;

      // Find the fitness of all our programs!
      for (unsigned int p = 0; p < param->population_size; p++)
      {
         GeneticLinearProgram_FitnessEvaluationOnTrainingSet(population[p], param, fitness_case_training_set, symbol_table);

         output_report->total_instructions_executed += population[p]->main_program->size();
      }

      // Take pointers to all the GeneticLinearProgram * (these are pointers to pointers to GeneticLinearProgram)
      // Place them in automatically sorted linked list structure.
      for (unsigned int n = 0; n < param->population_size; n++)
      {
         SortedGeneticLinearProgramPopulation_Insert(sorted_population, &population[n]);
      }

      do
      {
         GeneticLinearProgram *best_program = **SortedGeneticLinearProgramPopulation_At(sorted_population, param->population_size-1);

         if (best_program->fitness < best_fitness)
         {
            best_fitness = best_program->fitness;
         }

         if ((new_creation % param->reporting_granularity_in_new_creations) == 0)
         {
            if ((clock() - report_time) > (param->reporting_delay_in_seconds * CLOCKS_PER_SEC))
            {
               if (reporting)
               {
                  system("cls");

                  printf("Trying to evolve program:\r\n");

                  ParseTreeNode_Print_AsReadableExpression(target_program, symbol_table);

                  printf("\r\n");

                  SymbolTable_Print(symbol_table);

                  printf("\r\n");

                  float time_in_seconds = ((float)(clock() - start_time)) / ((float)CLOCKS_PER_SEC);

                  printf("-[ TIME %8.1fs - NEW_CREATION %7d - EFF GEN %4d - BEST FITNESS %10.3f - RUN %4d ]-\r\n", time_in_seconds, new_creation, new_creation / param->population_size, best_fitness, run);

                  for (unsigned int p = param->population_size - 1; p < param->population_size; p++)
                  {
                     GeneticLinearProgram **program = *SortedGeneticLinearProgramPopulation_At(sorted_population, p);

                     printf("% 12.2f, %2d hits:\r\n", (*program)->fitness, (*program)->hits);
                     InstructionString_Print((*program)->main_program, symbol_table, true);
                  }

                  report_time = clock();
               }
            }
         }

         // Determine what method of repopulation is going to be used:
         enum REPOPULATION_METHOD
         {
            REPOPULATE_CROSSOVER,
            REPOPULATE_MUTATION,
            REPOPULATE_REPRODUCTION,
         };

         REPOPULATION_METHOD repopulation_method;
         unsigned int new_individuals_count;

         float r = random_float() * 0.9999f;
         // Rounding errors may make upper bound of topmost cumulative rate not equal to 1.0f
         // So we make do with this.

         if (r < CUMULATIVE_RATE_OF_CROSSOVER)
         {
            repopulation_method = REPOPULATE_CROSSOVER;
            new_individuals_count = 2;
         }
         else
         if (r < CUMULATIVE_RATE_OF_REPRODUCTION)
         {
            repopulation_method = REPOPULATE_REPRODUCTION;
            new_individuals_count = 1;
         }
         else
         if (r < CUMULATIVE_RATE_OF_MUTATION)
         {
            repopulation_method = REPOPULATE_MUTATION;
            new_individuals_count = 1;
         }
         else
         {
            assert(false);
         }

         // We're going to make new_individuals_count new programs!
         GeneticLinearProgram *new_individuals[new_individuals_count];

         switch (repopulation_method)
         {
            case (REPOPULATE_CROSSOVER) :
            {
               // Use tournament selection!

               // We're going to want a series of unique indices into the sorted
               // population array.

               unsigned int index_of_parent_1_in_sorted_population_array;
               unsigned int index_of_parent_2_in_sorted_population_array;

               {
                  IndexSeries *tournament_indices = IndexSeries_ReturnTournament(param->population_size, param->tournament_size_for_crossover);

                  // Choose the individual from our tournament with the highest fitness:
                  // Don't bother to sort the tournament, just do this:
                  unsigned int index_of_parent_1_in_tournament = 0;
                  float fitness_of_parent_1_in_tournament = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[index_of_parent_1_in_tournament]))->fitness;

                  for (unsigned int tournament_index_check = 1; tournament_index_check < param->tournament_size_for_crossover; tournament_index_check++)
                  {
                     float fitness_of_index_check_program = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[tournament_index_check]))->fitness;

                     if (fitness_of_index_check_program < fitness_of_parent_1_in_tournament)
                     {
                        fitness_of_parent_1_in_tournament = fitness_of_index_check_program;
                        index_of_parent_1_in_tournament = tournament_index_check;
                     }
                  }

                  index_of_parent_1_in_sorted_population_array = tournament_indices->value[index_of_parent_1_in_tournament];

                  IndexSeries_DestroyIndexSeries(tournament_indices);
               }

               // Now we know the index of the first parent.
               // Find the index of the second parent the same way.
               {
                  IndexSeries *tournament_indices = IndexSeries_ReturnTournament(param->population_size, param->tournament_size_for_crossover);

                  // Choose the individual from our tournament with the highest fitness:
                  // Don't bother to sort the tournament, just do this:
                  unsigned int index_of_parent_2_in_tournament = 0;
                  float fitness_of_parent_2_in_tournament = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[index_of_parent_2_in_tournament]))->fitness;

                  for (unsigned int tournament_index_check = 2; tournament_index_check < param->tournament_size_for_crossover; tournament_index_check++)
                  {
                     float fitness_of_index_check_program = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[tournament_index_check]))->fitness;

                     if (fitness_of_index_check_program < fitness_of_parent_2_in_tournament)
                     {
                        fitness_of_parent_2_in_tournament = fitness_of_index_check_program;
                        index_of_parent_2_in_tournament = tournament_index_check;
                     }
                  }

                  index_of_parent_2_in_sorted_population_array = tournament_indices->value[index_of_parent_2_in_tournament];

                  IndexSeries_DestroyIndexSeries(tournament_indices);
               }

               //printf("Before new program pair.\r\n");
               GeneticLinearProgram_Crossover_Pair new_program_pair = GeneticLinearProgram_Crossover2((**SortedGeneticLinearProgramPopulation_At(sorted_population, index_of_parent_1_in_sorted_population_array)),
                                                                                                      (**SortedGeneticLinearProgramPopulation_At(sorted_population, index_of_parent_2_in_sorted_population_array)),
                                                                                                      param->discard_crossover_programs_of_size_above);

               new_individuals[0] = new_program_pair.program[0];
               new_individuals[1] = new_program_pair.program[1];
            }
            break;
            case (REPOPULATE_REPRODUCTION) :
            {
               // Reproduce a program (chosen by tournament!).

               // We're going to want a series of unique indices into the sorted
               // population array.

               unsigned int index_of_parent_1_in_sorted_population_array;

               {
                  IndexSeries *tournament_indices = IndexSeries_ReturnTournament(param->population_size, param->tournament_size_for_reproduction);

                  // Choose the individual from our tournament with the highest fitness:
                  // Don't bother to sort the tournament, just do this:
                  unsigned int index_of_parent_1_in_tournament = 0;
                  float fitness_of_parent_1_in_tournament = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[index_of_parent_1_in_tournament]))->fitness;

                  for (unsigned int tournament_index_check = 1; tournament_index_check < param->tournament_size_for_reproduction; tournament_index_check++)
                  {
                     float fitness_of_index_check_program = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[tournament_index_check]))->fitness;

                     if (fitness_of_index_check_program < fitness_of_parent_1_in_tournament)
                     {
                        fitness_of_parent_1_in_tournament = fitness_of_index_check_program;
                        index_of_parent_1_in_tournament = tournament_index_check;
                     }
                  }

                  index_of_parent_1_in_sorted_population_array = tournament_indices->value[index_of_parent_1_in_tournament];

                  IndexSeries_DestroyIndexSeries(tournament_indices);
               }

               new_individuals[0] = GeneticLinearProgram_DuplicateProgram((**SortedGeneticLinearProgramPopulation_At(sorted_population, index_of_parent_1_in_sorted_population_array)));
            }
            break;
            case (REPOPULATE_MUTATION) :
            {
               // Mutate a program (chosen by tournament!).

               // We're going to want a series of unique indices into the sorted
               // population array.
               unsigned int index_of_parent_1_in_sorted_population_array;

               {
                  IndexSeries *tournament_indices = IndexSeries_ReturnTournament(param->population_size, param->tournament_size_for_mutation);

                  // Choose the individual from our tournament with the highest fitness:
                  // Don't bother to sort the tournament, just do this:
                  unsigned int index_of_parent_1_in_tournament = 0;
                  float fitness_of_parent_1_in_tournament = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[index_of_parent_1_in_tournament]))->fitness;

                  for (unsigned int tournament_index_check = 1; tournament_index_check < param->tournament_size_for_mutation; tournament_index_check++)
                  {
                     float fitness_of_index_check_program = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[tournament_index_check]))->fitness;

                     if (fitness_of_index_check_program < fitness_of_parent_1_in_tournament)
                     {
                        fitness_of_parent_1_in_tournament = fitness_of_index_check_program;
                        index_of_parent_1_in_tournament = tournament_index_check;
                     }
                  }

                  index_of_parent_1_in_sorted_population_array = tournament_indices->value[index_of_parent_1_in_tournament];

                  IndexSeries_DestroyIndexSeries(tournament_indices);
               }

               new_individuals[0] = GeneticLinearProgram_MutateProgram((**SortedGeneticLinearProgramPopulation_At(sorted_population, index_of_parent_1_in_sorted_population_array)), instruction_set, symbol_table);
            }
            break;
            default :
            {
               // Bad repopulation method... :/
               assert(false);
            }
         }

         // Now that the new individuals have been produced, we need to delete
         // new_individuals_count members of the old population to make space

         // Keep track of the iterators of the ones we've identified for deletion.
         SortedGeneticLinearProgramPopulation::iterator individuals_deleted_iterators[new_individuals_count];

         for (unsigned int replaced_individual = 0; replaced_individual < new_individuals_count; replaced_individual++)
         {
            // Use tournament selection to find the worst of a bunch.

            unsigned int index_of_parent_1_in_sorted_population_array;

            // We need to do a test to ensure that we don't try to delete
            // the same individual multiple times: we use acceptable index
            // to make sure that this index is different from all the previous
            // indices!
            int acceptable_index = 0;

            while (acceptable_index == 0)
            {
               IndexSeries *tournament_indices = IndexSeries_ReturnTournament(param->population_size, param->tournament_size_for_deletion);

               // Choose the individual from our tournament with the LOWEST fitness:
               // Don't bother to sort the tournament, just do this:
               unsigned int index_of_parent_1_in_tournament = 0;
               float fitness_of_parent_1_in_tournament = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[index_of_parent_1_in_tournament]))->fitness;

               for (unsigned int tournament_index_check = 1; tournament_index_check < param->tournament_size_for_deletion; tournament_index_check++)
               {
                  float fitness_of_index_check_program = (**SortedGeneticLinearProgramPopulation_At(sorted_population, tournament_indices->value[tournament_index_check]))->fitness;

                  if (fitness_of_index_check_program > fitness_of_parent_1_in_tournament)
                  {
                     fitness_of_parent_1_in_tournament = fitness_of_index_check_program;
                     index_of_parent_1_in_tournament = tournament_index_check;
                  }
               }

               index_of_parent_1_in_sorted_population_array = tournament_indices->value[index_of_parent_1_in_tournament];

               IndexSeries_DestroyIndexSeries(tournament_indices);

               acceptable_index = 1;

               for (unsigned int previous_individual = 0; previous_individual < replaced_individual; previous_individual++)
               {
                  if (SortedGeneticLinearProgramPopulation_At(sorted_population, index_of_parent_1_in_sorted_population_array) == individuals_deleted_iterators[previous_individual])
                  {
                     acceptable_index = 0;
                     break;
                  }
               }

               individuals_deleted_iterators[replaced_individual] = SortedGeneticLinearProgramPopulation_At(sorted_population, index_of_parent_1_in_sorted_population_array);
            }
         }

         // Keep track of the slots vacated by the programs we destroy.
         GeneticLinearProgram **vacant_slots[new_individuals_count];

         for (unsigned int replaced_individual = 0; replaced_individual < new_individuals_count; replaced_individual++)
         {
            // Delete the individual ITSELF, and retrieve its slot.
            vacant_slots[replaced_individual] = *individuals_deleted_iterators[replaced_individual];

            GeneticLinearProgram_DestroyProgram(*vacant_slots[replaced_individual]);

            *vacant_slots[replaced_individual] = new_individuals[replaced_individual];

            // Check the fitness of the new individuals.
            GeneticLinearProgram_FitnessEvaluationOnTrainingSet(new_individuals[replaced_individual], param, fitness_case_training_set, symbol_table);

            output_report->total_instructions_executed += new_individuals[replaced_individual]->main_program->size();
         }

         for (unsigned int replaced_individual = 0; replaced_individual < new_individuals_count; replaced_individual++)
         {
            // Delete the iterator indirector
            sorted_population.erase(individuals_deleted_iterators[replaced_individual]);

            SortedGeneticLinearProgramPopulation_Insert(sorted_population, vacant_slots[replaced_individual]);
         }

         new_creation += new_individuals_count;

         if (best_program->hits == param->fitness_case_training_set_count)
         {
            if (best_program->fitness <= param->maximum_acceptable_fitness)
            {
               if (GeneticLinearProgram_AcceptanceTestOnTestSet(best_program, param, fitness_case_test_set, symbol_table))
               {
                  // Copy the acceptable program to the exit vector pointer
                  // and stop the evolution!

                  acceptable_program_found = true;
                  acceptable_program = InstructionString_DuplicateInstructionString(best_program->main_program);

                  break;
               }
            }

            // Refinement mode.
            if ((param->maximum_acceptable_fitness < 0.0f)
             && (new_creation >= param->max_creations_per_run))
            {
               // Copy the acceptable program to the exit vector pointer
               // and stop the evolution!

               acceptable_program_found = true;
               acceptable_program = InstructionString_DuplicateInstructionString(best_program->main_program);

               break;
            }
         }

         assert(sorted_population.size() == param->population_size);
      }
      while (new_creation < param->max_creations_per_run);

      output_report->total_candidates_processed += new_creation;

//      print = true;
      for (unsigned int p = param->population_size - 3; p < param->population_size; p++)
      {
//         printf("% 8.2f:\r\n", (*sorted_population[p])->fitness);
//         InstructionString_Print((*sorted_population[p])->main_program);
//         printf("\r\n");
      }
//      printf("-----------------\r\n");
//      printf("End of run %4d!\r\n", run);
//      printf("New creation %4d.\r\n", new_creation);
//      printf("-----------------\r\n");

      for (unsigned int n = 0; n < param->population_size; n++)
      {
         GeneticLinearProgram_DestroyProgram(population[n]);
         population[n] = NULL;
      }

      delete[] population;

      if (acceptable_program_found) break;
   }

   if (acceptable_program_found)
   {
      printf("\r\n");
      printf("END EVOLUTION.\r\n");
      printf("\r\n");
      printf("Produced program:\r\n");
      printf("\r\n");
      InstructionString_Print(acceptable_program, symbol_table, true);
      printf("\r\n");

      *output_string = acceptable_program;

      FILE *f = fopen(param->output_file.c_str(), "ab");
      fprintf(f, "%d %d %d %d verified\r\n",
                  output_report->total_candidates_processed,
                     acceptable_program->size(),
                        output_report->total_instructions_executed,
                           output_report->runs_attempted);
      fclose(f);
   }
   else
   {
      printf("\r\n");
      printf("END EVOLUTION.\r\n");
      printf("\r\n");
      printf("No acceptable program found.\r\n");
      printf("\r\n");

      *output_string = NULL;

      FILE *f = fopen(param->output_file.c_str(), "ab");
      fprintf(f, "%d %d %d %d aborted\r\n",
                  output_report->total_candidates_processed,
                     -1,
                        output_report->total_instructions_executed,
                           output_report->runs_attempted);
      fclose(f);
   }

   if (param->interactive)
   {
      wait_for_return();
   }
}



// Augment the SymbolTable symbol_table with new symbols representing
// the intermediate values produced by evaluating the operator-like
// interior nodes in the given tree.
void EvolutionSystem_Master_PrepareSymbolTableEntriesForBranchingPoints(
       const std::string &prefix_of_current_node,
           ParseTreeNode *target_program,
             SymbolTable *symbol_table)
{
   const unsigned int depth = ParseTreeNode_GetHighestDepth(target_program);

   // Add yourself as node.
   SymbolTable_InsertSymbolTableEntry(symbol_table, prefix_of_current_node, SYMBOL_TABLE_VARIABLE_ARCHETYPE_VARIABLE);

   if (depth >= 3)
   {
      const unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(target_program);

      for (unsigned int child = 0; child < arity; child++)
      {
         if (ParseTreeNode_GetHighestDepth(target_program->as_interior_node.child_node[child]) >= 2)
         {
            char suffix[2] = {'0', 0};

            suffix[0] += child;

            EvolutionSystem_Master_PrepareSymbolTableEntriesForBranchingPoints(prefix_of_current_node + suffix,
                                                                               target_program->as_interior_node.child_node[child],
                                                                               symbol_table);
         }
      }
   }
}

// The final result of this function will be the creation of an InstructionString
// containing the appended programs necessary to create the full program by
// considering partial programs and combining them.
// Construct a new parse tree program encapsulating the immediate actions
// of the root node of the given parse tree, then evolve a program that
// performs that action.
// Recursively evolve child programs.
void EvolutionSystem_EvolveInstructionStringFromParseTree_PerformIncrement(
               const std::string  &prefix_of_current_node,
                   ParseTreeNode  *target_program,
          EvolutionSystem_Report  *output_report,
               InstructionString **output_string,
      EvolutionSystem_Parameters  *param,
     InstructionSetProbabilistic  *instruction_set,
                     SymbolTable  *symbol_table)
{
   SymbolTable local_symbol_table = *symbol_table;

   const unsigned int arity = ParseTreeNode_OperatorLikeSymbol_ReturnArity(target_program);

   InstructionString *child_programs[arity];

   for (unsigned int child = 0; child < arity; child++)
   {
      child_programs[child] = NULL;

      if (target_program->as_interior_node.child_node[child]->nature == ParseTreeNode::NODE_NATURE_INTERIOR_NODE)
      {
         char suffix[2] = {'0', 0};

         suffix[0] += child;

         EvolutionSystem_EvolveInstructionStringFromParseTree_PerformIncrement(
                    prefix_of_current_node + suffix,
                    target_program->as_interior_node.child_node[child],
                    output_report,
                    &child_programs[child],
                    param,
                    instruction_set,
                    symbol_table);

         //printf("Child %d has a program of length %d.\r\n", child, InstructionString_ReturnLengthInInstructions(child_programs[child]));

         //InstructionString_Print(child_programs[child], &local_symbol_table, true);
      }
   }

   InstructionString *my_program;

   for (SymbolTableData::iterator i  = local_symbol_table.symbol_table_data.begin();
                                  i != local_symbol_table.symbol_table_data.end();
                                  i++)
   {
      SymbolicVariableData_ApplyArchetype(&(i->second), SYMBOL_TABLE_VARIABLE_ARCHETYPE_BYSTANDER);
   }

   SymbolicVariableData_ApplyArchetype(
      SymbolTable_QuerySymbolTableVarname(
           &local_symbol_table,
           prefix_of_current_node.c_str()),
      SYMBOL_TABLE_VARIABLE_ARCHETYPE_RESULT);

   {
      // Construct the microprogram that we're evolving.
      ParseTreeSourceProgram *parse_tree_micro_program = new ParseTreeSourceProgram;
      parse_tree_micro_program->symbol_table = new SymbolTable; // Not used.

      ParseTreeInteriorNodeSetMember *operator_like_symbol_assignment = ParseTreeInteriorNodeSetMember_NewParseTreeInteriorNodeSetMember_OperatorLikeSymbol(PARSE_TREE_OPERATOR_LIKE_SYMBOL_ASSIGNMENT);

      parse_tree_micro_program->interior_node_set["="] = operator_like_symbol_assignment;

      // Construct the master things.

      std::map<SYMBOL_TABLE_KEY, ParseTreeLeafNodeSetMember *> leaf_node_by_varkey;

      for (SymbolTableData::iterator r  = local_symbol_table.symbol_table_data.begin();
                                     r != local_symbol_table.symbol_table_data.end();
                                     r++)
      {
         SYMBOL_TABLE_KEY const &varkey = r->first;

         // The master nodes for all the varnames are contained in the
         // leaf_node_by_varkey map.
         leaf_node_by_varkey[varkey] = ParseTreeLeafNodeSetMember_NewParseTreeLeafNodeSetMember_VariableFromSymbolTableKey(varkey);

         // Put all of the master nodes into the leaf_node_set so
         // they can be easily deleted later.
         parse_tree_micro_program->leaf_node_set[varkey] = leaf_node_by_varkey[varkey];
      }

      ParseTreeNode *&assignment_root = parse_tree_micro_program->target_program;
      assignment_root = ParseTreeNode_NewIncompleteInteriorNodeFromParseTreeInteriorNodeSetMember(operator_like_symbol_assignment);

      assignment_root->as_interior_node.child_node[0] = ParseTreeNode_NewLeafNodeFromParseTreeLeafNodeSetMember(leaf_node_by_varkey[SymbolTable_ReverseLookupKeyFromVarname(&local_symbol_table, prefix_of_current_node.c_str())]);
      assignment_root->as_interior_node.child_node[1] = ParseTreeNode_DuplicateNode(target_program);

      for (unsigned int child = 0; child < arity; child++)
      {
         bool done_already = false;

         ParseTreeNode *&t = assignment_root->as_interior_node.child_node[1]->as_interior_node.child_node[child];

         std::string way;

         if (target_program->as_interior_node.child_node[child]->nature == ParseTreeNode::NODE_NATURE_LEAF_NODE)
         {
            t = ParseTreeNode_DuplicateTree(target_program->as_interior_node.child_node[child]);

            if (child == 0)
            {
               if (target_program->as_interior_node.interior_node_master->as_operator_like_symbol.operator_like_symbol == PARSE_TREE_OPERATOR_LIKE_SYMBOL_ASSIGNMENT)
               {
                  const ParseTreeLeafNodeSetMember *&lsm = t->as_leaf_node.leaf_node_master;
                  SYMBOL_TABLE_KEY symbol_table_key_for_this_child_node = lsm->as_varname.varname_symbol_table_key;

                  SymbolicVariableData_ApplyArchetype(SymbolTable_QuerySymbolTableKey(&local_symbol_table, symbol_table_key_for_this_child_node), SYMBOL_TABLE_VARIABLE_ARCHETYPE_RESULT);

                  done_already = true;
               }
            }
         }
         else
         {

            char suffix[2] = {'0', 0};

            suffix[0] += child;

            std::string child_node = prefix_of_current_node + suffix;

            t = ParseTreeNode_NewLeafNodeFromParseTreeLeafNodeSetMember(leaf_node_by_varkey[SymbolTable_ReverseLookupKeyFromVarname(&local_symbol_table, child_node.c_str())]);
         }

         const ParseTreeLeafNodeSetMember *&lsm = t->as_leaf_node.leaf_node_master;
         SYMBOL_TABLE_KEY symbol_table_key_for_this_child_node = lsm->as_varname.varname_symbol_table_key;

         if (!done_already)
         {
            SymbolicVariableData_ApplyArchetype(SymbolTable_QuerySymbolTableKey(&local_symbol_table, symbol_table_key_for_this_child_node), SYMBOL_TABLE_VARIABLE_ARCHETYPE_CONSTANT);
         }
      }

//      ParseTreeNode_Print_AsReadableExpression(assignment_root, symbol_table);
//      printf("\r\n");

      // Do evolve.
      EvolutionSystem_EvolveInstructionStringFromParseTree(
         parse_tree_micro_program->target_program,
         output_report,
         &my_program,
         param,
         false,
         instruction_set,
         &local_symbol_table,
         NULL);

      // Destroy program.
      ParseTreeSourceProgram_DestroySourceProgram(parse_tree_micro_program);
   }

   // Combine the programs we have.
   InstructionString *output_program = NULL;

   for (unsigned int child = 0; child < arity; child++)
   {
      if (child_programs[child] != NULL)
      {
         if (output_program == NULL)
         {
            output_program = InstructionString_DuplicateInstructionString(child_programs[child]);
         }
         else
         {
            InstructionString_ConcatenateInstructionStringsInPlace(output_program, child_programs[child]);
         }

         delete child_programs[child];
      }
   }

   // We've accumulated all the child programs and freed them.

   if (output_program == NULL)
   {
      output_program = InstructionString_DuplicateInstructionString(my_program);
   }
   else
   {
      InstructionString_ConcatenateInstructionStringsInPlace(output_program, my_program);
   }

   // Now output_program has all the program in it.

      FILE *f = fopen(param->output_file.c_str(), "ab");
      fprintf(f, "constructed %s\r\n",
                  ParseTreeNode_Print_ToString(target_program, &local_symbol_table).c_str());
      fprintf(f, "length %d\r\n",
                  output_program->size());
      fclose(f);

   *output_string = output_program;
}

// Attempt to evolve an InstructionString with the same semantics as
// the input program using an incremental recursive approach.
void EvolutionSystem_EvolveInstructionStringFromParseTree_Incremental(
                   ParseTreeNode  *target_program,
          EvolutionSystem_Report  *output_report,
               InstructionString **output_string,
      EvolutionSystem_Parameters  *param,
     InstructionSetProbabilistic  *instruction_set,
                     SymbolTable  *symbol_table)
{
   // We have received the master program that we want to evolve

   // and the global symbol table that contains information about
   // all the symbolic type things that appear in the tree.

   // Copy the symbol table.
   SymbolTable local_symbol_table = *symbol_table;

   // Prepare new symbol table entries for the
   // branching points.

   std::string MASTER_BRANCH_SYMBOLIC_VARIABLE = "M0";

   EvolutionSystem_Master_PrepareSymbolTableEntriesForBranchingPoints(MASTER_BRANCH_SYMBOLIC_VARIABLE, target_program, &local_symbol_table);

   // Given all we've prepared, do the actual evolution.

   // Debug out the symbol table.
   for (SymbolTableData::iterator i  = (&local_symbol_table)->symbol_table_data.begin();
                                  i != (&local_symbol_table)->symbol_table_data.end();
                                  i++)
   {
      printf("Symbol table key, name: %4d - %10s\r\n", i->first, i->second.variable_symbolic_name.c_str());
   }

   InstructionString *my_program;

   EvolutionSystem_EvolveInstructionStringFromParseTree_PerformIncrement(
              MASTER_BRANCH_SYMBOLIC_VARIABLE,
              target_program,
              output_report,
              &my_program,
              param,
              instruction_set,
              &local_symbol_table);

   *output_string = my_program;

   *symbol_table = local_symbol_table;
}

// Perform training and test set analysis on an InstructionString
// and determine if it passes both. This is used to test
// the viability of composite programs produced by
// incremental evolution.
void EvolutionSystem_VerifyInstructionStringFromParseTree(
                   ParseTreeNode  *target_program,
          EvolutionSystem_Report  *output_report,
               InstructionString  *input_string,
      EvolutionSystem_Parameters  *param,
                     SymbolTable  *symbol_table)
{
   assert(target_program  != NULL);
   assert(output_report   != NULL);
   assert(input_string    != NULL);
   assert(param           != NULL);
   assert(symbol_table    != NULL);

   INSTRUCTION_EXECUTIONS_MAX_COUNT = param->instruction_executions_max_count;

   const unsigned int FITNESS_CASE_TOTAL_COUNT = param->fitness_case_training_set_count
                                                + param->fitness_case_test_set_count;

   FitnessCase fitness_case[FITNESS_CASE_TOTAL_COUNT];

   FitnessCase *fitness_case_training_set = fitness_case;
   FitnessCase *fitness_case_test_set     = fitness_case + param->fitness_case_training_set_count;

   FitnessCase_InitialiseCases(target_program, symbol_table, fitness_case, FITNESS_CASE_TOTAL_COUNT);

   //printf("Testing fitness on training set for combined program:\r\n");

   GeneticLinearProgram p;

   p.main_program = input_string;

   GeneticLinearProgram_FitnessEvaluationOnTrainingSet(&p, param, fitness_case_training_set, symbol_table, true);

   if (p.hits != param->fitness_case_training_set_count)
   {
      //printf("Did not pass training set. (%d of %d)\r\n", p.hits, param->fitness_case_training_set_count);
      output_report->success_on_training_set = false;
   }
   else
   {
      //printf("Passed training set.\r\n");
      output_report->success_on_training_set = true;
   }

   if (!GeneticLinearProgram_AcceptanceTestOnTestSet(&p, param, fitness_case_test_set, symbol_table))
   {
      //printf("Did not pass test set.\r\n");
      //system("pause");
      output_report->success_on_test_set = false;
   }
   else
   {
      //printf("Passed test set! =^_^=\r\n");
      output_report->success_on_test_set = true;
   }
}


///////////////////

int main(int argc, char *argv[])
{
   if (argc < 6)
   {
      printf("Usage:\r\nprogram.exe source_program_file analysis_type parameters_file parameters_file2 output_file");
      exit(0);
   }

   std::string source_program_file        = std::string(argv[1]);
   std::string analysis_type              = std::string(argv[2]);
   std::string parameters_file_evolution  = std::string(argv[3]);
   std::string parameters_file_refinement = std::string(argv[4]);
   std::string output_file                = std::string(argv[5]);

   ParseTreeSourceProgram *parse_tree_source_program = ParseTreeSourceProgram_ConstructSourceProgramFromFile(source_program_file);

   EvolutionSystem_Parameters param_evolution  = EvolutionSystem_Parameters_LoadFromFile(parameters_file_evolution);
   EvolutionSystem_Parameters param_refinement = EvolutionSystem_Parameters_LoadFromFile(parameters_file_refinement);

   param_evolution.output_file  = output_file;
   param_refinement.output_file = output_file;

   InstructionSetProbabilistic instruction_set;

   instruction_set.push_back(InstructionProbability(VMI_A_ADD,   1.00f));
   instruction_set.push_back(InstructionProbability(VMI_A_SUB,   1.00f));
   instruction_set.push_back(InstructionProbability(VMI_A_MUL,   1.00f));
   instruction_set.push_back(InstructionProbability(VMI_A_DIVP,  1.00f));
   instruction_set.push_back(InstructionProbability(VMI_V_LOADV, 1.00f));
   instruction_set.push_back(InstructionProbability(VMI_M_LOADS, 1.00f));
   instruction_set.push_back(InstructionProbability(VMI_M_STORS, 1.00f));

   InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(&instruction_set);

   VirtualMachine_InitialiseStaticMnemonics();

   EvolutionSystem_Report report;

   // For this steady state experiment, the variable 'new_creation' refers to
   // the number of new entities created thus far in the steady state system.

   if ((analysis_type == "attempt_incremental")
    || (analysis_type == "attempt_incremental_with_refine"))
   {
      for (unsigned int attempt = 0; attempt < param_evolution.repeat_experiment; attempt++)
      {
         {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
         fprintf(f, "experiment %d\r\n", attempt);
         fclose(f);}

         InstructionString *output_instruction_string;

         EvolutionSystem_EvolveInstructionStringFromParseTree_Incremental(
            parse_tree_source_program->target_program,
            &report,
            &output_instruction_string,
            &param_evolution,
            &instruction_set,
            parse_tree_source_program->symbol_table);

         printf("Final result!\r\n");

         InstructionString_Print(output_instruction_string, parse_tree_source_program->symbol_table, true);

         SymbolTable_Print(parse_tree_source_program->symbol_table);

         if (0) /* This is already performed by the evolutionary system.*/
         EvolutionSystem_VerifyInstructionStringFromParseTree(
            parse_tree_source_program->target_program,
            &report,
            output_instruction_string,
            &param_evolution,
            parse_tree_source_program->symbol_table);

         if (analysis_type == "attempt_incremental_with_refine")
         {
            for (unsigned int attempt_refine = 0; attempt_refine < param_refinement.repeat_experiment; attempt_refine++)
            {
               {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
               fprintf(f, "refinement %d\r\n", attempt_refine);
               fclose(f);}

               InstructionString *output_refined_string;
               EvolutionSystem_Report report2;

               EvolutionSystem_EvolveInstructionStringFromParseTree(
                  parse_tree_source_program->target_program,
                  &report2,
                  &output_refined_string,
                  &param_refinement,
                  param_refinement.reporting,
                  &instruction_set,
                  parse_tree_source_program->symbol_table,
                  output_instruction_string);

               if (output_refined_string != NULL)
               {
                  if (param_evolution.interactive)
                  {
                     printf("Produced output:\r\n\r\n");

                     InstructionString_Print(output_refined_string, parse_tree_source_program->symbol_table, true);

                     wait_for_return();
                  }

                  delete output_refined_string;
               }

               {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
               fprintf(f, "refinement_over %d\r\n", attempt_refine);
               fclose(f);}
            }
         }

         if (output_instruction_string != NULL)
         {
            delete output_instruction_string;
         }

         {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
         fprintf(f, "experiment_over %d\r\n", attempt);
         fclose(f);}
      }
   }
   else
   if ((analysis_type == "attempt_standard")
    || (analysis_type == "attempt_standard_with_refine"))
   {
      for (unsigned int attempt = 0; attempt < param_evolution.repeat_experiment; attempt++)
      {
         {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
         fprintf(f, "experiment %d\r\n", attempt);
         fclose(f);}

         report = EvolutionSystem_Report();

         InstructionString *output_instruction_string;

         EvolutionSystem_EvolveInstructionStringFromParseTree(
            parse_tree_source_program->target_program,
            &report,
            &output_instruction_string,
            &param_evolution,
            param_evolution.reporting,
            &instruction_set,
            parse_tree_source_program->symbol_table,
            NULL);

         if (analysis_type == "attempt_standard_with_refine")
         {
            for (unsigned int attempt_refine = 0; attempt_refine < param_refinement.repeat_experiment; attempt_refine++)
            {
               {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
               fprintf(f, "refinement %d\r\n", attempt_refine);
               fclose(f);}

               InstructionString *output_refined_string;
               EvolutionSystem_Report report2;

               EvolutionSystem_EvolveInstructionStringFromParseTree(
                  parse_tree_source_program->target_program,
                  &report2,
                  &output_refined_string,
                  &param_refinement,
                  param_refinement.reporting,
                  &instruction_set,
                  parse_tree_source_program->symbol_table,
                  output_instruction_string);

               if (output_refined_string != NULL)
               {
                  if (param_evolution.interactive)
                  {
                     printf("Produced output:\r\n\r\n");

                     InstructionString_Print(output_refined_string, parse_tree_source_program->symbol_table, true);

                     wait_for_return();
                  }

                  delete output_refined_string;
               }

               {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
               fprintf(f, "refinement_over %d\r\n", attempt_refine);
               fclose(f);}
            }
         }

         if (output_instruction_string != NULL)
         {
            delete output_instruction_string;
         }

         {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
         fprintf(f, "experiment_over %d\r\n", attempt);
         fclose(f);}
      }
   }
   else
   if ((analysis_type == "tree_walking_compiler")
    || (analysis_type == "tree_walking_compiler_with_refine"))
   {
      InstructionString *compiled_instruction_string = new InstructionString;

      TreeWalkingCompiler_BasicCompileParseTreeSourceProgram(parse_tree_source_program,
                                                             parse_tree_source_program->target_program,
                                                             0,
                                                             compiled_instruction_string);

      printf("Final result!\r\n");
      printf("Tree walker's version:\r\n\r\n");

      InstructionString_Print(compiled_instruction_string, parse_tree_source_program->symbol_table, true);

      SymbolTable_Print(parse_tree_source_program->symbol_table);

      if (param_evolution.interactive)
      {
         wait_for_return();
      }

      FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
      fprintf(f, "treewalk %s\r\n",
                  ParseTreeNode_Print_ToString(parse_tree_source_program->target_program, parse_tree_source_program->symbol_table).c_str());
      fprintf(f, "%d %d\r\n",
                  compiled_instruction_string->size(),
                     InstructionString_ReturnHighestRegisterUsed(compiled_instruction_string));
      fclose(f);

      if (analysis_type == "tree_walking_compiler_with_refine")
      {
         for (unsigned int attempt_refine = 0; attempt_refine < param_refinement.repeat_experiment; attempt_refine++)
         {
            {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
            fprintf(f, "refinement %d\r\n", attempt_refine);
            fclose(f);}

            InstructionString *output_refined_string;
            EvolutionSystem_Report report2;

            EvolutionSystem_EvolveInstructionStringFromParseTree(
               parse_tree_source_program->target_program,
               &report2,
               &output_refined_string,
               &param_refinement,
               param_refinement.reporting,
               &instruction_set,
               parse_tree_source_program->symbol_table,
               compiled_instruction_string);

            if (param_evolution.interactive)
            {
               printf("Produced output:\r\n\r\n");

               InstructionString_Print(output_refined_string, parse_tree_source_program->symbol_table, true);

               wait_for_return();
            }

            if (output_refined_string != NULL)
            {
               delete output_refined_string;
            }

            {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
            fprintf(f, "refinement_over %d\r\n", attempt_refine);
            fclose(f);}
         }
      }

      delete compiled_instruction_string;
   }
   else
   if (analysis_type == "random_search")
   {
      vm_register_general_purpose_count = param_evolution.vm_register_general_purpose_count;
      vm_register_final_general_purpose = vm_register_general_purpose_count - 1;

      INSTRUCTION_EXECUTIONS_MAX_COUNT = param_evolution.instruction_executions_max_count;

   EvolutionSystem_Parameters *param = &param_evolution;
   ParseTreeNode *target_program = parse_tree_source_program->target_program;

   InstructionSetProbabilistic instruction_set_adapted;

   if (param->adaptive_instruction_set)
   {
      InstructionSetProbabilistic adaptation_contribution;

      std::map<const ParseTreeInteriorNodeSetMember *, unsigned int> interior_node_population;

      ParseTreeSourceProgram_CalculateOperatorLikeSymbolPopulations(target_program, interior_node_population);

      for (std::map<const ParseTreeInteriorNodeSetMember *, unsigned int>::iterator i  = interior_node_population.begin();
                                                                                    i != interior_node_population.end();
                                                                                    i++)
      {
         switch (i->first->as_operator_like_symbol.operator_like_symbol)
         {
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_PLUS) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_ADD,  (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MINUS) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_SUB,  (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_MULTIPLY) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_MUL,  (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_DIVIDE) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_A_DIVP, (float)i->second));
            }
            break;
            case (PARSE_TREE_OPERATOR_LIKE_SYMBOL_ASSIGNMENT) :
            {
               adaptation_contribution.push_back(InstructionProbability(VMI_M_STORS, (float)i->second));
            }
            break;
            default :
            {
               // Don't care.
            }
         }
      }
      InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(&adaptation_contribution);

      instruction_set_adapted = instruction_set;
      instruction_set_adapted.insert(instruction_set_adapted.end(), adaptation_contribution.begin(), adaptation_contribution.end());

      InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(&instruction_set_adapted);
   }
   else
   {
      instruction_set_adapted = instruction_set;

      InstructionSetProbabilistic_NormaliseAndCalculateCumulatives(&instruction_set_adapted);
   }

      EvolutionSystem_Report r;

      for (unsigned int length_in_instructions = param_evolution.initial_minimum_program_size_in_instructions;
                        length_in_instructions <= param_evolution.initial_maximum_program_size_in_instructions;
                        length_in_instructions++)
      {
         unsigned int successes = 0;

         for (unsigned int retry = 0; retry < param_evolution.repeat_experiment; retry++)
         {
            InstructionString *compiled_instruction_string =
                  InstructionString_NewRandomInstructionStringFromInstructionSetProbabilistic(&instruction_set_adapted,
                                                                                              parse_tree_source_program->symbol_table,
                                                                                              length_in_instructions);

            EvolutionSystem_VerifyInstructionStringFromParseTree(parse_tree_source_program->target_program,
                                                                 &r,
                                                                 compiled_instruction_string,
                                                                 &param_evolution,
                                                                 parse_tree_source_program->symbol_table);

            if (r.success_on_test_set)
            {
               successes++;
            }

            delete compiled_instruction_string;
         }

         {FILE *f = fopen(param_evolution.output_file.c_str(), "ab");
         fprintf(f, "at_length,%d,successes,%d,attempts,%d\r\n", length_in_instructions, successes, param_evolution.repeat_experiment);
         fclose(f);}
      }
   }
   else
   {
      printf("I didn't understand your request.\r\n");exit(0);
   }

   ParseTreeSourceProgram_DestroySourceProgram(parse_tree_source_program);
}
